/************************************************************************/
/* Accumulator.h: class Accumulator					*/
/************************************************************************/

/* An Accumulator is a simple device for computing running mean values
   and root-mean-square deviations. This implementation is based on
   Knuth's method.
   (-> https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance)

   This could also be set up as a template class, but so far we never
   had any use for averages of types other than double precision.
*/

#ifndef ACCUMULATOR_H
#define ACCUMULATOR_H

class Accumulator {
  private:
    long long int n;			// number of entries
    double xq,				// present mean value
      q2;				// aux. storage for rms. deviation

  public:
    Accumulator() {			// constructor
      n = 0;
      xq = q2 = 0.0;	
    }
    // rely on default destructor and copy constructor

    void add(double x) noexcept {	// add an entry
      double dx = x - xq;
      n++;
      xq += dx / n;
      q2 += dx * (x - xq);
    }

    int count() const noexcept {	// report the number of entries
      return n;
    }

    double mean() const {		// report the mean value
      if (n < 1) throw MthException(MthException::n_data);
      return xq;
    }
    double mean(int reset) {
      double x = xq;
      if (n < 1) throw MthException(MthException::n_data);
      if (reset) {
	n = 0;
	xq = q2 = 0.0;
      }
      return x;
    }

    void merge(const Accumulator a) noexcept {
      double d = a.xq - xq,
        h = static_cast<double>(a.n) / static_cast<double>(a.n + n);
      xq += h * d;
      q2 += a.q2 + n * h * sq(d);
      n += a.n;
    }
    void merge(int nh, double sum1, double sum2) noexcept {
      double xqh, d,			// Sum(x_i), Sum(x_i^2)
        h = static_cast<double>(nh) / static_cast<double>(nh + n);
      if (nh > 0) {
	xqh = sum1 / nh,
        d = xqh - xq,
	xq += h * d;
	q2 += sum2 - sum1 * xqh + n * h * sq(d);
	n += nh;
      }
    }

    void reset() noexcept {		// reset
      n = 0;
      xq = q2 = 0.0;
    }

    double rms() const {		// report the rms. value
      if (n < 1) throw MthException(MthException::n_data);
      return sqrt(q2 / n);
    }
    double rms(int reset) {
      if (n < 1) throw MthException(MthException::n_data);
      double x = sqrt(q2 / n);
      if (reset) {
	n = 0;
	xq = q2 = 0.0;
      }
      return x;
    }
};
#endif
