/************************************************************************/
/* Matrix.h: template class Matrix					*/
/************************************************************************/

/* Matrix is implemented as a derivative of Vector: A 2-dimensional matrix
   is a vector of vectors.
*/

#ifndef MATRIX_H
#define MATRIX_H


template <typename T>
class Matrix : public Vector<Vector<T> > {
  private:
    int ncol;

  public:
    typedef Vector<Vector<T> > super;
    Matrix(int i = 0, int j = 0);	// constructor

    int rows() const noexcept {		// show number of rows
      return super::size();
    }

    int cols() const noexcept {		// show number of columns
      return ncol;
    }

    T det() const;			// determinant
    Matrix<T>& I() noexcept;		// set to identity matrix
    void redim(int nrnew, int ncnew) noexcept;	// change matrix size
    void size(int nrnew, int ncnew) noexcept;	// change matrix size
    Matrix<T>& operator=(T s) noexcept;	// preset
    Matrix<T>& operator+=(T s) noexcept;// increment with scalar
    Matrix<T>& operator+=(const Matrix<T>& m);
    Matrix<T>& operator-=(T s) noexcept;// decrement with scalar
    Matrix<T>& operator-=(const Matrix<T>& m);
    Matrix<T>& operator*=(T s) noexcept;// multiply with scalar
    Matrix<T>& operator/=(T s) noexcept;// divide by scalar
};

// constructor
template <typename T>
Matrix<T>::Matrix(int i, int j) : Vector<Vector<T> >(i), ncol(j) {
  int k;
#if 0				// from U. Breymann, "C++ Einführung ..."
  Vector<T> h(j);
  for (k = 0; k < i; k++) super::operator[](k) = h;
#else
  for (k = 0; k < i; k++) (super::operator[](k)).size(j);
#endif
}

// determinant (by Laplace's formula: very slow for large n)
template <typename T>
T Matrix<T>::det() const {
  if (super::size() != ncol) throw MthException(MthException::square);
  switch (ncol) {
    case 0:
      return static_cast<T> (0);
    case 1:
      return super::operator[](0)[0];
    case 2:
      return super::operator[](0)[0] * super::operator[](1)[1]
      - super::operator[](0)[1] * super::operator[](1)[0];
    default:
      int i, j, jh, k, s;
      T h;
      Matrix<T> m(ncol - 1, ncol - 1);
      h = static_cast<T> (0);
      for (k = 0, s = 1; k < ncol; k++) {
	for (i = 1; i < ncol; i++) {
	  for (j = 0, jh = 0; j < ncol; j++) {
	    if (j == k) continue;
	    m[i - 1][jh++] = super::operator[](i)[j];
	  }
	}
	h += s * super::operator[](0)[k] * m.det();
	s = -s;
      }
      return h;
  }
}

// set to identity matrix
template <typename T>
Matrix<T>& Matrix<T>::I() noexcept {
  int i, j;
  for (i = 0; i < super::size(); i++) {
    for (j = 0; j < ncol; j++) {
      super::operator[](i)[j] = (i == j) ? static_cast<T> (1)
      : static_cast<T> (0);
    }
  }
  return *this;
}

// change matrix size (kept merely for compatibility)
template <typename T>
void Matrix<T>::redim(int nrnew, int ncnew) noexcept {
#if 0	// old version
  int i,
    nr = super::size();
  if (nrnew <= 0 || ncnew <= 0) {	// delete all content
    for (i = 0; i < nr; i++) super::operator[](i).redim(0);
    super::redim(0);
  }
  else {
    if (nrnew < nr) {			// fewer rows: redim, delete spares
      for (i = 0; i < nrnew; i++) super::operator[](i).redim(ncnew);
      super::redim(nrnew);
    }
    else if (nrnew == nr) {		// same no. of rows: redim
      for (i = 0; i < nrnew; i++) super::operator[](i).redim(ncnew);
    }
    else {				// more rows: redim existing rows
      Vector<T> h(ncnew);		// add empty additional ones
      super::redim(nrnew);
      for (i = 0; i < nr; i++) super::operator[](i).redim(ncnew);
      for (i = nr; i < nrnew; i++) super::operator[](i) = h;
    }
    ncol = ncnew;
  }
#endif
  this->size(nrnew, ncnew);
}

// change matrix size
template <typename T>
void Matrix<T>::size(int nrnew, int ncnew) noexcept {
  int i,
    nr = super::size();
  if (nrnew <= 0 || ncnew <= 0) {	// delete all content
    for (i = 0; i < nr; i++) super::operator[](i).size(0);
    super::size(0);
  }
  else {
    if (nrnew < nr) {			// fewer rows: resize, delete spares
      for (i = 0; i < nrnew; i++) super::operator[](i).size(ncnew);
      super::size(nrnew);
    }
    else if (nrnew == nr) {		// same no. of rows: resize
      for (i = 0; i < nrnew; i++) super::operator[](i).size(ncnew);
    }
    else {				// more rows: resize existing rows
      Vector<T> h(ncnew);		// add empty additional ones
      super::size(nrnew);
      for (i = 0; i < nr; i++) super::operator[](i).size(ncnew);
      for (i = nr; i < nrnew; i++) super::operator[](i) = h;
    }
    ncol = ncnew;
  }
}

// "m = s": set all elements to a fixed (scalar) value
template <typename T>
Matrix<T>& Matrix<T>::operator=(T s) noexcept {
  int i;
  for (i = 0; i < super::size(); i++) super::operator[](i) = s;
  return *this;
}

// "m += s": increment matrix elements with scalar s
template <typename T>
Matrix<T>& Matrix<T>::operator+=(T s) noexcept {
  int i;
  for (i = 0; i < super::size(); i++) super::operator[](i) += s;
  return *this;
}

// "m += mh": increment matrix elements with elements of another matrix
template <typename T>
Matrix<T>& Matrix<T>::operator+=(const Matrix<T>& m) {
  int i;
  if (super::size() != m.rows()) throw MthException(MthException::dimensions);
  for (i = 0; i < super::size(); i++) super::operator[](i) += m[i];
  return *this;
}

// "m -= s": decrement matrix elements with scalar s
template <typename T>
Matrix<T>& Matrix<T>::operator-=(T s) noexcept {
  int i;
  for (i = 0; i < super::size(); i++) super::operator[](i) -= s;
  return *this;
}

// "m -= mh": decrement matrix elements with elements of another matrix
template <typename T>
Matrix<T>& Matrix<T>::operator-=(const Matrix<T>& m) {
  int i;

  if (super::size() != m.rows()) throw MthException(MthException::dimensions);
  for (i = 0; i < super::size(); i++) super::operator[](i) -= m[i];
  return *this;
}

// "m *= s": multiply matrix v with scalar s
template <typename T>
Matrix<T>& Matrix<T>::operator*=(T s) noexcept {
  int i;

  for (i = 0; i < super::size(); i++) super::operator[](i) *= s;
  return *this;
}

// "m /= s": divide matrix elements by scalar s
template <typename T>
Matrix<T>& Matrix<T>::operator/=(T s) noexcept {
  int i;
  for (i = 0; i < super::size(); i++) super::operator[](i) /= s;
  return *this;
}


// global functions for operator overloading (binary operations)
template <typename T>
Matrix<T> operator+(const Matrix<T>& m, T s) noexcept;
template <typename T>
Matrix<T> operator+(T s, const Matrix<T>& m) noexcept;
template <typename T>
Matrix<T> operator+(const Matrix<T>& m, const Matrix<T>& mh);
template <typename T>
Matrix<T> operator-(const Matrix<T>& m, const Matrix<T>& mh);
template <typename T>
Matrix<T> operator*(const Matrix<T>& m, T s) noexcept;
template <typename T>
Matrix<T> operator*(T s, const Matrix<T>& m) noexcept;
template <typename T>
Matrix<T> operator*(const Matrix<T>& m, const Matrix<T>& mh);
template <typename T>
Matrix<T> operator&(const Matrix<T>& m, const Matrix<T>& mh);
template <typename T>
Vector<T> operator*(const Matrix<T>& m, const Vector<T>& v);

// "m + s", "s + m": sum of matrix m and scalar s
template <typename T>
Matrix<T> operator+(const Matrix<T>& m, T s) noexcept {
  Matrix<T> h = m;
  h += s;
  return h;
}

template <typename T>
Matrix<T> operator+(T s, const Matrix<T>& m) noexcept {
  Matrix<T> h = m;
  h += s;
  return h;
}

// "m + mh": sum of matrices m and mh
template <typename T>
Matrix<T> operator+(const Matrix<T>& m, const Matrix<T>& mh) {
  Matrix<T> h(m.rows(), m.cols());
  int i, j;

  if (m.rows() != mh.rows() || m.cols() != mh.cols())
    throw MthException(MthException::dimensions);
  for (i = 0; i < m.rows(); i++) {
    for (j = 0; j < m.cols(); j++) h[i][j] = m[i][j] + mh[i][j];
  }
  return h;
}

// "m - mh": difference of matrices m and mh
template <typename T>
Matrix<T> operator-(const Matrix<T>& m, const Matrix<T>& mh) {
  Matrix<T> h(m.rows(), m.cols());
  int i, j;

  if (m.rows() != mh.rows() || m.cols() != mh.cols())
    throw MthException(MthException::dimensions);
  for (i = 0; i < m.rows(); i++) {
    for (j = 0; j < m.cols(); j++) h[i][j] = m[i][j] - mh[i][j];
  }
  return h;
}

// "m * s", "s * m": product of matrix m and scalar s
template <typename T>
Matrix<T> operator*(const Matrix<T>& m, T s) noexcept {
  Matrix<T> h = m;
  h *= s;
  return h;
}

template <typename T>
Matrix<T> operator*(T s, const Matrix<T>& m) noexcept {
  Matrix<T> h = m;
  h *= s;
  return h;
}

// "m * mh": product of matrices m and mh
template <typename T>
Matrix<T> operator*(const Matrix<T>& m, const Matrix<T>& mh) {
  int i, j, k;
  T sum;
  Matrix<T> h(m.rows(), mh.cols());

  if (m.cols() != mh.rows()) throw MthException(MthException::dimensions);
  for (i = 0; i < m.rows(); i++) {
    for (j = 0; j < mh.cols(); j++) {
       sum = static_cast<T> (0);
       for (k = 0; k < m.cols(); k++) sum += m[i][k] * mh[k][j];
       h[i][j] = sum;
    }
  }
  return h;
}

// "m & mh": product of matrix m and transpose(mh)
template <typename T>
Matrix<T> operator&(const Matrix<T>& m, const Matrix<T>& mh) {
  int i, j;
  Matrix<T> h(m.rows(), mh.rows());

  if (m.cols() != mh.cols()) throw MthException(MthException::dimensions);
  for (i = 0; i < m.rows(); i++) {
    for (j = 0; j < mh.rows(); j++) h[i][j] = m[i] * mh[j];
  }
  return h;
}

// "m * v": product of a matrix and a vector
template <typename T>
Vector<T> operator*(const Matrix<T>& m, const Vector<T>& v) {
  int i;
  Vector<T> h(m.rows());

  if (m.cols() != v.size()) throw MthException(MthException::dimensions);
  for (i = 0; i < m.rows(); i++) h[i] = m[i] * v;
  return h;
}

#endif
