/************************************************************************/
/* Polynomial.h: template class Polynomial				*/
/************************************************************************/

/* A polynomial of degree n-1 is defined as
   p(x) = a_0 + a_1 x + ... a_n x^n .
   As a polynomial is described by the vector of its coefficients, we
   implement it here as a child of the Vector class.

   Polynomials are meaningful for variable types for which the operations
   +, *, = are defined, e.g., int or double. The scale() method should
   be used for type double only.
*/

#ifndef POLYNOMIAL_H
#define POLYNOMIAL_H


template <typename T>
class Polynomial : public Vector<T> {
  public:
    Polynomial(int i = 0) : Vector<T>(i) {}

    T eval(T x) const noexcept;		// evaluate
    T scale() noexcept;			// scale & normalize

    Polynomial<T>& operator=(T x) noexcept {
      *static_cast<Vector<T> *>(this) = x;
      return *this;
    }

    Polynomial<T>& operator=(const Vector<T>& vh) noexcept {
      *static_cast<Vector<T> *>(this) = vh;
      return *this;
    }

    Polynomial<T>& operator+=(T x) noexcept {
      *static_cast<Vector<T> *>(this) += x;
      return *this;
    }

    Polynomial<T>& operator+=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) += vh;
      return *this;
    }

    Polynomial<T>& operator-=(T x) noexcept {
      *static_cast<Vector<T> *>(this) -= x;
      return *this;
    }

    Polynomial<T>& operator-=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) -= vh;
      return *this;
    }

    Polynomial<T>& operator*=(T x) noexcept {
      *static_cast<Vector<T> *>(this) *= x;
      return *this;
    }

    Polynomial<T>& operator*=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) -= vh;
      return *this;
    }

    Polynomial<T>& operator/=(T x) noexcept {
      *static_cast<Vector<T> *>(this) /= x;
      return *this;
    }

    Polynomial<T>& operator/=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) /= vh;
      return *this;
    }
};

// v.eval(x): evaluate polynomial
template <typename T>
T Polynomial<T>::eval(T x) const noexcept {
  T *p = Vector<T>::a + Vector<T>::n - 1,
    *pmin = Vector<T>::a,
    sum = *p--;
  while (p >= pmin) sum = sum * x + *p--;
  return sum;
}

// v.scale(x): scale and normalize polynomial
//   multiply the coefficients with powers of a scaling factor s
//   so that they are of the order O(1). The roots of the original
//   polynomial are x_i = xs_i / s
template <typename T>
T Polynomial<T>::scale() noexcept {
  int i, e,
    emax = -1024;
  double s, h;
  T *p = Vector<T>::a + Vector<T>::n - 1,
    *pmin = Vector<T>::a;
  h = 1.0 / *p;
  *p-- = 1.0;
  i = 1;
  while (p >= pmin) {
    *p *= h;
    frexp(*p--, &e);			// faster than pwr()
    e /= i++;
    emax = std::max(emax, e);
  }
  e = -emax - 1;
  s = ldexp(1.0, e);
  i = 0;
  p = Vector<T>::a + Vector<T>::n - 2;
  while (p >= pmin) {
    i += e;
    *p = ldexp(*p, i);			// *= 2^i
    p--;
  }
  return s;
}


// global functions for operator overloading (binary operations)

template <typename T>
Polynomial<T> operator+(const Polynomial<T>& v, T s) noexcept;
template <typename T>
Polynomial<T> operator+(T s, const Polynomial<T>& v) noexcept;
template <typename T>
Polynomial<T> operator+(const Polynomial<T>& v, const Vector<T>& vh)
noexcept;
template <typename T>
Polynomial<T> operator-(const Polynomial<T>& v, T s) noexcept;
template <typename T>
Polynomial<T> operator-(T s, const Polynomial<T>& v) noexcept;
template <typename T>
Polynomial<T> operator-(const Polynomial<T>& v, const Vector<T>& vh)
noexcept;
template <typename T>
Polynomial<T> operator*(const Polynomial<T>& v, T s) noexcept;
template <typename T>
Polynomial<T> operator*(T s, const Polynomial<T>& v) noexcept;

template <typename T>
Polynomial<T> operator+(const Polynomial<T>& v, T s) noexcept {
  Polynomial<T> h(v);
  h += s;
  return h;
}

template <typename T>
Polynomial<T> operator+(T s, const Polynomial<T>& v) noexcept {
  Polynomial<T> h(v);
  h += s;
  return h;
}

template <typename T>
Polynomial<T> operator+(const Polynomial<T>& v, const Vector<T>& vh)
noexcept {
  Polynomial<T> h(v);
  h += vh;
  return h;
}

template <typename T>
Polynomial<T> operator-(const Polynomial<T>& v, T s) noexcept {
  Polynomial<T> h(v);
  h -= s;
  return h;
}

template <typename T>
Polynomial<T> operator-(T s, const Polynomial<T>& v) noexcept {
  Polynomial<T> h(v);
  h -= s;
  return h;
}

template <typename T>
Polynomial<T> operator-(const Polynomial<T>& v, const Vector<T>& vh)
noexcept {
  Polynomial<T> h(v);
  h -= vh;
  return h;
}

template <typename T>
Polynomial<T> operator*(const Polynomial<T>& v, T s) noexcept {
  Polynomial<T> h(v);
  h *= s;
  return h;
}

template <typename T>
Polynomial<T> operator*(T s, const Polynomial<T>& v) noexcept {
  Polynomial<T> h(v);
  h *= s;
  return h;
}

template <typename T>
Polynomial<T> operator/(const Polynomial<T>& v, T s) noexcept {
  Polynomial<T> h(v);
  h *= 1.0 / s;
  return h;
}

template <typename T>
Polynomial<T> operator/(T s, const Polynomial<T>& v) noexcept {
  Polynomial<T> h(v);
  h *= 1.0 / s;
  return h;
}

#endif
