/************************************************************************/
/* SortedVector.h: template class SortedVector				*/
/************************************************************************/

/* A sorted vector has its elements arranged in ascending order. We
   implement it as a child of the Vector class.

   Evidently, the class should be used for variable types for which
   "ascending order" makes sense.
*/

#ifndef SORTEDVECTOR_H
#define SORTEDVECTOR_H


template <typename T>
class SortedVector : public Vector<T> {
  public:
    typedef Vector<T> super;		// abbreviation for parent class
    SortedVector(int i = 0) : Vector<T>(i) {}

    int floor_idx(T x) const noexcept;
    bool is_sorted() const noexcept;

    SortedVector<T>& operator=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) = vh;
      if (!this->is_sorted) throw MthException(MthException::ascending);
    }

    SortedVector<T>& operator+=(T x) noexcept {
      *static_cast<Vector<T> *>(this) += x;
    }

    SortedVector<T>& operator+=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) += vh;
      if (!this->is_sorted) throw MthException(MthException::ascending);
    }

    SortedVector<T>& operator-=(T x) noexcept {
      *static_cast<Vector<T> *>(this) -= x;
    }

    SortedVector<T>& operator-=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) -= vh;
      if (!this->is_sorted) throw MthException(MthException::ascending);
    }

    SortedVector<T>& operator*=(T x) {
      *static_cast<Vector<T> *>(this) *= x;
      if (x < -x) throw MthException(MthException::ascending);
    }

    SortedVector<T>& operator*=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) *= vh;
      if (!this->is_sorted) throw MthException(MthException::ascending);
    }

    SortedVector<T>& operator/=(T x) {
      *static_cast<Vector<T> *>(this) /= x;
      if (x < -x) throw MthException(MthException::ascending);
    }

    SortedVector<T>& operator/=(const Vector<T>& vh) {
      *static_cast<Vector<T> *>(this) /= vh;
      if (!this->is_sorted) throw MthException(MthException::ascending);
    }
};

// v.floor_idx: index of largest element a[i] <= x
template <typename T>
int SortedVector<T>::floor_idx(T x) const noexcept {
  int i, high, low;

  if (x < super::a[0])
    return -1;
  else if (x < super::a[super::n - 1]) {// bisection method
    low = 0;
    high = super::n - 1;
    while (high - low > 1) {
      i = (low + high) / 2;
      if (x >= super::a[i])
	low = i;
      else
	high = i;
    }
    return low;
  }
  else
    return super::n - 1;
}

// v.is_sorted: check if elemts are (still) in ascending order
template <typename T>
bool SortedVector<T>::is_sorted() const noexcept {
  T *p = super::a + 1,
  *pmax = super::a + super::n - 1;
  while (p <= pmax) {
    if (*p++ <= *(p - 1)) return false;
  }
  return true;
}

#endif
