/************************************************************************/
/* Vector.h: template class Vector					*/
/************************************************************************/

/* A Vector is an ordered set of primitive data (e.g., char, int, double).
   It is possible to construct Vectors of other data types, but then some
   of the class functions may not work properly. It is assumed that the
   "=" operation is defined for the Vector elements.

   It is furthermore assumed that the operations "+", "-", "*", and "/"
   are defined; care should be taken not to invoke them for data of
   wrong type.
*/

#ifndef VECTOR_H
#define VECTOR_H

extern double RND() noexcept;		// RND defined in mathC.h


template <typename T>
class Vector {
  protected:
    int n;				// number of elements
    T *a;				// pointer to a chunk of data

  public:
    Vector(int i = 0): n(std::max(0, i)) {	// constructor
      if (n > 0) a = new T[n];		// do not allow zero-length space
    }
    Vector(const Vector<T>& vh);	// copy constructor
    ~Vector() { 			// destructor
      if (n > 0) delete [] a;
    }

    int size() const noexcept {		// show size
      return n;
    }

    T* first() noexcept {		// pointer to 1st element
      return a;
    }
    const T* first() const noexcept {	// ... version for constant vectors
      return a;
    }

    T* last() noexcept {		// pointer to last element
      return a + n - 1;
    }
    const T* last() const noexcept {	// ... version for constant vectors
      return a + n - 1;
    }

    T& operator[](int i) noexcept {	// index operator
      return a[i];
    }
    const T& operator[](int i) const noexcept {	// ... for constant vectors
      return a[i];
    }

    int find_idx(MthKey::mth_key op) const noexcept;// index of min/max element
    T find_val(MthKey::mth_key op) const noexcept;// value of min/max
    void func(T (*f)(T));		// apply function
    T polynomial(T x) const noexcept;	// eval. polynomial with coeffs. a[i]
    int print(FILE *fp, const char *pre, const char *fmt, const char *post)
      const noexcept;			// print a vector
    int print(FILE *fp, const char *fmt) const noexcept;
    void ramp(T x, T dx) noexcept;	// fill with arithmetic sequence
    int read(FILE *fp, const char *fmt);// read a vector
    void set(T x) noexcept;		// fill with constant value;
    int size(int nnew) noexcept;	// change size
#if 0
    void redim(int nnew) noexcept;	//   obsolete version
#endif
    T sum() const noexcept;		// sum of elements
    T sum2() const noexcept;		// sum of squares
    void urand() noexcept;		// unit vector of random direction
    Vector<T>& operator=(T x) noexcept;
    Vector<T>& operator=(const Vector<T>& vh) noexcept;
    Vector<T>& operator+=(T x) noexcept;
    Vector<T>& operator+=(const Vector<T>& vh);
    Vector<T>& operator-=(T x) noexcept;
    Vector<T>& operator-=(const Vector<T>& vh);
    Vector<T>& operator*=(T x) noexcept;
    Vector<T>& operator*=(const Vector<T>& vh);
    Vector<T>& operator/=(T x) noexcept;
    Vector<T>& operator/=(const Vector<T>& vh);
};

// copy constructor
template <typename T>
Vector<T>::Vector(const Vector<T>& vh): n(vh.n) {
  if (n > 0) {
    T *p = a = new T[vh.n],
      *pmax = a + n - 1,
      *ph = vh.a;
    while (p <= pmax) *p++ = *ph++;
  }
}

// v.find_idx(i): report index of smallest/largest element
template <typename T>
int Vector<T>::find_idx(MthKey::mth_key op) const noexcept {
  int i = 1,
      iref = 0;
  T *p = a + 1,
    *pmax = a + n - 1;

  switch (op) {
    case MthKey::min:
      while (p <= pmax) {
	if (*p++ < *(a + iref)) iref = i;
	i++;
      }
      break;
    case MthKey::max:
      while (p <= pmax) {
	if (*p++ > *(a + iref)) iref = i;
	i++;
      }
      break;
  }
  return iref;
}

// v.find_val(i): report value of smallest/largest element
template <typename T>
T Vector<T>::find_val(MthKey::mth_key op) const noexcept {
  int i = 1,
      iref = 0;
  T x = *a,
    *p = a + 1,
    *pmax = a + n - 1;

  switch (op) {
    case MthKey::min:
      while (p <= pmax) {
	if (*p < x) x = *p;
	p++;
      }
      break;
    case MthKey::max:
      while (p <= pmax) {
	if (*p > x) x = *p;
	p++;
      }
      break;
  }
  return x;
}

// v.func(f): apply function f to all vector elements
template <typename T>
void Vector<T>::func(T (*f)(T)) {
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) *p++ = (*f)(*p);
}

// v.polynomial(x): calc. polynomial having the vector elements as coefficients
template <typename T>
T Vector<T>::polynomial(T x) const noexcept {
  T *p = a + n - 1,
    *pmin = a,
    sum = *p--;
  while (p >= pmin) sum = sum * x + *p--;
  return sum;
}

//v.print(fp, fmt): print vector elements to I/O channel fp using the
//		    format fmt
template <typename T>
int Vector<T>::print(FILE *fp, const char *fmt) const noexcept {
  int bytes = 0;
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) bytes += fprintf(fp, fmt, *p++);
  return bytes;
}
template <typename T>
int Vector<T>::print(FILE *fp, const char *pre, const char *fmt,
const char *post) const noexcept {
  int bytes;
  T *p = a,
    *pmax = a + n - 1;
  bytes = fprintf(fp, "%s", pre);
  while (p <= pmax) bytes += fprintf(fp, fmt, *p++);
  bytes += fprintf(fp, "%s", post);
  return bytes;
}

// v.ramp(x, dx): fill with arithmetic sequence
template <typename T>
void Vector<T>::ramp(T x, T dx) noexcept {
  T *p = a,
    *pmax = a + n - 1;

  while (p <= pmax) {
    *p++ = x;
    x += dx;
  }
}

//v.read(fp, fmt): read vector elements to I/O channel fp using the
//		   format fmt
template <typename T>
int Vector<T>::read(FILE *fp, const char *fmt) {
  int n_read = 0;
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) n_read += fscanf(fp, fmt, p++);
  return n_read;
}

// v.set(x): fill with constant value
template <typename T>
void Vector<T>::set(T x) noexcept {
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) *p++ = x;
}

// v.size(nnew): change vector size
template <typename T>
int Vector<T>::size(int nnew) noexcept {
  int i;
  T *p;

  if (nnew <= 0) {
    if (n > 0) delete [] a;
    n = 0;
  }
  else if (nnew != n) {
    p = new T[nnew];
    for (i = 0; i < std::min(n, nnew); i++) p[i] = a[i];
    if (n > 0) delete [] a;
    n = nnew;
    a = p;
  }
  return n;
}

// v.sum(): sum of all elements
template <typename T>
T Vector<T>::sum() const noexcept {
  T sum = static_cast<T> (0),
    *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) sum += *p++;
  return sum;
}

// v.sum2(): sum of squares of all elements
template <typename T>
T Vector<T>::sum2() const noexcept {
  T sum = static_cast<T> (0),
    *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) {
    sum += *p * *p;
    p++;
  }
  return sum;
}

// "v = s": set all vector elements to a fixed (scalar) value
template <typename T>
Vector<T>& Vector<T>::operator=(T x) noexcept {
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) *p++ = x;
  return *this;
}

// "v = vh": initialize v with elements of vh
template <typename T>
Vector<T>& Vector<T>::operator=(const Vector<T>& vh) noexcept {
  int i;
  T *h;

  if (n != vh.n) {			// size mismatch?
    h = (vh.n > 0) ? new T[vh.n] : NULL;	// yes, make a new vector
    for (i = 0; i < vh.n; i++) h[i] = vh.a[i];
    if (n > 0) delete [] a;
    n = vh.n;
    a = h; 
  }
  else {					// no, just copy
    for (i = 0; i < n; i++) a[i] = vh.a[i];
  }
  return *this;
}

// "v += s": add scalar s to each element of v
template <typename T>
Vector<T>& Vector<T>::operator+=(T x) noexcept {
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) *p++ += x;
  return *this;
}

// "v += vh": add element of vh to each element of v
template <typename T>
Vector<T>& Vector<T>::operator+=(const Vector<T>& vh) {
  T *p = a,
    *pmax = a + n - 1,
    *ph = vh.a;
  if (n != vh.n) throw MthException(MthException::dimensions);
  while (p <= pmax) *p++ += *ph++;
  return *this;
}

// "v -= s": substract scalar s from each element of v
template <typename T>
Vector<T>& Vector<T>::operator-=(T x) noexcept {
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) *p++ -= x;
  return *this;
}

// "v -= vh": substract element of vh from each element of v
template <typename T>
Vector<T>& Vector<T>::operator-=(const Vector<T>& vh) {
  T *p = a,
    *pmax = a + n - 1,
    *ph = vh.a;
  if (n != vh.n) throw MthException(MthException::dimensions);
  while (p <= pmax) *p++ -= *ph++;
  return *this;
}

// "v *= s": multiply each element of v with scalar s
template <typename T>
Vector<T>& Vector<T>::operator*=(T x) noexcept {
  T *p = a,
    *pmax = a + n - 1;
  while (p <= pmax) *p++ *= x;
  return *this;
}

// "v *= vh": multiply each element of v with element of vh
template <typename T>
Vector<T>& Vector<T>::operator*=(const Vector<T>& vh) {
  T *p = a,
    *pmax = a + n - 1,
    *ph = vh.a;
  if (n != vh.n) throw MthException(MthException::dimensions);
  while (p <= pmax) *p++ *= *ph++;
  return *this;
}

// "v /= s": divide each element of v by scalar s 
template <typename T>
Vector<T>& Vector<T>::operator/=(T x) noexcept {
  T *p = a,
    *pmax = a + n - 1;
  double xh = 1.0 / x;
  while (p <= pmax) *p++ *= xh;
  return *this;
}

// "v /= vh": divide each element of v by element of vh 
template <typename T>
Vector<T>& Vector<T>::operator/=(const Vector<T>& vh) {
  T *p = a,
    *pmax = a + n - 1,
    *ph = vh.a;

  if (n != vh.n) throw MthException(MthException::dimensions);
  while (p <= pmax) *p++ /= *ph++;
  return *this;
}

// "v.urand()": unit vector of random direction (Jansoone's method)
template <typename T>
void Vector<T>::urand() noexcept {
  double s, *p,
    *pmax = a + n - 1;

  do {
    s = 0.0;
    p = a;
    while (p <= pmax) {			// random vector in cube [-1; +1]
      *p = 2.0 * RND() - 1.0;
      s += sq(*p);
      p++;
    }
  } while (s <= 0.0 || s > 1.0);	// discard if outside unit sphere
  s = 1.0 / sqrt(s);
  p = a;
  while (p <= pmax) *p++ *= s;		// normalize
}
  

// global functions for operator overloading (binary operations)

template <typename T>
Vector<T> operator+(const Vector<T>& v, T s) noexcept;
template <typename T>
Vector<T> operator+(T s, const Vector<T>& v) noexcept;
template <typename T>
Vector<T> operator+(const Vector<T>& v, const Vector<T>& vh);
template <typename T>
Vector<T> operator-(const Vector<T>& v, T s) noexcept;
template <typename T>
Vector<T> operator-(T s, const Vector<T>& v) noexcept;
template <typename T>
Vector<T> operator-(const Vector<T>& v, const Vector<T>& vh);
template <typename T>
Vector<T> operator*(const Vector<T>& v, T s) noexcept;
template <typename T>
Vector<T> operator*(T s, const Vector<T>& v) noexcept;
template <typename T>
T operator*(const Vector<T>& v, const Vector<T>& vh);
template <typename T>
Vector<T> operator/(const Vector<T>& v, T s) noexcept;
template <typename T>
Vector<T> operator^(const Vector<T>& v, const Vector<T>& vh);

// "v + s": return vector whose elements are v[i] + s
template <typename T>
Vector<T> operator+(const Vector<T>& v, T s) noexcept {
  Vector<T> h(v);
  h += s;
  return h;
}

// "s + v": return vector whose elements are v[i] + s
template <typename T>
Vector<T> operator+(T s, const Vector<T>& v) noexcept {
  Vector<T> h(v);
  h += s;
  return h;
}

// "v + vh": return vector whose elements are v[i] + vh[i]
template <typename T>
Vector<T> operator+(const Vector<T>& v, const Vector<T>& vh) {
  Vector<T> h(v);
  h += vh;
  return h;
}

// "v - s": return vector whose elements are v[i] - s
template <typename T>
Vector<T> operator-(const Vector<T>& v, T s) noexcept {
  Vector<T> h(v);
  h -= s;
  return h;
}

// "s - v": return vector whose elements are -v[i] + s
template <typename T>
Vector<T> operator-(T s, const Vector<T>& v) noexcept {
  Vector<T> h(v);
  h += s;
  h -= (v + v);
  return h;
}

// "v - vh": return vector whose elements are v[i] - vh[i]
template <typename T>
Vector<T> operator-(const Vector<T>& v, const Vector<T>& vh) {
  Vector<T> h(v);
  h -= vh;
  return h;
}

// "v * s": return vector whose elements are v[i] * s
template <typename T>
Vector<T> operator*(const Vector<T>& v, T s) noexcept {
  Vector<T> h(v);
  h *= s;
  return h;
}

// "s * v": return vector whose elements are v[i] * s
template <typename T>
Vector<T> operator*(T s, const Vector<T>& v) noexcept {
  Vector<T> h(v);
  h *= s;
  return h;
}

// "v * vh": inner product
template <typename T>
T operator*(const Vector<T>& v, const Vector<T>& vh) {
  Vector<T> h(v);
  h *= vh;
  return h.sum();
}

// "v / s": return vector whose elements are v[i] / s
template <typename T>
Vector<T> operator/(const Vector<T>& v, T s) noexcept {
  Vector<T> h(v);
  h *= 1.0 / s;
  return h;
}

// "v ^ vh": cross product (for 3D vector only!)
template <typename T>
Vector<T> operator^(const Vector<T>& v, const Vector<T>& vh) {
  if (v.size() != 3 || vh.size() != 3)
    throw MthException(MthException::dimensions);
  Vector<T> h(v);
  h[0] = v[1] * vh[2] - v[2] * vh[1];
  h[1] = v[2] * vh[0] - v[0] * vh[2];
  h[2] = v[0] * vh[1] - v[1] * vh[0];
  return h;
}

#endif
