/************************************************************************/
/* mathC.h: main header file of the mathC library			*/
/************************************************************************/
#ifndef MATHC_H
#define MATHC_H

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <functional>			// has function<>
#include <utility>			// has swap()
#include <complex>

#define RND	ran_shuffle		// preferred random number generator

typedef std::complex<double> complex_d;

// numerical constants
#if 0 // elegant construct, but confuses the Mac OSX (Darwin) loader
#ifdef MAIN
extern const double
  mth_pi   = 3.14159265358979323846,
  mth_high = 1.0e+99,
  mth_eps0 = 1.0e-12,			// threshold for floating-point "=="
  mth_eps1 = 1.0e-9;			// relative precision in iterations
#else
extern const double mth_pi, mth_high, mth_eps0, mth_eps1;
#endif // of MAIN/subroutine switch
#else // of Darwin-caused workaround
const double
  mth_pi   = 3.14159265358979323846,
  mth_high = 1.0e+99,
  mth_eps0 = 1.0e-12,			// threshold for floating-point "=="
  mth_eps1 = 1.0e-09;			// relative precision in iterations
#endif // of Darwin-caused workaround

enum errcode {				// function return values:
  mth_GOOD,				// null problemo
  mth_FAIL				// generic error
};


// structs
struct point2 {
  double x, y;
};


#include <MthException.h>
#include <MthKey.h>


// template definitions
#include <simple.h>			// simple template functions
#include <functions.h>			// inline functions
#include <Vector.h>			// vector operations
#include <Matrix.h>			// matrix operations
#include <Polynomial.h>			// polynomial operations
#include <SortedVector.h>		// ops for ordered vectors
#include <Accumulator.h>		// running averages
#include <multicomplex.h>		// multicomplex algebra


// function prototypes
Matrix<double> diff_2all(std::function<double(double, double, void *)>,
  void *, double, double, int, int, double, double);

double diff_cauchy(std::function<complex_d(const complex_d&, void *)>, void *,
  int, double, double, double);
double diff_cauchy(std::function<complex_d(const complex_d&)>, int, double,
  double, double);
double diff_cauchy(std::function<complex_d(const Vector<complex_d>&, void *)>,
  void *, int, const Vector<double>&, const Vector<double>&, double, double);
double diff_cauchy(std::function<complex_d(const Vector<complex_d>&)>,
  int, const Vector<double>&, const Vector<double>&, double, double);
Vector<double>
  diff_cauchy_all(std::function<complex_d(const Vector<complex_d>&, void *)>,
  void *, int, const Vector<double>&, double, double);
Vector<double>
  diff_cauchy_all(std::function<complex_d(const Vector<complex_d>&)>,
  int, const Vector<double>&, double, double);

Vector<double> diff_gradient(std::function<double(const Vector<double>&,
  void *)>, void *, const Vector<double>&, int, const Vector<double>&);
Vector<double> diff_gradient(std::function<double(const Vector<double>&)>,
  const Vector<double>&, int, const Vector<double>&);

Matrix<double> diff_hessian(std::function<double(const Vector<double>&,
  void *)>, void *, const Vector<double>&, int, const Vector<double>&);
Matrix<double> diff_hessian(std::function<double(const Vector<double>&)>,
  const Vector<double>&, int, const Vector<double>&);

double diff_lm(std::function<double(const complex_d&, void *)>, void *, int,
  double, double, double);
double diff_lm(std::function<double(const complex_d&)>, int, double, double,
  double);
double diff_lm(std::function<double(const Vector<complex_d>&, void *)>,
  void *, int, const Vector<double>&, const Vector<double>&, double,
  double);
double diff_lm(std::function<double(const Vector<complex_d>&)>, int,
  const Vector<double>&, const Vector<double>&, double, double);
Vector<double> diff_lm_all(std::function<double(const Vector<complex_d>&,
  void *)>, void *, int, const Vector<double>&, double, double);
Vector<double> diff_lm_all(std::function<double(const Vector<complex_d>&)>,
  int, const Vector<double>&, double, double);

double diff_mcx(std::function<multicomplex(const multicomplex&, void *)>,
  void *, unsigned int, double);
double diff_mcx(std::function<multicomplex(const multicomplex&)>,
  unsigned int, double);
double diff_mcx(std::function<multicomplex(const Vector<multicomplex>&,
  void *)>, void *, unsigned int, const Vector<double>&,
  const Vector<double>&);
double diff_mcx(std::function<multicomplex(const Vector<multicomplex>&)>,
  unsigned int, const Vector<double>&, const Vector<double>&);
double diff_mcx(std::function<multicomplex(const Vector<multicomplex>&,
  void *)>, void *, const Vector<unsigned int>&, const Vector<double>&);
double diff_mcx(std::function<multicomplex(const Vector<multicomplex>&)>,
  const Vector<unsigned int>&, const Vector<double>&);

double diff_ridders(std::function<double(double, void *)>, void *, int,
  double, double, double&);
double diff_ridders(std::function<double(double)>, int, double, double,
  double&);
double diff_ridders(std::function<double(const Vector<double>&, void*)>,
  void *, int, const Vector<double>&, const Vector<double>&, double,
  double&);
double diff_ridders(std::function<double(const Vector<double>&)>,
  int, const Vector<double>&, const Vector<double>&, double, double&);

double diff_romberg(std::function<double(double, void *)>, void *, int,
  double, double, int, double, double);
double diff_romberg(std::function<double(double)>, int, double, double,
  int, double, double);

double diff_st(std::function<complex_d(const complex_d&, void *)>, void *,
  double);
double diff_st(std::function<complex_d(const complex_d&)>, double);
double diff_st(std::function<complex_d(const Vector<complex_d>&, void *)>,
  void *, const Vector<double>&, const Vector<double>&);
double diff_st(std::function<complex_d(const Vector<complex_d>&)>,
  const Vector<double>&, const Vector<double>&);
Vector<double> diff_st_all(std::function<complex_d(const Vector<complex_d>&,
    void *)>, void *, const Vector<double>&);
Vector<double> diff_st_all(std::function<complex_d(const Vector<complex_d>&)>,
  const Vector<double>&);

double diff_x(std::function<double(double, void *)>, void *, double, double);

int eqn_quadratic(const Polynomial<double>&, Vector<double>&) noexcept;

int eqn_cubic(const Polynomial<double>&, Vector<double>&) noexcept;

int eqn_polynomial(const Polynomial<double>&, int, Vector<double>&);

int eqn_quartic(const Polynomial<double>&, Vector<double>&) noexcept;

int eqn_quintic(const Polynomial<double>&, Vector<double>&) noexcept;

double eqn_regulafalsi(std::function<double(double, void *)>, void *,
  double, double, double, double, double);
double eqn_regulafalsi(std::function<double(double, void *)>, void *,
  double, int, double);

double eqn_seek(std::function<double(double, void *)>, void *, double&,
  double);

double eqn_steffensen(std::function<double(double, void *)>, void *,
  double, int, double);
complex_d eqn_steffensen(std::function<complex_d(complex_d, void *)>, void *,
  const complex_d&, int, double);
multicomplex eqn_steffensen(std::function<multicomplex(multicomplex, void *)>,
  void *, const multicomplex&, int, double);

double f_bessel_sph(double, int);

double f_gammaln(double);
double f_dfac(int);
int f_ibinom_c(int, int);

double f_legendre(double, int);
Polynomial<double> f_legendre_c(int);

double f_log2(double);

double fit_marquardt(std::function<Vector<double>(const Vector<double>&,
  void *)>, Vector<double>&, int, void *, int, int);

double fit_powell(std::function<Vector<double>(const Vector<double>&,
  void *)>, Vector<double>&, int, void *, int, int);
double min_powell(std::function<double(const Vector<double>&, void *)>,
  Vector<double>&, void *, int, int);

double fit_simplex(std::function<Vector<double>(const Vector<double>&,
  void *)>, Vector<double>&, const Vector<double>&, int, void *, int, int);

double int_gauss(std::function<double(double, void *)>, void *, int,
  double, double, int);
double int_gauss(std::function<double(double)>, int, double, double, int);

double int_ptrapez(std::function<double(double, void *)>, void *, double,
  double);
double int_ptrapez(std::function<double(double)>, double, double);

double int_romberg(std::function<double(double, void *)>, void *, double,
  double, double);
double int_romberg(std::function<double(double)>, double, double, double);

double int_simpson(std::function<double(double, void *)>, void *, double,
  double, int);
double int_simpson(std::function<double(double)>, double, double, int);
double int_simpson(const Vector<double>& y, double a, double b);

double int_spline(const SortedVector<double>&, const Vector<double>&,
  const Matrix<double>&, double, double) noexcept;

double int_trapez(std::function<double(double, void *)>, void *, double,
  double, int);
double int_trapez(std::function<double(double)>, double, double, int);
double int_trapez(const Vector<double>&, double, double);

void ipl_akima(const Vector<double>&, const Vector<double>&,
  Matrix<double>&);

double ipl_eval(double, const SortedVector<double>&,
  const Vector<double>&, const Matrix<double>&, int) noexcept;

Polynomial<double> ipl_lagrange(const Vector<double>&,
  const Vector<double>&);

double ipl_peval(double, const SortedVector<double>&,
  const Vector<double>&, const Matrix<double>&, int) noexcept;

void ipl_pspline(const SortedVector<double>&, const Vector<double>&,
  Matrix<double>&);

void ipl_reinsch(const SortedVector<double>&, const Vector<double>&,
  const Vector<double>&y, double, Vector<double>&, Matrix<double>&);

void ipl_spline(const SortedVector<double>&, const Vector<double>&,
  Matrix<double>&);
void ipl_spline(const SortedVector<double>&, const Vector<double>&,
  int, Matrix<double>&);

void ipl_track(double, Vector<double>&, int, const Vector<int>&, int&);
void ipl_track(double, double&, int, int, int&);

void lin_eigen2(const Matrix<double>&, Vector<double>&, Matrix<double>&);

Vector<double> lin_gauss(const Matrix<double>&, const Vector<double>&);

void lin_jacobi(Matrix<double>, Vector<double>&, Matrix<double>&);

void lin_matinv(Matrix<double>&);

void lin_rayleigh(const Matrix<double>&, double&, Vector<double>&);

double min_brent(std::function<double(double, void *)>, void *, double,
  double, double, double, double&);
double min_parabola(point2 a, point2 b, point2 c);

double min_descent(std::function<double(const Vector<double>&, void *)>,
  Vector<double>&, void *, int, int);

int min_scan1(std::function<double(double, void *)>, void *, double, double,
  double, double, double&, double&);

double ode_ck(std::function<double(double, double, void *)>, void *,
  double, double, double, double);
Vector<double> ode_ck(std::function<Vector<double>(double,
  const Vector<double>&, void *)>, void *, double, const Vector<double>&,
  double, double);

double ode_runge_kutta(std::function<double(double, double, void *)>,
  void *, double, double, double, double);
Vector<double> ode_runge_kutta(std::function<Vector<double>(double,
  const Vector<double>&, void *)>, void *, double, const Vector<double>&,
  double, double);

double ran_cgr48() noexcept;

double ran_ecuyer() noexcept;

double ran_mt() noexcept;

int ran_seed() noexcept;

Vector<int> ran_sequence(int) noexcept;

double ran_shuffle() noexcept;

double ran_stauffer() noexcept;

void sort_index(const Vector<double>&, Vector<int>&) noexcept;
#endif
