/************************************************************************/
/* multicomplex.h: class multicomplex (multicomplex arithmetics)	*/
/************************************************************************/
#ifndef MULTICOMPLEX_H
#define MULTICOMPLEX_H

/* A complex number can be written as z = x0 + i x1, where x0 and x1 are
   real numbers and i is the imaginary unit. Multicomplex numbers use more
   than one imaginary unit. A multicomplex number of level l is defined as
      m{l} = m0{l-1} + i{l} m1{l-1}, l > 0,
   where m0{l-1} and m1{l-1} are multicomplex numbers of a lower level and
   i{l} is the imaginary unit of level l. Level 1 corresponds to a complex
   number (i{1} = i is the usual imaginary unit), and Level 0 (not allowed
   here!) to a real number.

   Binary operations and assignments involving multicomplex numbers of
   unequal level are permitted. If necessary, the level of the result is
   increased.
*/

class multicomplex {
  private:
    unsigned int l;			// level
    multicomplex *m0, *m1;		// real & imaginary parts
    complex_d *z;
    
  public:
    multicomplex() {			// constructors
      l = 0;
      m0 = m1 = NULL;
      z = NULL;
    }
    multicomplex(double d) {
      l = 1;
      m0 = m1 = NULL;
      z = new complex_d;
      *z = d;
    }
    multicomplex(complex_d c) {
      l = 1;
      m0 = m1 = NULL;
      z = new complex_d;
      *z = c;
    }
    multicomplex(const multicomplex& a) {
      l = 0;
      init(a.l);
      copy(a);
    }
    ~multicomplex() {			// destructor
      purge();
    }

    void init(unsigned int i) {		// recursive initialization
      if (i > 15) throw MthException(MthException::range);
      if (i != l) {
        if (l > 0) this->purge();	// ... re-initialization
	l = i;
	if (l > 1) {
	  m0 = new multicomplex;
	  m0->init(l - 1);
	  m1 = new multicomplex;
	  m1->init(l - 1);
	}
	else
	  z = new complex_d;
      }
    }

    void copy(const multicomplex& a) noexcept {
      if (l > 1) {
        m0->copy(*a.m0);
        m1->copy(*a.m1);
      }
      else
        *z = *a.z;
    }
    
    void promote(unsigned int lnew) {	// increase the level
      if (lnew < l)
        throw MthException(MthException::data_loss);
      else if (lnew == l)
        return;
      else {
        if (l == 0)
	  this->init(lnew);
	else {
	  while (l < lnew) {
	    multicomplex *mh = new multicomplex;
	    mh->m0 = m0;
	    mh->m1 = m1;
	    mh->z = z;
	    mh->l = l;
	    m0 = mh;
	    m1 = new multicomplex;
	    m1->init(l);
	    l++;
	  }
	}
      }
    }

    void purge() noexcept {		// recursively free allocated memory
      if (l > 1) {
	delete m0;
	delete m1;
      }
      else
        delete z;
    }

    unsigned int level() const noexcept {
      return l;
    }
 
    multicomplex* p0() const noexcept {	// pointer to part
      return m0;
    }
    multicomplex* p1() const noexcept {
      return m1;
    }

    complex_d c_idx(unsigned int i) const {
      if (l > 1) {
	unsigned int k = 0x01;
	k <<= l - 1;
	return ((i & k) ? m1->c_idx(i & (~k)) : m0->c_idx(i));
      }
      else {
	if (i > 1) throw MthException(MthException::range);
	return *z;			// odd and even i give same result
      }
    }
    complex_d c_idx(unsigned int i, complex_d c) {
      unsigned int k = 0x01;
      if (l > 1) {
	k <<= l - 1;
	return ((i & k) ? m1->c_idx(i & (~k), c) : m0->c_idx(i, c));
      }
      else {
	if (i > 1) throw MthException(MthException::range);
	*z = c;				// ignore odd/even in i
	return c;
      }
    }

    double r_idx(unsigned int i) const {
      if (l > 1) {
	unsigned int k = 0x01;
	k <<= l - 1;
	return ((i & k) ? m1->r_idx(i & (~k)) : m0->r_idx(i));
      }
      else {
	if (i > 1) throw MthException(MthException::range);
	return (i == 1) ? z->imag() : z->real();
      }
    }
    double r_idx(unsigned int i, double d) {
      if (l > 1) {
	unsigned int k = 0x01;
	k <<= l - 1;
	return ((i & k) ? m1->r_idx(i & (~k), d) : m0->r_idx(i, d));
      }
      else {
	if (i > 1) throw MthException(MthException::range);
	if (i == 1)
	  z->imag(d);
	else
	  z->real(d);
	return d;
      }
    }

    double real() const {		// report/set "most real" component
      return this->r_idx(0);
    }
    double real(double d) {
      return this->r_idx(0, d);
    }

    double unreal() const {		// report/set "most imaginary" comp.
      unsigned int k = (0x01 << l) - 1;
      return this->r_idx(k);
    }
    double unreal(double d) {
      unsigned int k = (0x01 << l) - 1;
      return this->r_idx(k, d);
    }

    double norm() const {		// report sum of squares
      unsigned int i,
	k = 0x01;
      double y;
      if (l == 0)
        throw MthException(MthException::initialization);
      else {
	k <<= l;
	for (i = 0, y = 0.0; i < k; i++) y += sq(this->r_idx(i));
	return y;
      }
    }
      
    multicomplex& operator=(double d) noexcept {
      if (l > 1) {
	*m0 = d;
	*m1 = 0.0;
      }
      else {
        if (l == 0) init(1);
	*z = d;
      }
      return *this;
    }
    multicomplex& operator=(const complex_d& c) noexcept {
      if (l > 1) {
	*m0 = c;
	*m1 = 0.0;
      }
      else {
        if (l == 0) init(1);
	*z = c;
      }
      return *this;
    }
    multicomplex& operator=(const multicomplex& q) {
      if (q.l == 0) { }
      else if (q.l < l) {
        *m0 = q;
	*m1 = 0.0;
      }
      else {
	if (q.l > l) this->promote(q.l);
	if (l > 1) {
	  *m0 = *q.m0;
	  *m1 = *q.m1;
	}
	else
	  *z = *q.z;
      }
      return *this;
    }

    multicomplex& operator+=(double d) noexcept {
      if (l > 1)
	*m0 += d;
      else {
        if (l == 0) init(1);
	*z += d;
      }
      return *this;
    }
    multicomplex& operator+=(const multicomplex& q) {
      if (q.l == 0)
        throw MthException(MthException::initialization);
      if (q.l < l)
	*m0 += q;
      else {
	if (q.l > l) this->promote(q.l);
	if (l > 1) {
	  *m0 += *q.m0;
	  *m1 += *q.m1;
	}
	else
	  *z += *q.z;
      }
      return *this;
    }

    multicomplex& operator-=(double d) noexcept {
      if (l > 1)
	*m0 -= d;
      else {
        if (l == 0) init(1);
	*z -= d;
      }
      return *this;
    }
    multicomplex& operator-=(const multicomplex& q) {
      if (q.l == 0)
        throw MthException(MthException::initialization);
      if (q.l < l)
	*m0 -= q;
      else {
	if (q.l > l) this->promote(q.l);
	if (l > 1) {
	  *m0 -= *q.m0;
	  *m1 -= *q.m1;
	}
	else
	  *z -= *q.z;
      }
      return *this;
    }

    multicomplex& operator*=(double d) noexcept {
      if (l > 1) {
	*m0 *= d;
	*m1 *= d;
      }
      else {
        if (l == 0) init(1);
	*z *= d;
      }
      return *this;
    }
    multicomplex& operator*=(const multicomplex& q) {
      if (q.l == 0)
        throw MthException(MthException::initialization);
      if (q.l < l) {
	*m0 *= q;
	*m1 *= q;
      }
      else {
	if (q.l > l) this->promote(q.l);
	if (l > 1) {
	  multicomplex m00(*m0), m01(*m0), m10(*m1), m11(*m1);
	  m00 *= *q.m0;
	  m01 *= *q.m1;
	  m10 *= *q.m0;
	  m11 *= *q.m1;
	  m00 -= m11;			// m0 * q.m0 - m1 * q.m1
	  m01 += m10;			// m0 * q.m1 + m1 * q.m0
	  *m0 = m00;
	  *m1 = m01;
	}
	else
	  *z *= *q.z;
      }
      return *this;
    }

    multicomplex& operator/=(double d) noexcept {
      if (l > 1) {
	double dh = 1.0 / d;		// hope that d != 0
	*m0 *= dh;
	*m1 *= dh;
      }
      else {
        if (l == 0) init(1);
	*z /= d;
      }
      return *this;
    }
    multicomplex& operator/=(const multicomplex& q) {
      if (q.l == 0)
        throw MthException(MthException::initialization);
      if (q.l < l) {
	*m0 /= q;
	*m1 /= q;
      }
      else {
	if (q.l > l) this->promote(q.l);
	if (l > 1) {
	  multicomplex r2(*q.m0), qh(*q.m1), t0(*m0), t1(*m1);
	  r2 *= *q.m0;
	  qh *= *q.m1;
	  r2 += qh;			// sq(q.m0) + sq(q.m1)
	  t0 *= *q.m0;
	  qh = *m1;
	  qh *= *q.m1;
	  t0 += qh;
	  t0 /= r2;			// (m0 * q.m0 + m1 * q.m1) / r2
	  t1 *= *q.m0;
	  qh = *m0;
	  qh *= *q.m1;
	  t1 -= qh;
	  t1 /= r2;			// (m1 * q.m0 - m0 * q.m1) / r2
	  *m0 = t0;
	  *m1 = t1;
	}
	else
	  *z /= *q.z;
      }
      return *this;
    }
};

/************************************************************************/
/* unary operator overloads						*/
/************************************************************************/
multicomplex operator+(const multicomplex& m) noexcept; 
multicomplex operator-(const multicomplex& m) noexcept; 

/************************************************************************/
/* binary operator overloads						*/
/************************************************************************/
multicomplex operator+(const multicomplex& m, double d) noexcept; 
multicomplex operator+(double d, const multicomplex& m) noexcept; 
multicomplex operator+(const multicomplex& ma, const multicomplex& mb); 
multicomplex operator-(const multicomplex& m, double d) noexcept; 
multicomplex operator-(double d, const multicomplex& m) noexcept; 
multicomplex operator-(const multicomplex& ma, const multicomplex& mb); 
multicomplex operator*(const multicomplex& m, double d) noexcept; 
multicomplex operator*(double d, const multicomplex& m) noexcept; 
multicomplex operator*(const multicomplex& ma, const multicomplex& mb); 
multicomplex operator/(const multicomplex& m, double d) noexcept; 
multicomplex operator/(double d, const multicomplex& m) noexcept;
multicomplex operator/(const multicomplex& ma, const multicomplex& mb); 
bool operator==(const multicomplex& m, double d) noexcept; 
bool operator==(double d, const multicomplex& m) noexcept; 
bool operator==(const multicomplex& ma, const multicomplex& mb) noexcept;
bool operator!=(const multicomplex& m, double d) noexcept; 
bool operator!=(double d, const multicomplex& m) noexcept; 
bool operator!=(const multicomplex& ma, const multicomplex& mb) noexcept;


/************************************************************************/
/* special functions							*/
/************************************************************************/
double real(const multicomplex& m) noexcept;
multicomplex sin(const multicomplex& m) noexcept;
multicomplex cos(const multicomplex& m) noexcept;
multicomplex tan(const multicomplex& m) noexcept;
multicomplex sinh(const multicomplex& m) noexcept;
multicomplex cosh(const multicomplex& m) noexcept;
multicomplex tanh(const multicomplex& m) noexcept;
multicomplex exp(const multicomplex& m) noexcept;
multicomplex log(const multicomplex& m);
multicomplex log1p(const multicomplex& m);
multicomplex atan(const multicomplex& m);
multicomplex atanh(const multicomplex& m);
multicomplex pow(const multicomplex& m, double d);
multicomplex pow(const multicomplex& m, const multicomplex& e);
multicomplex sqrt(const multicomplex& m);
multicomplex nthrt(const multicomplex& m, int n);
#endif
