/*************************************************************31.08.2001*/
/* diff_2all: numerical differentiation of a 2-dimensional function - 	*/
/*	      all derivatives up to a specified order			*/
/*									*/
/* Derivatives are calculated by difference expressions using 0-2 steps	*/
/* of extrapolation by Ridder's method.					*/
/*======================================================================*/
/* f(x, y, s)	(i)  function to be differentiated
   s		(i)  structure passed to the function
   x, y		(i)  at this location
   n		(i)  order of differentiation (0 <= n <= 4)
   r		(i)  number of extrapolation steps (0 <= r <= 2)
   dx, dy	(i)  (minimal) differentiation increments
   RESULT	(o)  matrix of derivatives: f0x0y f0x1y f0x2y ...
   					    f1x0y f1x1x f1x2y ...
					    ...
   exceptions	     option: n or r value not available
   		     exceptions thrown by f				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int
  max_n = 4,				// max. order of differentiation
  max_r = 2;				// max. number of extrapolations
const double q = 1.6;			// ratio for expolation scheme


// function prototypes

static void diff_core(function<double(double, double, void *)>, void *,
  double, double, int, double, double, double *);


Matrix<double> diff_2all(function<double(double, double, void *)> f,
void *s, double x, double y, int n, int r, double dx, double dy) {
  int i, ih, ip, j, k;
  double q2, d[max_r + 1][((max_n + 2) * (max_n + 1)) / 2];
  Matrix<double> deriv(n + 1, n + 1);

  if (n < 0 || n > max_n || r < 0 || r > max_r)
    throw MthException(MthException::option);

  for (k = 0; k <= r; k++) {
    diff_core(f, s, x, y, n, dx, dy, &d[k][0]);
    dx *= q;
    dy *= q;
  }
  for (i = ip = 0; i <= n; i++) {
    for (ih = 0; ih <= i; ih++, ip++) {
      for (j = r, q2 = sq(q); j > 0; j--, q2 *= sq(q)) {
	for (k = 0; k < j; k++) d[k][ip] = (q2 * d[k][ip] - d[k + 1][ip])
	  / (q2 - 1.0);
      }
      deriv[i - ih][ih] = d[0][ip];
    }
  }
  return deriv; 
}

/*************************************************************30.08.2001*/
/* diff_core: numerical differentiation of a 2-dimensional function -	*/
/*	      all derivatives up to a specified order			*/
/*======================================================================*/
/* f(x, y, s)	(i)  function to be differentiated
   s		(i)  structure passed to the function
   x, y		(i)  at this location
   n		(i)  order of differentiation (0 <= n <= 4)
   dx, dy	(i)  (minimal) differentiation increments
   deriv	(o)  result array (length 1, 3, 6, or 10)		*/
/************************************************************************/
static void diff_core(function<double(double, double, void *)> f, void *s,
double x, double y, int n, double dx, double dy, double *deriv) {
  int i, j, jmax,
    c = (n + 1) / 2;			// center index of function table
  double dx2, dy2;
  Matrix<double> a(2 * c + 1, 2 * c + 1);

  *(deriv++) = a[c][c] = f(x, y, s);	// 0th order
  if (n == 0) return;

  for (i = -1; i <= 1; i++) {		// 1st order
    jmax = 1 - abs(i);
    for (j = -jmax; j <= jmax; j += 2)
      a[c + i][c + j] = f(x + i * dx, y + j * dy, s);
  }
  *(deriv++) = (a[c + 1][c] - a[c - 1][c]) / (2.0 * dx);	// 1x0y
  *(deriv++) = (a[c][c + 1] - a[c][c - 1]) / (2.0 * dy);	// 0x1y
  if (n == 1) return;

  for (i = -1; i <= 1; i += 2) {	// 2nd order
    for (j = -1; j <= 1; j += 2) a[c + i][c + j] = f(x + i * dx,
    y + j * dy, s);
  }
  dx2 = sq(dx);
  dy2 = sq(dy);
  *(deriv++) = (a[c + 1][c] - 2.0 * a[c][c] + a[c - 1][c]) / dx2;// 2x0y
  *(deriv++) = (a[c + 1][c + 1] - a[c + 1][c - 1]
  - a[c - 1][c + 1] + a[c - 1][c - 1]) / (4.0 * dx * dy);	// 1x1y
  *(deriv++) = (a[c][c + 1] - 2.0 * a[c][c] + a[c][c - 1]) / dy2;// 0x2y
  if (n == 2) return;

  for (i = -2; i <= 2; i += 2) {	// 3rd order
    jmax = 2 - abs(i);
    for (j = -jmax; j <= jmax; j += 4)
      a[c + i][c + j] = f(x + i * dx, y + j * dy, s);
  }
  *(deriv++) = (a[c + 2][c] - 2.0 * a[c + 1][c] + 2.0 * a[c - 1][c]
  - a[c - 2][c]) / (2.0 * cb(dx));				// 3x0y
  *(deriv++) = (a[c - 1][c + 1] - 2.0 * a[c][c + 1] + a[c + 1][c + 1]
  - a[c - 1][c - 1] + 2.0 * a[c][c - 1] - a[c + 1][c - 1])
  / (2.0 * dx2 * dy);						// 2x1y
  *(deriv++) = (a[c + 1][c - 1] - 2.0 * a[c + 1][c] + a[c + 1][c + 1]
  - a[c - 1][c - 1] + 2.0 * a[c - 1][c] - a[c - 1][c + 1])
  / (2.0 * dx * dy2);						// 1x2y
  *(deriv++) = (a[c][c + 2] - 2.0 * a[c][c + 1] + 2.0 * a[c][c - 1]
  - a[c][c - 2]) / (2.0 * cb(dy));				// 0x3y
  if (n == 3) return;

  for (i = -2; i <= 2; i += 1) {	// 4th order
    if (i == 0) continue;
    j = 3 - abs(i);
    a[c + i][c + j] = f(x + i * dx, y + j * dy, s);
    a[c + i][c - j] = f(x + i * dx, y - j * dy, s);
  }
  *(deriv++) = (a[c - 2][c] - 4.0 * a[c - 1][c] + 6.0 * a[c][c]
  - 4.0 * a[c + 1][c] + a[c + 2][c]) / sq(dx2);			// 4x0y
  *(deriv++) = (a[c - 2][c + 1] - 2.0 * a[c - 1][c + 1] + 2.0 * a[c + 1][c + 1]
  - a[c + 2][c + 1] - a[c - 2][c - 1] + 2.0 * a[c - 1][c - 1]
  - 2.0 * a[c + 1][c - 1] + a[c + 2][c - 1]) / (-4.0 * cb(dx) * dy);// 3x1y
  *(deriv++) = (a[c - 1][c - 1] - 2.0 * a[c][c - 1] + a[c + 1][c - 1]
  - 2.0 * a[c - 1][c] + 4.0 * a[c][c] - 2.0 * a[c + 1][c] + a[c - 1][c + 1]
  - 2.0 * a[c][c + 1] + a[c + 1][c + 1]) / (dx2 * dy2);		// 2x2y
  *(deriv++) = (a[c + 1][c - 2] - 2.0 * a[c + 1][c - 1] + 2.0 * a[c + 1][c + 1]
  - a[c + 1][c + 2] - a[c - 1][c - 2] + 2.0 * a[c - 1][c - 1]
  - 2.0 * a[c - 1][c + 1] + a[c - 1][c + 2]) / (-4.0 * dx * cb(dy));// 1x3y
  *(deriv++) = (a[c][c - 2] - 4.0 * a[c][c - 1] + 6.0 * a[c][c]
  - 4.0 * a[c][c + 1] + a[c][c + 2]) / sq(dy2);			// 0x4y
}
