/************************************************************************/
/* differentiation using Cauchy's integral theorem			*/
/************************************************************************/
#include <mathC.h>

#define Y2(I, J)	y[I * m + J]
#define Y3(I, J, K)	y[I * m2 + J * m + K]

using namespace std;


const int factorial[] = {1, 1, 2, 6};


/*************************************************************03.01.2018*/
/* diff_cauchy: single derivatives of a function of a scalar		*/
/*======================================================================*/
/* f(x, u)	(i)  complex function to be differentiated
		     x: a complex scalar
   u		(i)  pointer to auxiliary data
   n		(i)  order of differentiation (0 <= n <= 3)
   x		(i)  (real) location at which derivative is required
   r		(i)  radius of integration circle
   eps		(i)  desired precision (> 0: absolute, < 0: relative)
   RESULT	(o)  (real part of) d^n f(x)/d x^n
   exceptions	     convergence: failed to obtain desired precision
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_cauchy(function<complex_d(const complex_d&, void *)> f, void *u,
int n, double x, double r, double eps) {
  double y,
    c = factorial[n] / (2.0 * mth_pi);
  if (eps > 0.0) eps /= c;

  y = c * int_ptrapez([&](double phi, void *u) -> double {
    int i;
    complex_d zpwr(1.0, 0.0),
      z = polar(r, phi),
      a = x + z;
    for (i = 1; i <= n; i++) zpwr *= z;
    return real(f(a, u) / zpwr);
  }, NULL, 2.0 * mth_pi, eps); 
  return y;
}

double diff_cauchy(function<complex_d(const complex_d&)> f, int n, double x,
double r, double eps) {
  return diff_cauchy([&](const complex_d& z, void *u) -> complex_d {
    return f(z);
  }, NULL, n, x, r, eps);
}

/*************************************************************30.12.2017*/
/* diff_cauchy: single derivatives of a function of a vector		*/
/*======================================================================*/
/* f(x, u)	(i)  complex function to be differentiated
		     x(m): a complex vector
   u		(i)  pointer to auxiliary data
   n		(i)  order of differentiation (0 <= n <= 3)
   dir		(i)  direction of differentiation; e.g., for m = 3,
		     (1.0, 0, 0) means differentiation w.r.t. x[0], or
		     (1.0, 1.0, 0) means differentiation w.r.t. x[0] + x[1]
   x(m)		(i)  (real) location (vector) at which derivative is required
   r		(i)  radius of integration circle
   eps		(i)  desired precision (> 0: absolute, < 0: relative)
   RESULT	(o)  (real part of) (d f/d z[k])
   exceptions	     convergence: failed to obtain desired precision
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_cauchy(function<complex_d(const Vector<complex_d>&, void *)> f,
void *u, int n, const Vector<double>& dir, const Vector<double>& x,
double r, double eps) {
  int m = x.size();
  double y,
    c = factorial[n] / (2.0 * mth_pi);
  if (eps > 0.0) eps /= c;

  y = c * int_ptrapez([&](double phi, void *u) -> double {
    int i,
      m = dir.size();
    complex_d zpwr(1.0, 0.0),
      z = polar(r, phi);
    Vector<complex_d> a(m);
    for (i = 0; i < m; i++) a[i] = x[i] + dir[i] * z;
    for (i = 1; i <= n; i++) zpwr *= z;
    return real(f(a, u) / zpwr);
  }, NULL, 2.0 * mth_pi, eps); 
  return y;
}

double diff_cauchy(function<complex_d(const Vector<complex_d>&)> f,
int n, const Vector<double>& dir, const Vector<double>& x, double r,
double eps) {
  return diff_cauchy([&](const Vector<complex_d>& z, void *u) -> complex_d {
    return f(z);
  }, NULL, n, dir, x, r, eps);
}

/*************************************************************28.12.2017*/
/* diff_cauchy_all: all derivatives of a given order			*/
/*======================================================================*/
/* f(z, u)	(i)  complex function to be differentiated
		     z(m): a complex vector
   u		(i)  pointer to auxiliary data
   n		(i)  order of differentiation (0 <= n <= 3)
   x(m)		(i)  (real) location (vector) at which derivative is required
   r		(i)  radius of integration circle
   eps		(i)  desired precision (> 0: absolute, < 0: relative)
   RESULT	(o)  vector of length m^n containing n-th derivatives of f()
   exceptions	     option: illegal order of differentiation
   		     convergence: failed to obtain desired precision
		     exceptions thrown by f()				*/
/************************************************************************/
Vector<double> diff_cauchy_all(function<complex_d(const Vector<complex_d>&,
void *)> f, void *u, int n, const Vector<double>& x, double r, double eps) {
  int i, j, k, fact, ny,
    m = x.size(),
    m2 = sq(m);

  if (n < 0 or n > 3) throw MthException(MthException::option);

  Vector<double> direction(m);
  double c = factorial[n] / (2.0 * mth_pi);
  if (eps > 0.0) eps /= c;
  auto fz_real = [&](double phi, void *u) -> double {
    int i,
      m = x.size();
    complex_d zpwr(1.0, 0.0),
      z = polar(r, phi);
    Vector<complex_d> a(m);
    for (i = 0; i < m; i++) a[i] = x[i] + direction[i] * z;
    for (i = 1; i <= n; i++) zpwr *= z;
    return real(f(a, u) / zpwr);
  };

  for (i = 1, ny = 1; i <= n; i++) ny *= m;	// size of output
  Vector<double> y(ny);

  direction = 0.0;
  switch (n) {
    case 0:
      direction[0] = 1.0;
      y[0] = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
      break;
    case 1:
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
	y[i] = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	direction[i] = 0.0;
      }
      break;
    case 2:
      double q;
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
	Y2(i, i) = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	direction[i] = 0.0;
      }
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
        for (j = 0; j < i; j++) {
	  direction[j] = 1.0;
	  q = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	  direction[j] = 0.0;
	  Y2(i, j) = Y2(j, i) = 0.5 * q - 0.5 * (Y2(i, i) + Y2(j, j));
	}
	direction[i] = 0.0;
      }
      break;
    case 3:
      double qp, qm;
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
	Y3(i, i, i) = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	direction[i] = 0.0;
      }
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
        for (j = 0; j < i; j++) {
	  direction[j] = 1.0;
	  qp = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	  direction[j] = -1.0;
	  qm = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	  direction[j] = 0.0;
	  Y3(i, j, j) = Y3(j, i, j) = Y3(j, j, i)
	  = (qp + qm - 2.0 * Y3(i, i, i)) / 6.0;
	  Y3(i, i, j) = Y3(i, j, i) = Y3(j, i, i)
	  = (qp - qm - 2.0 * Y3(j, j, j)) / 6.0;
	}
	direction[i] = 0.0;
      }
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
        for (j = 0; j < i; j++) {
	  direction[j] = 1.0;
	  for (k = 0; k < j; k++) {
	    direction[k] = 1.0;
	    qp = c * int_ptrapez(fz_real, NULL, 2.0 * mth_pi, eps); 
	    direction[k] = 0.0;
	    Y3(i, j, k) = Y3(i, k, j) = Y3(j, i, k) = Y3(j, k, i)
	    = Y3(k, i, j) = Y3(k, j, i)
	    = (qp - Y3(i, i, i) - Y3(j, j, j) - Y3(k, k, k)
	    - 3.0 * (Y3(i, i, j) + Y3(i, j, j) + Y3(i, i, k) + Y3(i, k, k)
	    + Y3(j, j, k) + Y3(j, k, k))) / 6.0;
	  }
	  direction[j] = 0.0;
	}
	direction[i] = 0.0;
      }
      break;
  }
  return y;
}

Vector<double> diff_cauchy_all(function<complex_d(const Vector<complex_d>&)> f,
int n, const Vector<double>& x, double r, double eps) {
  return diff_cauchy_all([&](const Vector<complex_d>& z, void *u) -> complex_d {
    return f(z);
  }, NULL, n, x, r, eps);
}
