/*************************************************************31.10.2010*/
/* diff_gradient: gradient vector of an n-dimensional function		*/
/*									*/
/* Derivatives are calculated by difference expressions using 0-2 steps	*/
/* of extrapolation by Romberg's method.				*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  pointer to auxiliary data
   x		(i)  vector of arguments
   r		(i)  number of extrapolation steps (0 <= r <= 2)
   dx		(i)  vector of (minimal) differentiation increments
   RESULT	(o)  vector of 1st derivatives
   exceptions	     option: n or r value not available
   		     exceptions thrown by f				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int max_r = 3;			// max. number of extrapolations
const double q = 1.6;			// ratio for expolation scheme


// function prototypes
static Vector<double> diff_core(function<double(const Vector<double>&,
void *)>, void *, const Vector<double>&, const Vector<double>&);


Vector<double> diff_gradient(function<double(const Vector<double>&,
void *)> f, void *u, const Vector<double>& x, int r,
const Vector<double>& dx) {
  int i, j, k,
    n = x.size();
  double q2, qf;
  Vector<double> deriv(n);
  Matrix<double> d(max_r + 1, n);

  if (r < 0 || r > max_r)
    throw MthException(MthException::option);

  for (k = 0, qf = 1.0; k <= r; k++) {
    d[k] = diff_core(f, u, x, qf * dx);
    qf *= q;
  }
  for (i = 0; i < n; i++) {
    for (j = r, q2 = sq(q); j > 0; j--, q2 *= sq(q)) {
      for (k = 0; k < j; k++) d[k][i] = (q2 * d[k][i]
      - d[k + 1][i]) / (q2 - 1.0);
    }
    deriv[i] = d[0][i];
  }
  return deriv; 
}

Vector<double> diff_gradient(function<double(const Vector<double>&)> f,
const Vector<double>& x, int r, const Vector<double>& dx) {
  return diff_gradient([&](const Vector<double>& xh, void *u) -> double {
    return f(xh);
  }, NULL, x, r, dx);
}

/*************************************************************31.10.2010*/
/* diff_core: numerical differentiation of an n-dimensional function	*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  pointer to auxiliary data
   x		(i)  vector of arguments
   dx		(i)  vector of (minimal) differentiation increments	*/
/************************************************************************/
static Vector<double> diff_core(function<double(const Vector<double>&,
void *)> f, void *u, const Vector<double>& x, const Vector<double>& dx) {;
  int i,
    n = x.size();
  double fl, fh;
  Vector<double> xh = x;
  Vector<double> d(n);

  for (i = 0; i < n; i++) {
    xh[i] += dx[i];
    fh = f(xh, u);
    xh[i] = x[i] - dx[i];
    fl = f(xh, u);
    xh[i] = x[i];
    d[i] = 0.5 * (fh - fl) / dx[i];
  }
  return d;
}
