/*************************************************************27.05.2009*/
/* diff_hessian: Hessian matrix of an n-dimensional function		*/
/*									*/
/* Derivatives are calculated by difference expressions using 0-2 steps	*/
/* of extrapolation by Romberg's method.				*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  pointer to auxiliary data
   x		(i)  vector of arguments
   r		(i)  number of extrapolation steps (0 <= r <= 3)
   dx		(i)  vector of (minimal) differentiation increments
   RESULT	(o)  matrix of 2nd derivatives
   exceptions	     option: r value not available
   		     exceptions thrown by f				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int max_r = 3;			// max. number of extrapolations
const double q = 1.6;			// ratio for expolation scheme


// function prototypes
static Matrix<double> diff_core(function<double(const Vector<double>&,
void *)>, void *, const Vector<double>&, const Vector<double>&);


Matrix<double> diff_hessian(function<double(const Vector<double>&,
void *)> f, void *u, const Vector<double>& x, int r,
const Vector<double>& dx) {
  int i, ih, j, k,
    n = x.size();
  double q2, qf;
  Matrix<double> deriv(n, n), d[max_r + 1];

  if (r < 0 || r > max_r)
    throw MthException(MthException::option);

  for (k = 0, qf = 1.0; k <= r; k++) {
    d[k].size(n, n);
    d[k] = diff_core(f, u, x, qf * dx);
    qf *= q;
  }
  deriv.size(n, n);
  for (i = 0; i < n; i++) {
    for (ih = 0; ih <= i; ih++) {
      for (j = r, q2 = sq(q); j > 0; j--, q2 *= sq(q)) {
	for (k = 0; k < j; k++) d[k][i][ih] = (q2 * d[k][i][ih]
	- d[k + 1][i][ih]) / (q2 - 1.0);
      }
      deriv[i][ih] = d[0][i][ih];
      if (ih < i) deriv[ih][i] = deriv[i][ih];
    }
  }
  return deriv; 
}

Matrix<double> diff_hessian(function<double(const Vector<double>&)> f,
const Vector<double>& x, int r, const Vector<double>& dx) {
  return diff_hessian([&](const Vector<double>& xh, void *u) -> double {
    return f(xh);
  }, NULL, x, r, dx);
}

/*************************************************************27.05.2009*/
/* diff_core: numerical differentiation of an n-dimensional function	*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  pointer to auxiliary data
   x		(i)  vector of arguments
   dx		(i)  vector of (minimal) differentiation increments	*/
/************************************************************************/
static Matrix<double> diff_core(function<double(const Vector<double>&,
void *)> f, void *u, const Vector<double>& x, const Vector<double>& dx) {
  int i, j,
    n = x.size();
  double fll, flh, fhl, fhh, fl0, f00, fh0;
  Vector<double> xh = x;
  Matrix<double> d(n, n);

  f00 = f(x, u);

  for (i = 0; i < n; i++) {		// main diagonal
    xh[i] += dx[i];
    fh0 = f(xh, u);
    xh[i] = x[i] - dx[i];
    fl0 = f(xh, u);
    xh[i] = x[i];
    d[i][i] = (fh0 - 2.0 * f00 + fl0) / sq(dx[i]);
  }

  for (i = 0; i < n; i++) {		// off-diagonal elements
    for (j = 0; j < i; j++) {
      xh[i] += dx[i];
      xh[j] += dx[j];
      fhh = f(xh, u);
      xh[j] = x[j] - dx[j];
      fhl = f(xh, u);
      xh[i] = x[i] - dx[i];
      fll = f(xh, u);
      xh[j] = x[j] + dx[j];
      flh = f(xh, u);
      xh[i] = x[i];
      xh[j] = x[j];
      d[j][i] = (fhh - fhl - flh + fll) / (4.0 * dx[i] * dx[j]);
      d[i][j] = d[j][i];
    }
  }
  return d;
}
