/************************************************************************/
/* differentiation using the Lyness-Moler method			*/
/*									*/
/* The operand function must take complex arguments and return a real	*/
/* result.								*/
/* [1] J.N. Lyness, C.B. Moler, SIAM J. Numer Anal. 4 (1967) 202-210.	*/
/************************************************************************/
#include <mathC.h>

#define Y2(I, J)	y[I * m + J]
#define Y3(I, J, K)	y[I * m2 + J * m + K]

using namespace std;


const int nmoebius = 20,		// Moebius numbers
  moebius[] = {+1, -1, -1,  0, -1, +1, -1,  0,  0, +1,	//  1-10
	       -1,  0, -1, +1, +1,  0, -1,  0, -1,  0};	// 11-20


/*************************************************************05.01.2018*/
/* diff_lm: (single) derivatives of a function of a scalar		*/
/*======================================================================*/
/* f(x, u)	(i)  real function; x: complex scalar
   u		(i)  pointer to auxiliary data
   n		(i)  order of differentiation (n >= 1)
   x		(i)  (real) location at which derivative is required
   r		(i)  radius of integration circle
   eps		(i)  desired precision (> 0: absolute, < 0: relative)
   RESULT	(o)  (real part of) d^n f(x)/d x^n
   exceptions	     option: illegal order of differentiation
   		     convergence: failed to obtain desired precision
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_lm(function<double(const complex_d&, void *)> f, void *u, int n,
double x, double r, double eps) {
  int i, m, fact;
  double y, y_old, ymax, h, trapez, f0, c;
  complex_d z,
    a = x;

  if (n <= 0) throw MthException(MthException::option);

  for (i = 2, fact = 1; i <= n; i++) fact *= i;	// factorial
  c = fact * pow(r, -n);
  if (eps > 0.0) eps /= c;
  f0 = f(a, u);
  for (m = 1, y = ymax = 0.0, y_old = mth_high; m <= nmoebius; m++) {
    if (moebius[m - 1] == 0) continue;
    h = 1.0 / (n * m);
    for (i = 0, trapez = 0.0; i <= n * m - 1; i++) {
      z = a + std::polar(r, 2.0 * mth_pi * h * i);
      trapez += f(z, u);
    }
    trapez /= (n * m);
    y += moebius[m - 1] * (trapez - f0);
    ymax = fmax(ymax, fabs(y));		// positive size estimate
    if (fabs(y - y_old) < ((eps > 0.0) ? eps : fabs(eps * ymax))) break;
    y_old = y;
  }
  if (m > nmoebius) throw MthException(MthException::convergence);
  return y * c;
}

double diff_lm(function<double(const complex_d&)> f, int n, double x,
double r, double eps) {
  return diff_lm([&](const complex_d& z, void *u) -> double {
    return f(z);
  }, NULL, n, x, r, eps);
}

/*************************************************************05.01.2018*/
/* diff_lm: (single) derivatives of a function of a vector		*/
/*======================================================================*/
/* f(x, u)	(i)  real function; x(k): complex vector
   u		(i)  pointer to auxiliary data
   n		(i)  order of differentiation (n >= 1)
   dir(k)	(i)  direction of differentiation, i.e., differentiation
   		     is performed w.r.t. z = dir · x
   x(k)		(i)  (real) location at which derivative is required
   r		(i)  radius of integration circle
   eps		(i)  desired precision (> 0: absolute, < 0: relative)
   RESULT	(o)  (real part of) d^n f(z)/d z^n
   exceptions	     option: illegal order of differentiation
   		     convergence: failed to obtain desired precision
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_lm(function<double(const Vector<complex_d>&, void *)> f,
void *u, int n, const Vector<double>& dir, const Vector<double>& x,
double r, double eps) {
  int i, j, m, fact,
    k = x.size();
  double y, y_old, ymax, h, trapez, f0, c;
  complex_d zeta;
  Vector<complex_d> z(k);

  if (n <= 0) throw MthException(MthException::option);

  for (i = 2, fact = 1; i <= n; i++) fact *= i;	// factorial
  c = fact * pow(r, -n);
  if (eps > 0.0) eps /= c;
  for (j = 0; j < k; j++) z[j] = x[j];
  f0 = f(z, u);				// function value at x
  for (m = 1, y = ymax = 0.0, y_old = mth_high; m <= nmoebius; m++) {
    if (moebius[m - 1] == 0) continue;
    h = 1.0 / (n * m);
    for (i = 0, trapez = 0.0; i <= n * m - 1; i++) {
      zeta = std::polar(r, 2.0 * mth_pi * h * i);
      for (j = 0; j < k; j++) z[j] = x[j] + dir[j] * zeta;
      trapez += f(z, u);
    }
    trapez /= (n * m);
    y += moebius[m - 1] * (trapez - f0);
    ymax = fmax(ymax, fabs(y));		// positive size estimate
    if (fabs(y - y_old) < ((eps > 0.0) ? eps : fabs(eps * ymax))) break;
    y_old = y;
  }
  if (m > nmoebius) throw MthException(MthException::convergence);
  return y * c;
}

double diff_lm(function<double(const Vector<complex_d>&)> f, int n,
const Vector<double>& dir, const Vector<double>& x, double r, double eps) {
  return diff_lm([&](const Vector<complex_d>& z, void *u) -> double {
    return f(z);
  }, NULL, n, dir, x, r, eps);
}

/*************************************************************05.01.2018*/
/* diff_lm_all: all derivatives of a given order			*/
/*======================================================================*/
/* f(z, u)	(i)  real function; z(m): a complex vector
   u		(i)  pointer to auxiliary data
   n		(i)  order of differentiation (1 <= n <= 3)
   x(m)		(i)  (real) location at which derivative is required
   r		(i)  radius of integration circle
   eps		(i)  desired precision (> 0: absolute, < 0: relative)
   RESULT	(o)  vector of length m^n containing n-th derivatives of f()
   exceptions	     option: illegal order of differentiation
   		     convergence: failed to obtain desired precision
		     exceptions thrown by f()				*/
/************************************************************************/
Vector<double> diff_lm_all(function<double(const Vector<complex_d>&,
void *)> f, void *u, int n, const Vector<double>& x, double r, double eps) {
  int i, j, k, fact, ny,
    m = x.size(),
    m2 = sq(m);
  double q, qp, qm;
  const int factorial[] = {1, 1, 2, 6};

  if (n < 1 or n > 3) throw MthException(MthException::option);

  for (i = 1, ny = 1; i <= n; i++) ny *= m;	// size of output
  Vector<double> y(ny);

  Vector<double> direction(m);
  direction = 0.0;
  switch (n) {
    case 1:
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
	y[i] = diff_lm(f, u, n, direction, x, r, eps); 
	direction[i] = 0.0;
      }
      break;
    case 2:
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
	Y2(i, i) = diff_lm(f, u, n, direction, x, r, eps); 
	direction[i] = 0.0;
      }
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
        for (j = 0; j < i; j++) {
	  direction[j] = 1.0;
	  q = diff_lm(f, u, n, direction, x, r, eps); 
	  direction[j] = 0.0;
	  Y2(i, j) = Y2(j, i) = 0.5 * q - 0.5 * (Y2(i, i) + Y2(j, j));
	}
	direction[i] = 0.0;
      }
      break;
    case 3:
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
	Y3(i, i, i) = diff_lm(f, u, n, direction, x, r, eps); 
	direction[i] = 0.0;
      }
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
        for (j = 0; j < i; j++) {
	  direction[j] = 1.0;
	  qp = diff_lm(f, u, n, direction, x, r, eps); 
	  direction[j] = -1.0;
	  qm = diff_lm(f, u, n, direction, x, r, eps); 
	  direction[j] = 0.0;
	  Y3(i, j, j) = Y3(j, i, j) = Y3(j, j, i)
	  = (qp + qm - 2.0 * Y3(i, i, i)) / 6.0;
	  Y3(i, i, j) = Y3(i, j, i) = Y3(j, i, i)
	  = (qp - qm - 2.0 * Y3(j, j, j)) / 6.0;
	}
	direction[i] = 0.0;
      }
      for (i = 0; i < m; i++) {
	direction[i] = 1.0;
        for (j = 0; j < i; j++) {
	  direction[j] = 1.0;
	  for (k = 0; k < j; k++) {
	    direction[k] = 1.0;
	    qp = diff_lm(f, u, n, direction, x, r, eps); 
	    direction[k] = 0.0;
	    Y3(i, j, k) = Y3(i, k, j) = Y3(j, i, k) = Y3(j, k, i)
	    = Y3(k, i, j) = Y3(k, j, i)
	    = (qp - Y3(i, i, i) - Y3(j, j, j) - Y3(k, k, k)
	    - 3.0 * (Y3(i, i, j) + Y3(i, j, j) + Y3(i, i, k) +Y3(i, k, k)
	    + Y3(j, j, k) + Y3(j, k, k))) / 6.0;
	  }
	  direction[j] = 0.0;
	}
	direction[i] = 0.0;
      }
      break;
  }
  return y;
}

Vector<double> diff_lm_all(function<double(const Vector<complex_d>&)> f,
int n, const Vector<double>& x, double r, double eps) {
  return diff_lm_all([&](const Vector<complex_d>& z, void *u) -> double {
    return f(z);
  }, NULL, n, x, r, eps);
}
