/************************************************************************/
/* differentiation using the multicomplex step method [1, 2]		*/
/* (a generalization of the Squire-Trapp method [3, 4])			*/
/*									*/
/* The operand function must take multicomplex arguments and return a	*/
/* multicomplex result, but return a real value for a real argument.	*/
/* [1] G. Lantoine, R. P. Russell, Th. Dargent, ACM Trans. Math. Softw.	*/
/*     38 (2012) no. 16.						*/
/* [2] A. Verheyleweghen, report, NTNU, 2014.				*/
/* [3] W. Squire, G. Trapp, SIAM Rev. 40 (1998) 110-112.		*/
/* [4] N. Higham, SIAM News 51(5) (2018) 2.				*/
/*									*/
/* Note: This is experimental software; even the mathematical theory is	*/
/* not complete. Caution is advised when it is applied to functions	*/
/* containing log() or pow().						*/
/* Note: The method can handle arbitrary orders of differentiation, but	*/
/* for n > 16 the internal step size exceeds 1.0e-12, which may be too	*/
/* large for comfort.							*/
/************************************************************************/
#include <mathC.h>

using namespace std;


inline double increment(unsigned int l) {
  return ldexp(1.0, -664 / static_cast<int>(l));
}


/*************************************************************11.04.2019*/
/* diff_mcx: arbitrary-order derivatives of a function of a scalar	*/
/*======================================================================*/
/* f(x, s)	(i)  multicomplex function; x: multicomplex scalar
   u		(i)  pointer to auxiliary data
   l		(i)  order of differentiation
   x		(i)  (real) location at which derivative is required
   RESULT	(o)  (real part of) d^L f(x)/d x^L
   exceptions	     option: differentiation order < 1
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_mcx(function<multicomplex(const multicomplex&, void *)> f,
void *u, unsigned int l, double x) {
  if (l < 1) throw MthException(MthException::option);
  unsigned int i, j;
  double hl, h;
  multicomplex xh, yh;
  xh.init(l);
  yh.init(l);
  h = increment(l);
  xh = 0.0;
  xh.r_idx(0, x);
  for (i = 1, j = 1, hl = 1.0; i <= l; i++) {
    xh.r_idx(j, h);
    hl *= h;				// h^l
    j *= 2;
  }
  yh = f(xh, u);
  return yh.r_idx(j - 1) / hl;
}

double diff_mcx(function<multicomplex(const multicomplex&)> f,
unsigned int l, double x) {
  return diff_mcx([&](const multicomplex& z, void *u) -> multicomplex {
    return f(z);
  }, NULL, l, x);
}

/*************************************************************14.04.2019*/
/* diff_mcx: arbitrary-order directional derivatives of a function of a	*/
/*	     vector							*/
/*======================================================================*/
/* f(x, u)	(i)  multicomplex function; x(n): multicomplex vector
   u		(i)  pointer to auxiliary data
   l		(i)  order of differentiation
   dir(n)	(i)  direction of differentiation, i.e., differentiation
   		     is performed w.r.t. z = dir · x
   x(n)		(i)  (real) location at which the derivative is required
   RESULT	(o)  (real part of) d^l f(z)/d z^l
   exceptions	     option: differentiation order < 1
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_mcx(function<multicomplex(const Vector<multicomplex>&,
void *)> f, void *u, unsigned int l, const Vector<double>& dir,
const Vector<double>& x) {
  if (l < 1) throw MthException(MthException::option);
  int i, j, k,
    n = x.size();
  double hl, h;
  multicomplex yh;
  Vector<multicomplex> xh(n);
  yh.init(l);
  h = increment(l);
  for (k = 0; k < n; k++) {
    xh[k].init(l);
    xh[k] = 0.0;
    (xh[k]).r_idx(0, x[k]);
    for (i = 1, j = 1, hl = 1.0 ; i <= l; i++) {
      (xh[k]).r_idx(j, h * dir[k]);
      hl *= h;
      j *= 2;
    }
  }
  yh = f(xh, u);
  return yh.r_idx(j - 1) / hl;
}

double diff_mcx(function<multicomplex(const Vector<multicomplex >&)> f,
unsigned int l, const Vector<double>& dir, const Vector<double>& x) {
  return diff_mcx([&](const Vector<multicomplex>& z, void *u)
  -> multicomplex {
    return f(z);
  }, NULL, l, dir, x);
}

/*************************************************************23.04.2019*/
/* diff_mcx: arbitrary-order derivatives of a function of a vector	*/
/*	     including mixed derivatives				*/
/*======================================================================*/
/* f(x, u)	(i)  multicomplex function; x(n): multicomplex vector
   u		(i)  pointer to auxiliary data
   ord(n)	(i)  order of differentiation (w.r.t. x[i]: ord[i])
   x(n)		(i)  (real) location at which the derivative is required
   RESULT	(o)  d^l f(z)/(d x[0]^ord[0] d x[1]^ord[1] …), l = Sum ord[]
   exceptions	     option: differentiation order < 1
		     exceptions thrown by f()				*/
/************************************************************************/
double diff_mcx(function<multicomplex(const Vector<multicomplex>&,
void *)> f, void *u, const Vector<unsigned int>& ord,
const Vector<double>& x) {
  unsigned int m,
    l = ord.sum();			// total order of differentiation
  if (l < 1) throw MthException(MthException::option);
  int i, k,
    n = x.size();
  double hl, h;
  multicomplex yh;
  Vector<multicomplex> xh(n);
  yh.init(l);
  h = increment(l);
  for (k = 0, m = 0x01, hl = 1.0; k < n; k++) {
    xh[k].init(l);
    xh[k] = 0.0;
    (xh[k]).r_idx(0, x[k]);
    for (i = 1; i <= ord[k]; i++) {
      (xh[k]).r_idx(m, h);
      m <<= 1;
      hl *= h;
    }
  }
  yh = f(xh, u);
  return yh.r_idx(m - 1) / hl;
}

double diff_mcx(function<multicomplex(const Vector<multicomplex>&)> f,
const Vector<unsigned int>& ord, const Vector<double>& x) {
  return diff_mcx([&](const Vector<multicomplex>& z, void *u)
  -> multicomplex {
    return f(z);
  }, NULL, ord, x);
}
