/*************************************************************13.08.1995*/
/* diff_ridders: numerical differentiation of a 1-dimensional function	*/
/*		 using the algorithm of Ridders				*/
/* Ref.: W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,	*/
/* Numerical Recipes, 2nd Edition, p. 180ff, Cambridge University	*/
/* Press, 1992.								*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  pointer to auxiliary data
   n	 	(i)  order of differentiation (0 <= n <= ndmx)
   x		(i)  location at which derivative is required
   dx		(i)  initial differentiation increment (a range over which
		     f changes SIGNIFICANTLY; note that the increment is
		     decreased internally)
   dev		(o)  estimate of the error of the derivative
   RESULT	(o)  nth derivative of f at location x
   EXCEPTIONS        option:    n too large (extend fdiff1()!)
		     exceptions thrown by f()				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  contr = 1.4,				// interval contraction factor */
  contr2 = contr * contr,
  safe = 2.0;				// termination safety factor
const int
  ntab = 10,				// size of interpolation table
  ndmx = 5;				// max. differentiation order


static double fdiff1(function<double(double, void *)>, void *, int,
  double, double);


double diff_ridders(function<double(double, void*)> f, void *u, int n,
double x, double dx, double& dev) {
  int i, j;
  double devt, q, e1, e2, a[ntab][ntab], fd, h;

  h = fabs(dx);
  if (n < 0 || n > ndmx) throw MthException(MthException::option);

  if (n == 0) {				// trivial: return function value
    fd = f(x, u);
    dev = 0.0;
    return fd;
  }

  a[0][0] = fdiff1(f, u, n, x, h);	// calculate initial estimate
  dev = mth_high;

  for (i = 1; i < ntab; i++) {
    h /= contr;				// reduce increment size
    a[i][0] = fdiff1(f, u, n, x, h);	// calculate new derivative
    for (j = 1, q = contr2; j <= i; j++) {// extrapolate
      a[i][j] = (a[i][j - 1] * q - a[i - 1][j - 1]) / (q - 1.0);
      q *= contr2;
      e1 = fabs(a[i][j] - a[i][j - 1]);
      e2 = fabs(a[i][j] - a[i - 1][j - 1]);
      devt = (e1 > e2) ? e1 : e2;
      if (devt <= dev) {		// error decreased?
	dev = devt;			// yes: save last result
	fd = a[i][j];
      }
    }
    if (fabs(a[i][i] - a[i - 1][i - 1]) >= safe * dev) break;
  }
  return fd;
}

double diff_ridders(function<double(double)> f, int n, double x, double dx,
double& dev) {
  return diff_ridders([&](double xh, void *u) -> double {
    return f(xh);
  }, NULL, n, x, dx, dev);
}

/*************************************************************13.08.1995*/
/* fdiff1: auxiliary function; recipes for numerical differentiation	*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  structure for passing additional information
   n		(i)  order of derivative (1 <= n <= ndmx)
   x		(i)  argument of f
   h		(i)  increment of differentiation
   RESULT	(o)  raw estimate of nth derivative at location x	*/
/************************************************************************/
static double fdiff1(function<double(double, void *)> f, void *u, int n,
double x, double h) {
  switch (n) {
  case 1:
    return (f(x + h, u) - f(x - h, u)) / (2.0 * h);
    break;
  case 2:
    return (f(x + 2.0 * h, u) - 2.0 * f(x, u) + f(x - 2.0 * h, u))
    / sq(2.0 * h);
    break;
  case 3:
    return ((f(x + 3.0 * h, u) - f(x - 3.0 * h, u))
    - 3.0 * (f(x + h, u) - f(x - h, u))) / cb(2.0 * h);
    break;
  case 4:
    return ((f(x + 4.0 * h, u) + f(x - 4.0 * h, u))
    - 4.0 * (f(x + 2.0 * h, u) + f(x - 2.0 * h, u))
    + 6.0 * f(x, u)) / sq(sq(2.0 * h));
    break;
  case 5:
    return ((f(x + 4.0 * h, u) - f(x - 4.0 * h, u))
    - 10.0 * (f(x + 2.0 * h, u) - f(x - 2.0 * h, u))
    + 16.0 * (f(x + h, u) - f(x - h, u))) / (12.0 * h * sq(sq(h)));
  default:
    return 0.0;
  }
}

/*************************************************************16.02.2018*/
/* diff_ridders: (single) derivatives of a function of a vector		*/
/*======================================================================*/
/* f(x, u)	(i)  real function; x(k): real vector
   dir(k)	(i)  direction of differentiation, i.e., differentiation
   		     is performed w.r.t. z = dir · x 
   x(k)		(i)  location at which derivative is required		*/
/************************************************************************/
double diff_ridders(function<double(const Vector<double>&, void*)> f,
void *u, int n, const Vector<double>& dir, const Vector<double>& x,
double dx, double& dev) {
  int i, j,
    k = x.size();
  double devt, q, e1, e2, a[ntab][ntab], fd, h;
  Vector<double> xh(k);

  h = fabs(dx);
  if (n < 0 || n > ndmx) throw MthException(MthException::option);

  if (n == 0) {				// trivial: return function value
    fd = f(x, u);
    dev = 0.0;
    return fd;
  }

  auto f1 = [&](double z, void *uh) -> double {
    xh = x + dir * z;
    return f(xh, u);
  };
  a[0][0] = fdiff1(f1, u, n, 0.0, h);	// calculate initial estimate
  dev = mth_high;

  for (i = 1; i < ntab; i++) {
    h /= contr;				// reduce increment size
    a[i][0] = fdiff1(f1, u, n, 0.0, h);	// calculate new derivative
    for (j = 1, q = contr2; j <= i; j++) {// extrapolate
      a[i][j] = (a[i][j - 1] * q - a[i - 1][j - 1]) / (q - 1.0);
      q *= contr2;
      e1 = fabs(a[i][j] - a[i][j - 1]);
      e2 = fabs(a[i][j] - a[i - 1][j - 1]);
      devt = (e1 > e2) ? e1 : e2;
      if (devt <= dev) {		// error decreased?
	dev = devt;			// yes: save last result
	fd = a[i][j];
      }
    }
    if (fabs(a[i][i] - a[i - 1][i - 1]) >= safe * dev) break;
  }
  return fd;
}

double diff_ridders(function<double(const Vector<double>&)> f, int n,
const Vector<double>& dir, const Vector<double>& x, double dx,
double& dev) {
  return diff_ridders([&](const Vector<double>& xh, void *u) -> double {
    return f(xh);
  }, NULL, n, dir, x, dx, dev);
}
