/*************************************************************24.01.2003*/
/* diff_romberg: numerical differentiation of a 1-dimensional function	*/
/*		 for functions with a limited x range			*/
/*									*/
/* symmetric case:  extrapolation scheme of Romberg			*/
/*		    G. Jordan-Engeln, F. Reutter: Numerische Mathematik	*/
/*		    für Ingenieure, BI Hochschultaschenbücher Bd. 104,	*/
/*		    Bibliographisches Institut, Mannheim 1973, pp. 242ff*/
/* asymmetric case: differentiation of Lagrange interpolation polynomial*/
/*		    ibid., pp. 198ff					*/
/*======================================================================*/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  structure for passing additional information
   n	 	(i)  order of derivative (0 <= n <= ndmx)
   x		(i)  argument of f
   h		(i)  initial step size of differentiation
		     (subsequent steps use larger increments)
   maxord	(i)  max. order of extrapolation
   xmin, xmax	(i)  boundaries of x
   RESULT	(o)  nth derivative of f at location x
   exceptions	     option: n out of range (edit fdiff1!)
		     increment: initial step size too small
		     interval:  x range too narrow or xmax < xmin
		     exceptions thrown by f() or ipl_lagrange()		*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  contr = 1.4142135,			// interval contraction factor
  thr = 1.0e-10,			// lower increment threshold
  xdiff = 1.0e-3;
const int
  ndmx = 4;				// max. differentiation order


static double fdiff1(function<double(double, void *)>, void *, int,
  double, double);


double diff_romberg(function<double(double, void *)> f, void *u, int n,
double x, double dx, int maxord, double xmin, double xmax) {
  int i, j, ord;
  double q, aextr, fd, h;
  enum diffmode {free, bounded} mode;

  h = fabs(dx);
  if (n < 0 || n > ndmx) throw MthException(MthException::option);
  if (h < thr * fabs(x)) throw MthException(MthException::increment);

  if (n == 0) {				// trivial: return function value
    fd = f(x, u);
    return fd;
  }

  q = pow(contr, maxord) * h;		// x range for symmetric formula
  mode = (x > xmin && x < xmax) ? free : bounded;
  if (x - q < xmin) {			// asymmetric (left bounded)
    ord = 2 * (maxord + 1) + 1;
    { Vector<double> a(ord), b(ord), c(ord);
      a[0] = 0.0;
      a[1] = aextr = h;
      for (i = 2; i < ord; i++) {
	if (mode == free) {
	  a[i] = (a[i - 1] > 0.0) ? -a[i - 1] : -contr * a[i - 1];
	  if (a[i] > aextr) aextr = a[i];
	  if (x + a[i] < xmin + xdiff * h) {
	    a[i] = xmin - x;
	    mode = bounded;
	  }
	}
	else 
	  a[i] = contr * ((a[i - 1] < 0.0) ? aextr : a[i - 1]);
      }
      if (x + a[ord - 1] > xmax)
	throw MthException(MthException::interval);
      for (i = 0; i < ord; i++) b[i] = f(x + a[i], u);
      c = ipl_lagrange(a, b);
      for (i = n, fd = c[n]; i > 1; i--) fd *= i;
    }
  }
  else if (x + q <= xmax) {		// symmetric (Romberg) method
    ord = maxord  + 1;
    { Vector<double> a(ord), b(ord);
      a[0] = fdiff1(f, u, n, x, h);	//calculate initial estimate
      for (i = 1; i < ord; i++) {
	h *= contr;
	b[0] = fdiff1(f, u, n, x, h);
	for (j = 1, q = 1.0 / sq(contr); j <= i; j++) {
	  b[j] = (a[j - 1] - q * b[j - 1]) / (1.0 - q);
	  q /= sq(contr);
	}
	for (j = 0; j <= i; j++) a[j] = b[j];
      }
      fd = b[ord - 1];
    }
  }
  else {				//asymmetric (right bounded)
    ord = 2 * (maxord + 1) + 1;
    { Vector<double> a(ord), b(ord), c(ord);
      a[0] = 0.0;
      a[1] = aextr = -h;
      for (i = 2; i < ord; i++) {
	if (mode == free) {
	  a[i] = (a[i - 1] < 0.0) ? -a[i - 1] : -contr * a[i - 1];
	  if (a[i] < aextr) aextr = a[i];
	  if (x + a[i] > xmax - xdiff * h) {
	    a[i] = xmax - x;
	    mode = bounded;
	  }
	}
	else 
	  a[i] = contr * ((a[i - 1] > 0.0) ? aextr : a[i - 1]);
      }
      if (x + a[ord - 1] < xmin)
	throw MthException(MthException::interval);
      for (i = 0; i < ord; i++) b[i] = f(x + a[i], u);
      c = ipl_lagrange(a, b);
      for (i = n, fd = c[n]; i > 1; i--) fd *= i;
    }
  }
  return fd;
}

double diff_romberg(function<double(double)> f, int n, double x,
double dx, int maxord, double xmin, double xmax) {
  return diff_romberg([&](double xh, void *u) -> double {
    return f(xh);
  }, NULL, n, x, dx, maxord, xmin, xmax);
}

/*************************************************************13.08.1995*/
/* fdiff1: auxiliary function; recipes for numerical differentiation	*/
/************************************************************************/
/* f(x, u)	(i)  function to be differentiated
   u		(i)  structure for passing additional information
   n		(i)  order of derivative (1 <= n <= ndmx)
   x		(i)  argument of f
   h		(i)  increment of differentiation
   RESULT	(o)  raw estimate of nth derivative at location x	*/
/*======================================================================*/
static double fdiff1(function<double(double, void *)> f, void *u, int n,
double x, double h) {
  switch (n) {
  case 1:
    return ((f(x + h, u) - f(x - h, u)) / (2.0 * h));
    break;
  case 2:
    return ((f(x + 2.0 * h, u) - 2.0 * f(x, u) + f(x - 2.0 * h, u))
    / sq(2.0 * h));
    break;
  case 3:
    return ((f(x + 3.0 * h, u) - 3.0 * f(x + h, u)
    + 3.0 * f(x - h, u) - f(x - 3.0 * h, u)) / cb(2.0 * h));
    break;
  case 4:
    return ((f(x + 4.0 * h, u) - 4.0 * f(x + 2.0 * h, u)
    + 6.0 * f(x, u) - 4.0 * f(x - 2.0 * h, u) + f(x - 4.0 * h, u))
    / sq(sq(2.0 * h)));
    break;
  default:
    return 0.0;
  }
}
