/************************************************************************/
/* differentiation using the Squire-Trapp method			*/
/*									*/
/* The operand function must take complex arguments and return a	*/
/* complex result, but return a real value for a real argument.		*/
/* [1] W. Squire, G. Trapp, SIAM Rev. 40 (1998) 110-112.		*/
/* [2] N. Higham, SIAM News 51(5) (2018) 2.				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double dx = 1.0e-100;		// differentiation increment


/*************************************************************04.07.2018*/
/* diff_st: (single) derivative of a function of a scalar		*/
/*======================================================================*/
/* f(x, s)	(i)  complex function; x: complex scalar
   u		(i)  pointer to auxiliary data
   x		(i)  (real) location at which derivative is required
   RESULT	(o)  (real part of) d f(x)/d x
   exceptions	     exceptions thrown by f()				*/
/************************************************************************/
double diff_st(function<complex_d(const complex_d&, void *)> f, void *u,
double x) {
  complex_d xh(x, dx),
    yh = f(xh, u);
  return imag(yh) / dx;
}

double diff_st(function<complex_d(const complex_d&)> f, double x) {
  return diff_st([&](const complex_d& z, void *u) -> complex_d {
    return f(z);
  }, NULL, x);
}

/*************************************************************05.07.2018*/
/* diff_st: (single) derivatives of a function of a vector		*/
/*======================================================================*/
/* f(x, u)	(i)  complex function; x(k): complex vector
   u		(i)  pointer to auxiliary data
   dir(k)	(i)  direction of differentiation, i.e., differentiation
   		     is performed w.r.t. z = dir · x
   x(k)		(i)  (real) location at which the derivative is required
   RESULT	(o)  (real part of) d f(z)/d z
   exceptions	     exceptions thrown by f()				*/
/************************************************************************/
double diff_st(function<complex_d(const Vector<complex_d>&, void *)> f,
void *u, const Vector<double>& dir, const Vector<double>& x) {
  int i,
    k = x.size();
  Vector<complex_d> z(k);
  for (i = 0; i < k; i++) z[i] = complex_d(x[i], dx * dir[i]);
  return imag(f(z, u)) / dx;
}

double diff_st(function<complex_d(const Vector<complex_d>&)> f,
const Vector<double>& dir, const Vector<double>& x) {
  return diff_st([&](const Vector<complex_d>& z, void *u) -> complex_d {
    return f(z);
  }, NULL, dir, x);
}

/*************************************************************05.07.2018*/
/* diff_st_all: all 1st-order derivatives				*/ 
/*======================================================================*/
/* f(z, u)	(i)  complex function; z(m): a complex vector
   u		(i)  pointer to auxiliary data
   x(m)		(i)  (real) location at which the derivative is required
   RESULT	(o)  vector containing 1st-order derivatives of f()
   exceptions	     exceptions thrown by f()				*/
/************************************************************************/
Vector<double> diff_st_all(function<complex_d(const Vector<complex_d>&,
void *)> f, void *u, const Vector<double>& x) {
  int i,
    m = x.size();
  Vector<double> y(m), direction(m);
  direction = 0.0;
  for (i = 0; i < m; i++) {
    direction[i] = 1.0;
    y[i] = diff_st(f, u, direction, x); 
    direction[i] = 0.0;
  }
  return y;
}

Vector<double> diff_st_all(function<complex_d(const Vector<complex_d>&)> f,
const Vector<double>& x) {
  return diff_st_all([&](const Vector<complex_d>& z, void *u) -> complex_d {
    return f(z);
  }, NULL, x);
}
