/*************************************************************27.06.2005*/
/* eqn_cubic: real roots of a cubic equation				*/
/*									*/
/* default version: iterative method of Deiters and Macias-Salinas	*/
/* (iterative search for 1st root, then deflation and direct solution	*/
/* of quadratic equation)						*/
/*									*/
/* alternatives:							*/
/* Cardano's formula (activate with -DEQN_CUBIC=1)			*/
/* analytic method of Zhao, Wang & Hong (-DEQN_CUBIC=2)			*/
/*									*/
/* [1] U. K. Deiters, R. Macias-Salinas, Ind. Eng. Chem. Res. 53 (2014) */
/*     2529-2536.							*/
/* [2] T. Zhao, D. Wang, H. Hong, J. Symb. Comput. 46 (2011) 904-918.	*/
/*									*/
/* Equations of lesser degree are solved by the appropriate formulas.	*/
/* The solutions are arranged in ascending order.			*/
/*======================================================================*/
/* a		(i)  vector containing the polynomial coefficients
   x		(o)  result vector
   RESULT	(o)  number of valid solutions				*/
/************************************************************************/
#include <mathC.h>

#ifndef METHOD				// set default method
  #define METHOD 1			// Deiters & Macias-Salinas
#endif

using namespace std;


const double rec3 = 1.0 / 3.0;		// 1/3


#if METHOD == 0	//  Cardano's method

int eqn_cubic(const Polynomial<double>& a, Vector<double>& x) noexcept {
  int i, nreal;
  double w, h, p, q, d, phi, s, c;
  const double w3 = sqrt(3.0);

  if (a.size() < 4 || a[3] == 0.0) return eqn_quadratic(a, x);

  w = a[2] / (3.0 * a[3]);		// x = y - w, y^3 + b1 y + b0 = 0
  p = cb(a[1] / (3.0 * a[3]) - sq(w));			// (b1/3)^3
  q = -0.5 * (2.0 * cb(w) - (a[1] * w - a[0]) / a[3]);	// -b0/2
  d = sq(q) + p;			// discriminant
  if (d < 0.0) {			// 3 real solutions
    h = q / sqrt(-p);
    phi = acos(fmax(-1.0, fmin(1.0, h)));
    sincos(rec3 * phi, &s, &c);
    p = 2.0 * pow(-p, 1.0 / 6.0);
    x[0] = 0.5 * p * (-c - w3 * s) - w;
    x[1] = 0.5 * p * (-c + w3 * s) - w;
    x[2] = 0.5 * p * 2.0 * c - w;
    nreal = 3;
  }
  else {				// only one real solution
    d = sqrt(d);
    x[0] = cbrt(q + d) + cbrt(q - d) - w;
    nreal = 1;
  }

  for (i = 0; i < nreal; i++) {		// 1 Newton iteration step
    if ((h = a[1] + x[i] * (2.0 * a[2] + 3.0 * x[i] * a[3])) != 0.0)
    x[i] -= (a[0] + x[i] * (a[1] + x[i] * (a[2] + x[i] * a[3]))) / h;
  }

#if 1	// should be superfluous, but sometimes good for closely spaced roots
  if (nreal == 3) {			// sort results
    if (x[1] < x[0]) std::swap(x[0], x[1]);
    if (x[2] < x[1]) std::swap(x[1], x[2]);
    if (x[1] < x[0]) std::swap(x[0], x[1]);
  }
#endif
  return nreal;
}

#elif METHOD == 2	// method of Zhao, Wang & Hong

int eqn_cubic(const Polynomial<double>& a, Vector<double>& x) noexcept {
  int i, nreal;
  double w, h, p, q, p1, p2, s, c, phi;
  const double w3 = sqrt(3.0);

  if (a.size() < 4 || a[3] == 0.0) return eqn_quadratic(a, x);

  w = a[2] / (3.0 * a[3]);		// offset: x = y - w
  p = a[1] / a[3] - 3.0 * sq(w);
  q = (2.0 * cb(w) - (a[1] * w - a[0]) / a[3]);
  p1 = -4.0 / 27.0 * cb(p) - sq(q);	// "p1" / 27
  p2 = -q;				// "p2" from Zhao article
  if (p1 <= 0.0) {
    s = sqrt(-p1);
    x[0] = (cbrt(0.5 * (p2 + s)) + cbrt(0.5 * (p2 - s))) - w;
    nreal = 1;
  }
  else {
    s = sqrt(p1);
    phi = atan2(s, p2);
    h = pow(sq(0.5 * s)+ sq(0.5 * p2), 1.0 / 6.0); // r^(1/3)
    if (phi < 0.5 * mth_pi)		// Zhao et al. "real convention"
      phi = (phi - 2.0 * mth_pi) * rec3;
    else if (phi == -0.5 * mth_pi)
      phi = 0.5 * mth_pi;
    else if (phi < 0.5 * mth_pi)
      phi *= rec3;
    else if (phi == 0.5 * mth_pi)
      phi = -0.5 * mth_pi;
    else
      phi = (phi + 2.0 * mth_pi) * rec3;
    sincos(phi, &s, &c);
    x[0] = -w - h * c;
    x[1] = -w - h * (c + w3 * s);
    x[2] = -w - h * (c - w3 * s);
    nreal = 3;
  }

  for (i = 0; i < nreal; i++) {		// 1 Newton iteration step
    if ((h = a[1] + x[i] * (2.0 * a[2] + 3.0 * x[i] * a[3])) != 0.0)
    x[i] -= (a[0] + x[i] * (a[1] + x[i] * (a[2] + x[i] * a[3]))) / h;
  }

  if (nreal == 3) {			// sort results
    if (x[1] < x[0]) std::swap(x[0], x[1]);
    if (x[2] < x[1]) std::swap(x[1], x[2]);
    if (x[1] < x[0]) std::swap(x[0], x[1]);
  }
  return nreal;
}

#else	// method of Deiters & Macias-Salinas

inline double aux_newton1(const double *a, double x) {
  double y, y1;
  y = a[2] + x;	
  y1 = 2.0 * y + x;
  y1 = x * y1 + a[1];
  y = (x * y + a[1]) * x + a[0];
#if 0	// Newton
  if (y1 != 0.0) x -= y / y1;
#else	// Halley
  if (y != 0.0) x -= y * y1 / (sq(y1) - y * (a[2] + 3.0 * x));
#endif
  return x;
}

inline int aux_quadratic(const double *a, double *x, int i, int j) {
  double p = -0.5 * a[1],
    d = sq(p) - a[0];
  if (d >= 0.0) {
    d = sqrt(d);
    x[i] = p - d;
    x[j] = p + d;
    return 2;
  }
  else
    return 0;
}

inline int aux_quadratic(const double *a, const double *o, double *x,
int i, int j) {
  double p = -0.5 * a[1],
    d = sq(p) - a[0];
  if (d >= 0.0) {
    d = sqrt(d);
    if (p < 0.0) {
      x[i] = aux_newton1(o, p - d);
      x[j] = p + d;
    }
    else {
      x[i] = p - d;
      x[j] = aux_newton1(o, p + d);
    }
    return 2;
  }
  else
    return 0;
}

int eqn_cubic(const Polynomial<double>& a, Vector<double>& x) noexcept {
  int i_slope, i_loc;
  double h, y, y1, y2, yref, dx, b[3], c[2], d, *pb;
  const double *pa,
    precy = 1.0e-15,
    prec = 1.0e-6;		// termination criterion


  if (a.size() < 4 || a[3] == 0.0) return eqn_quadratic(a, x);

  h = 1.0 / a[3];			// normalize
  pa = a.first();
  pb = b;
  while (pb <= b + 2) *pb++ = *pa++ * h;

  if (b[0] == 0.0) {			// one root at zero?
    x[0] = 0.0;
    if (aux_quadratic(b + 1, x.first(), 1, 2) == 0)
      return 1;
    else {
      if (x[2] < 0.0) {			// sort results
        x[0] = x[1];
	x[1] = x[2];
	x[2] = 0.0;
      }
      else if (x[1] < 0.0) {
	x[0] = x[1];
	x[1] = 0.0;
      }
      return 3;
    }
  }

  h = -b[2] / 3.0;			// inflection point
  y = b[0] + h * (b[1] + h * (b[2] + h));
  yref = fabs(b[0]) + fabs(h) * (fabs(b[1]) + fabs(h)
  * (fabs(b[2]) + fabs(h)));
  if (fabs(y) < precy * yref) {		// inflection point = root?
    x[0] = x[1] = h;
    c[1] = h + b[2];			// deflation
    c[0] = c[1] * h + b[1];
    return 1 + aux_quadratic(c, x.first(), 0, 2);
  }

  i_loc = (y > 0.0);
  d = sq(b[2]) - 3.0 * b[1];
  if ((i_slope = signum(d)) == 1)	// 2 extrema?
    h += ((i_loc) ? -2.0 / 3.0 : 2.0 / 3.0) * sqrt(d);
  else if (i_slope == 0) {		// saddle point?
    x[0] = h - cbrt(y);
    return 1;
  }
  else {				// monotonous?
    y2 = 3.0 * h;
    y1 = 2.0 * b[2] + y2;
    y2 += y1;
    y1 = h * y1 + b[1];
    dx = y * y1 / (sq(y1) - 0.5 * y * y2);
    h -= dx;
  }

  do {					// 1st root by Halley's method
    y = b[2] + h;			// evaluate function and derivatives
    y1 = 2.0 * y + h;
    y2 = y1 + 3.0 * h;
    y1 = h * y1 + b[1];
    y = (h * y + b[1]) * h + b[0];	// ... looks odd, but saves CPU time
    dx = y * y1 / (sq(y1) - 0.5 * y * y2);
    h -= dx;
  } while (fabs(dx) > prec * fabs(h));
  x[0] = x[2] = h;

  if (i_slope == 1) {
    c[1] = h + b[2];			// deflation; c[2] is 1
    c[0] = c[1] * h + b[1];
    return 1 + aux_quadratic(c, b, x.first(), i_loc, i_loc + 1);
  }
  return 1;
}
#endif	// METHOD selection
