/*************************************************************29.01.2011*/
/* eqn_polynomial: real roots of a polynomial or its derivatives	*/
/*									*/
/* method: The algorithm constructs all derivatives of the polynomial.	*/
/*	   The roots of the jth derivative are the locations of the	*/
/* extrema of the (j-1)th derivative; its roots must lie between these	*/
/* extrema or to the left or right of the outermost extrema, and can be	*/
/* located with the regula falsi. Once they are known, the (j-2)th	*/
/* derivative can be tackled, etc. For an mth order polynomial (having	*/
/* n = m + 1 coefficients), the (m-1)th derivative is a linear function.*/
/*======================================================================*/
/* a		(i)  vector containing the polynomial coefficients
   nd		(i)  order of derivative (0 = roots. 1 = extrema,
   		     2 = inflection points, etc.)
   x		(o)  vector of solutions (its size is increased if
   		     necessary, but not decreased)
   RESULT	(o)  number of valid solutions
   EXCPTIONS	     "option": nd negative or too large			*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double eps = 1.0e-12;


static double poly(double x, void *ptr);


int eqn_polynomial(const Polynomial<double>& a, int nd, Vector<double>& x) {
  int i, j, jref,
    n = a.size();
  double r, s, xa, xb, ya, yb;

  if (nd < 0) throw MthException(MthException::option);

  for (i = n - 1; i >= 0; i--) {
    if (a[i] == 0.0)
      n--;
    else
      break;
  }

  if (n <= 1 || nd > n - 2) return 0;

  Vector<int> k(n - 1);			// number of roots of derivative
  Vector<Polynomial<double> > b(n - 1);	// coefficients of derivatives
  Matrix<double> t(n - 1, n - 1);	// locations of roots

  b[0] = a;
  s = b[0].scale();

  for (i = 0, r = 1.0; i <= n - 2; i++) r = fmax(r, fabs(b[0][i]));
  r += 1.0;				// radius of root circle

  for (j = 1; j <= n - 2; j++) {	// calculate derivatives
    b[j].size(n - j);
    for (i = 0; i < n - j; i++) b[j][i] = (i + 1) * b[j - 1][i + 1];
  }

  jref = n - 2;				// highest derivative: linear
  t[jref][0] = -b[jref][0] / b[jref][1];
  k[jref] = 1;

  for (j = n - 3; j >= nd; j--) {
    k[j] = 0;
    xa = -r;				// left boundary
    ya = b[j].eval(xa);
    xb = t[jref][0];
    yb = b[j].eval(xb);
    if (ya * yb <= 0.0) {
      t[j][0] = eqn_regulafalsi(poly, &b[j], xa, xb, ya, yb, eps * r);
      k[j]++;
    }
    for (i = 1; i < k[jref]; i++) {
      xa = xb;
      ya = yb;
      xb = t[jref][i];
      yb = b[j].eval(xb);
      if (ya * yb <= 0.0) {
	t[j][k[j]] = eqn_regulafalsi(poly, &b[j], xa, xb, ya, yb, eps * r);
	k[j]++;
      }
    }
    xa = xb;				// right boundary
    ya = yb;
    xb = r;
    yb = b[j].eval(xb);
    if (ya * yb <= 0.0) {
      t[j][k[j]] = eqn_regulafalsi(poly, &b[j], xa, xb, ya, yb, eps * r);
      k[j]++;
    }
    if (k[j] > 0) jref--;
  }

  if (x.size() < k[nd]) x.size(k[nd]);
  for (i = 0; i < k[nd]; i++) x[i] = t[nd][i] / s;
  return k[nd];
}

/************************************************************************/
/* poly: nth order polynomial						*/
/************************************************************************/
/* x		(i)  abscissa
   ptr		(i)  coefficients					*/
/*======================================================================*/
static double poly(double x, void *ptr) {
  Polynomial<double> *b;
  b = static_cast<Polynomial<double> *>(ptr);
  return b->eval(x);
}
