/*************************************************************05.03.2002*/
/* eqn_quartic: real roots of a quartic polynomial			*/
/* Equations of lesser degree are solved by the appropriate formulas.	*/
/* The solutions are arranged in ascending order.			*/
/*									*/
/* Ref.: R. Zurmuehl, Praktische Mathematik fuer Ingenieure und		*/
/* Physiker, Springer, Berlin 1965, p. 60ff				*/
/*									*/
/* with coefficient scaling in order to minimize round-off errors	*/
/*======================================================================*/
/* a		(i)  vector containing the polynomial coefficients
   x		(o)  result vector
   RESULT	(o)  number of valid solutions				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


int eqn_quartic(const Polynomial<double>& a, Vector<double>& x) noexcept {
  int i, ix, nreal;
  double d0, d1, h, t, z, sc;
  Polynomial<double> c(4), b = a;

  if (a.size() < 5 || a[4] == 0.0) return eqn_cubic(a, x);

  b.size(5);
  sc = b.scale();

  c[3] = 1.0;				// cubic resolvent
  c[2] = b[2];
  c[1] = b[1] * b[3] - 4 * b[0];
  c[0] = sq(b[1]) + b[0] * (sq(b[3]) - 4 * b[2]);
  if ((i = eqn_cubic(c, x)) == 0) return 0;
  z = x[i - 1];				// take the largest root

  nreal = 0;
  t = sqrt(0.25 * sq(z) - b[0]);
  for (i = -1, ix = 0; i <= 1; i += 2) {
    d0 = -0.5 * z + i * t;		// coeffs. of quadratic factor
    d1 = (t != 0.0) ? i * 0.5 * (d0 * b[3] - b[1]) / t
    : 0.5 * b[3] + i * sqrt(0.25 * sq(b[3]) - z - b[2]);
    h = 0.25 * sq(d1) - d0;
    if (h >= 0.0) {
      h = sqrt(h);
      nreal += 2;
      x[ix++] = -0.5 * d1 - h;
      x[ix++] = -0.5 * d1 + h;
    }
  }

  for (i = 0; i < nreal; i++) {		// single Newton step
    d0 = b[0] + x[i] * (b[1] + x[i] * (b[2] + x[i] * (b[3] + x[i])));
    d1 = b[1] + x[i] * (2 * b[2] + x[i] * (3 * b[3] + 4 * x[i]));
    if (d1 != 0.0) x[i] -= d0 / d1;
    x[i] /= sc;				// reverse scaling
  }

  if (nreal == 4) {			// sort results
    if (x[2] < x[0]) std::swap(x[0], x[2]);
    if (x[3] < x[1]) std::swap(x[1], x[3]);
    if (x[1] < x[0]) std::swap(x[0], x[1]);
    if (x[3] < x[2]) std::swap(x[2], x[3]);
    if (x[2] < x[1]) std::swap(x[1], x[2]);
  }
  return nreal;
}
