/*************************************************************09.10.2005*/
/* eqn_quintic: real roots of a 5th order algebraic equation		*/
/*									*/
/* method: The 1st root is obtained with a regula falsi iteration.	*/
/*	   Deflation of the polynomial with this root yields a 4th	*/
/*	   order polynomial, which is then solved analytically.		*/
/*									*/
/* Equations of lesser degree are solved by the appropriate formulas.	*/
/* The solutions are arranged in ascending order.			*/
/*======================================================================*/
/* a		(i)  vector containing the polynomial coefficients
   x		(o)  vector of roots
   RESULT	(o)  number of valid solutions				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


static double poly5(double x, void *ptr);


int eqn_quintic(const Polynomial<double>& a, Vector<double>& x) noexcept {
  int i, nreal;
  double r, xa, xb, ya, yb, h;
  Polynomial<double> b(6);

  if (a.size() < 6 || a[5] == 0.0) return eqn_quartic(a, x);

  for (i = 0, r = 1.0, h = 1.0 / a[5]; i <= 4; i++) {
    b[i] = a[i] * h;			// reduce polynomial
    r = fmax(r, fabs(b[i]));
  }
  b[5] = 1.0;
  r += 1.0;				// radius of root circle

  xa = -r;				// find 1st root
  ya = poly5(xa, &b);
  xb = r;
  yb = poly5(xb, &b);
  h = eqn_regulafalsi(poly5, &b, xa, xb, ya, yb, mth_eps0 * r);

  b[5] = 0.0;				// deflation of a-polynomial
  b[4] = a[5];
  for (i = 3; i >= 0; i--) b[i] = a[i + 1] + h * b[i + 1];
  nreal = 1 + eqn_quartic(b, x);

  x[nreal - 1] = h;			// add the 1st root
  for (i = nreal - 2; i >= 0; i--) {	// sort results
    if (x[i + 1] > x[i])
      break;
    else
      std::swap(x[i], x[i + 1]);
  }
  return nreal;
}

/************************************************************************/
/* poly5: 5th order polynomial						*/
/************************************************************************/
/* x		(i)  abscissa
   ptr		(i)  coefficients					*/
/*======================================================================*/
static double poly5(double x, void *ptr) {
  Polynomial<double> *b;
  b = static_cast<Polynomial<double> *>(ptr);
  return b->eval(x);
}
