/*************************************************************09.10.2005*/
/* eqn_regulafalsi: fast 1-dimensional regula falsi iteration (Illinois	*/
/*		    algorithm, a.k.a. pegasus algorithm)		*/
/*									*/
/* The input parameters are not changed. Convergence is certain if	*/
/* ya and yb have opposite signs.					*/
/*======================================================================*/
/* f(x, s)	(i)  object function (y = f(x,s) -> 0)
   s		(i)  structure passing additional data to f()
   xa, xb	(i)  initial values
   ya, yb	(i)  residuals at xa and xb (ya = f(xa), yb = f(xb))
   eps		(i)  desired absolute precision
   RESULT	(o)  x value for which y = 0
   EXCEPTIONS	     ch_sign: ya and yb have the same signs
		     exceptions thrown by f()				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


double eqn_regulafalsi(function<double(double, void *)> f, void *s,
double xa, double xb, double ya, double yb, double eps) {
  double x, dx, h, y;

  if (ya * yb > 0.0) throw MthException(MthException::ch_sign);
  if (ya == 0.0) return xa;		// for the trivial case ya = yb = 0
  if (yb == 0.0) return xb;
  while (1) {				// loop terminated from within!
    dx = ya * (xa - xb) / (ya - yb);
    x = xa - dx;
    if (fabs(dx) < eps) break;
    y = f(x, s);
    h = y * ya;
    if (h < 0.0) {
      xb = xa;
      yb = ya;
    }
    else if (h == 0.0)
      break;
    else
      yb *= 0.5;
    xa = x;
    ya = y;
  }
  return x;
}

/*************************************************************24.09.2017*/
/* eqn_regulafalsi variant using a single initial value and a simple	*/
/*		   search strategy before the regula falsi iteration	*/
/*									*/
/* This is for cases where xa is known to be close to the solution and	*/
/* f() does not have extrema within the search range.			*/
/*======================================================================*/
/* nsteps	(i)  max. number of search steps			*/
/************************************************************************/
double eqn_regulafalsi(function<double(double, void *)> f, void *s,
double xa, int nsteps, double eps) {
  int i;
  double xb, ya, yb, dx;
  const double
    f1 = 0.001,				// rel. increment for differentiation
    f2 = 1.2;				// Newton step acceleration

  if ((ya = f(xa, s)) == 0.0) return xa;
  dx = eps + xa * f1;
  xb = xa + dx;
  if ((yb = f(xb, s)) == 0.0) return xb;
  if (ya * yb < 0.0) return eqn_regulafalsi(f, s, xa, xb, ya, yb, eps);
  for (i = 0; i < nsteps; i++) {
    dx = f2 * yb * (xa - xb) / (ya - yb);
    xa = xb;
    ya = yb;
    xb -= dx;
    if ((yb = f(xb, s)) == 0.0) return xb;
    if (ya * yb < 0.0) return eqn_regulafalsi(f, s, xa, xb, ya, yb, eps);
  }
  throw MthException(MthException::ch_sign);
}
