/************************************************************************/
/* root of a 1-dimensional function using Steffensen's accelerated	*/
/* substitution algorithm - complex and multicomplex variants		*/
/************************************************************************/
#include <mathC.h>

using namespace std;


/*************************************************************17.03.2018*/
/* eqn_steffensen: 1-dimensional root finder using Steffensen's		*/
/*		   accelerated substitution algorithm			*/
/*======================================================================*/
/* f(x, u)	(i)  iteration step function (solution x = f(x, u))
   u		(i)  structure for passing additional information to f()
   x		(i)  initial value
   kmax		(i)  number of iteration steps
   eps		(i)  convergence criterion (max. x deviation)
   RESULT	(o)  x value
   exceptions	     inflection:  algorithm failure (inflection point)
		     convergence: no convergence after kmax steps
		     exceptions thrown by f()				*/
/************************************************************************/
complex_d eqn_steffensen(function<complex_d(complex_d, void *)> f, void *u,
const complex_d& x, int kmax, double eps) {
  int i;
  complex_d x1, x2, d, ds,
    x0 = x;

  for (i = 0; i < kmax; i++) {
    x1 = f(x0, u);
    if (norm(x1 - x0) < sq(mth_eps0 * norm(x0))) return x0;
    x2 = f(x1, u);
    d = (x2 - x1) - (x1 - x0);
    if (d == 0.0) throw MthException(MthException::inflection);
    ds = sq(x1 - x0) / d;
    x0 -= ds;
    if (norm(ds) < sq(eps)) return x0;
  }
  throw MthException(MthException::convergence);
}

/*************************************************************18.06.2019*/
/* eqn_steffensen: multicomplex variant					*/
/************************************************************************/
multicomplex eqn_steffensen(function<multicomplex(multicomplex, void *)> f,
void *u, const multicomplex& x, int kmax, double eps) {
  int i;
  multicomplex x1, x2, d, ds,
    x0 = x;
  unsigned int l = x.level();
  x1.init(l);
  x2.init(l);

  for (i = 0; i < kmax; i++) {
    x1 = f(x0, u);
    if ((x1 - x0).norm() < sq(mth_eps0 * x0.norm())) return x0;
    x2 = f(x1, u);
    d = (x2 - x1) - (x1 - x0);
    if (d == 0.0) throw MthException(MthException::inflection);
    ds = sq(x1 - x0) / d;
    x0 -= ds;
    if (ds.norm() < sq(eps)) return x0;
  }
  throw MthException(MthException::convergence);
}
