/*************************************************************19.12.2001*/
/* fit_marquardt: least squares approximation for systems of nonlinear	*/
/*		  equations by a modified Marquardt-Levenberg method	*/
/*======================================================================*/
/* y = f(x,aux)	(i)  subroutine containing the object functions: f
		     calculates the deviation vector y(n) for a given
		     parameter set x(m).
   x		(io) vector of initial values/results; if x is of
   		     size 0, only the r.m.s. deviation is calculated
   n		(i)  number of equations
   aux		(i)  structure containing auxiliary data for the
		     object function (only passed through)
   maxit	(i)  max. number of iteration cycles (recomm'd: 20*m)
   iprot	(i)  protocol control: 0 = no protocol
				       1 = print intermediate results
   RESULT	(o)  root of mean squared deviations
   exceptions	     convergence: too many iteration cycles
   		     n_data:     more unknowns than equations
		     irrelevant: irrelevant parameter
		     exceptions thrown by f()				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


/* xl is the Marquardt parameter; it is increased or decreased by
   xfact at every iteration step. If xl becomes less than xlcrit, it
   is set to zero. The Jacobian matrix is recomputed after m success-
   ful iteration steps, or after each unsuccessful step.
*/
const double
  xinc = 1.0e-5,			// differentiation increment
  eps = 1.0e-6,				// termination criterion
  epsaux = 1.0e-4,
  xlinit = 0.1,				// Marquardt parameter initial value
  xlmax = 2000.0,			// Marquardt parameter max. value
  xlthr = 0.001,			// Marquardt parameter threshold
  factinit = 10.0;			// Marquardt parameter initial factor


double fit_marquardt(function<Vector<double>(const Vector<double>&,
void *)> f, Vector<double>& x, int n, void *aux, int maxit, int iprot) {
  int i, j, k, age, inv_flag, success,
    m = x.size();
  double xl, xlcrit, xfact, ynorm, ynorm2, h, delta, grdnt, yscal;
  Vector<double> trace(m), grad(m), dx(m), xnew(m), y(n), y0(n), yh(n);
  Matrix<double> asq(m, m), ainv(m, m), jacobian(m, n);

  if (m > n || n <= 0 ) throw MthException(MthException::n_data);
				
  y0 = f(x, aux);			// trial evaluation of f()
  ynorm = sqrt(y0.sum2() / n);
  if (m <= 0) return ynorm;		// only rms calculation?

  xl = xlinit;				// set control parameters
  xfact = factinit;
  xlcrit = xlthr;
  age = m;
  trace = 0.0;

  for (k = 0; k < maxit; k++) {		// main iteration loop
    if (iprot == 1) {
      printf("\nrms: %e\n", ynorm);
      x.print(stdout, "%16.7e");
      printf("\n");
    }
    if (age >= m) {			// calc. new Jacobian
      age = 0;
      for (i = 0; i < m; i++) {		// differentiate
	h = ynorm * sqrt(m * trace[i]);
	delta = fmax(sq(xinc), h);	// differentiation increment
	delta = xinc * fmax(delta, fabs(x[i]));
	h = x[i];
	x[i] = h + delta;
	y = f(x, aux);
	x[i] = h - delta;
	yh = f(x, aux);
	x[i] = h;
	jacobian[i] = (y - yh) * (1.0 / (2 * delta));
      }

      asq = jacobian & jacobian;	// jacobian * transpose(jacobian)

      for (i = 0; i < m; i++) {
	if (asq[i][i] <= 0.0)
	  throw MthException(MthException::irrelevant, i);
	trace[i] = 1.0 / sqrt(asq[i][i]);// scale square matrix
	for (j = 0; j < m; j++) asq[i][j] *= trace[i];
	for (j = 0; j < m; j++) asq[j][i] *= trace[i];
      }
      ainv = asq;
      inv_flag = 1;
    }
    grad = jacobian * y0;		// actual gradient vector
    grad *= trace;
    grdnt = sqrt(grad.sum2() / m) / (xinc + ynorm);

    do {				// determine correction vector
      do {
	if (xl > 2.0) {			// by gradient method
	  dx = grad * (1.0 / (1.0 + xl));
	  break;
	}
	else {				// by Marquardt's method
	  for (i = 0; i < m; i++) asq[i][i] = 1.0 + xl;
	  try {
	    dx = lin_gauss(asq, grad);
	    break;
	  }
	  catch (MthException& e) {
	    xl = (xl < xlcrit) ? xlcrit : xl * xfact;
	  }
	}
      } while (1);
      yscal = grad * dx;
      if (iprot == 1) printf("lambda=%e   gradient=%e\n", xl, grdnt);
      if (xl > xlmax) return ynorm;	// no gradient? finish!
      dx *= trace;
      xnew = x - dx;			// new parameter set - try it!

      try {
	y = f(xnew, aux);
      }
      catch (...) {
	xl = (xl < xlcrit)? xlcrit : xl * xfact;// get out of forbidden area
	success = 0;
	continue;
      }
      if ((ynorm2 = sqrt(y.sum2() / n)) > (ynorm * (1.0 - eps))) {
	success = 0;			// no improvement?
	if (xl < xlcrit) {		// calc. new lambda threshold
	  if (inv_flag != 0) {
	    lin_matinv(ainv);
	    inv_flag = 0;
	  }
	  for (i = 0, xlcrit = 0.0; i < m; i++) xlcrit += ainv[i].sum2();
	  xlcrit = m / sqrt(xlcrit);
	}
	xl = (xl < xlcrit) ? xlcrit : xl * xfact;// increase lambda
	age *= m;			// Jacobian matrix too old?
	if (age >= 1) break;
	if (grdnt < 1.0e-3 || xl > 20.0) {// convergence?
	  for (i = 0, success = 1; i < m; i++) {
	    if (fabs(dx[i]) / (epsaux + fabs(x[i])) > eps) success = 0;
	  }
	  if (success) return ynorm;
	}
      }
      else {				// successful step
	success = 1;
	x = xnew;
	y0 = y;
	if (ynorm2 <= 0.0) return ynorm2;// exact solution?
	xfact = 2.0 + n * fabs((sq(ynorm2) - sq(ynorm)) / yscal);
	xfact = fmax(2.0, fmin(10.0, xfact));
	xl /= xfact;			// decrease Marquardt parameter
	if (xl > 1.0)
	  xl = 1.0;
	else if (xl < xlcrit)
	  xl = 0.0;
	ynorm = ynorm2;
	age++;
      }
    } while (!success);
  }
  throw MthException(MthException::convergence);
}
