/************************************************************************/
/* Minimization and parameter fitting with Powell's method		*/
/*									*/
/* adapted from "Numerical Recipes"					*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  eps = 1.0e-7,				// desired precision
  gold = 1.618034,			// sectio aurea ratio
  cgold = 2.0 - gold,
  glimit = 100.0;			// limit for parabola extrapolation
const int kmax_linmin = 100;


// auxiliary structures
struct auxdata2 {			// for wrap_f()
  Vector<double> *x, *xa;
  const Vector<double> *u;
  void *aux;
  double (*f)(const Vector<double>&, void*);
};


// prototypes of auxiliary subroutines
static double linmin(function<double(const Vector<double>&, void*)>,
  Vector<double>&, const Vector<double>&, void *s);


/*************************************************************31.12.2001*/
/* min_powell: minimization of a multidimensional function by Powell's	*/
/*	       method (successive 1-D minimization with optimizing the	*/
/*	       search direction)					*/
/*======================================================================*/
/* y=f(x, s)	(i)  object function, calculates scalar property y for
		     parameter vector x; y will be minimized.
   x		(io) parameter vector
   s		(i)  structure for passing auxiliary data to the object
		     function
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
				    1 = print intermediate results
   RESULT	(o)  f() value at the minimum
   exceptions	     convergence					*/
/************************************************************************/
double min_powell(function<double(const Vector<double>&, void *)> f,
Vector<double>& x, void *s, int maxit, int iprot) {
  int i, k, iopt,
    m = x.size();
  double y2, y1, y0, gain, h;
  Vector<double> xinit(m), uavg(m), xextr(m);
  Matrix<double> u(m, m);

  u.I();				// set direction vectors

  y2 = f(x, s);

  for (k = 0; k < maxit; k++) {		// main iteration loop
    xinit = x;				// save initial parameters
    y0 = y2;
    gain = 0.0;
    for (i = 0, iopt = 0; i < m; i++) {	// minimize along each
      y1 = y2;				// direction successively
      y2 = linmin(f, x, u[i], s);
      if (iprot > 0) {
	printf("%d %d y = %e\n", k, i, y2);
	x.print(stdout, "", " %12.5e", "\n");
      }
      if (y1 - y2 > gain) {		// Did this give the largest
	gain = y1 - y2;			// decrease in value?
	iopt = i;
      }
    }

    if (y0 - y2 < eps * 0.5 * fabs(y0 + y2)) return y2;

    uavg = x - xinit;			// average direction
    if (uavg.sum2() < sq(eps) * x.sum2()) return y2;
    xextr = x + uavg;			// extrapolate along uavg
    try {
      y1 = f(xextr, s);
    }
    catch (...) {
      y1 = mth_high;
    }
    if (iprot > 0) {
      printf("%d %d y = %e\n", k, m, y1);
      xextr.print(stdout, "", " %12.5e", "\n");
    }
    if (y1 < y2) {			// successful extrapolation
      x = xextr;			// use it as new starting point
      y2 = y1;
      h = 2.0 * (y0 - 2.0 * y2 + y1) * sq(y0 - y2 - gain)
      - gain * sq(y0 - y1);
      if (h <= 0.0) {			// minimize along average direction
	y2 = linmin(f, x, uavg, s);
	if (iprot > 0) {
	  printf("%d %d y = %e\n", k, m + 1, y2);
	  x.print(stdout, "", " %12.5e", "\n");
	}
	u[iopt] = uavg;
      }
    }
  }
  throw MthException(MthException::convergence);
}

/*************************************************************31.12.2001*/
/* linmin: one-dimensional minimization of a multidimensional function	*/
/*======================================================================*/
/* y=f(x, s)	(i)  object function
   x		(io) parameter vector
   u		(i)  direction vector
   s		(i)  structure passing auxiliary data to f()
   RESULT	(o)  object function value at the minimum
   exceptions	     failure: no local minimum detected			*/
/************************************************************************/
static double linmin(function<double(const Vector<double>&, void *)> f,
Vector<double>& x, const Vector<double>& u, void *s) {
  int i, k,
    m = x.size();
  double d, pxlim;
  point2 a, b, c, p;

  auto wrap_f = [&](double q, void *sv) -> double {
    Vector<double> xa = u * q + x;
    return f(xa, s);
  };

  /* The problem is to find a scalar parameter q for which the function
     value at x + q * u is a minimum.
     The first step is the calculation of brackets for this minimum.
  */
  a.x = 0.0;				// 1st guess
  a.y = wrap_f(a.x, s);
  for (i = 0, b.x = 0.0; i < m; i++) {
    d = fabs(u[i]) / (fabs(x[i]) + eps);// nonzero direction vector?
    if (d > b.x) b.x = d;
  }
  if (b.x == 0.0) throw MthException(MthException::failure);
  b.x = eps / b.x;			// 2nd guess
  
  while (true) {			// safety catch in case f()
    try {				// cannot be evaluated
      b.y = wrap_f(b.x, s);
      break;
    }
    catch (...) {
      b.x = 0.5 * (a.x + b.x);
    }
  }
  if (b.y > a.y) std::swap(a, b);	// wrong direction?
  c.x = b.x + gold * (b.x - a.x);	// 3rd guess
  while (true) {			// safety catch in case f()
    try {				// cannot be evaluated
      c.y = wrap_f(c.x, s);
      break;
    }
    catch (...) {
      c.x = 0.5 * (b.x + c.x);
    }
  }

  for (k = 0; k < kmax_linmin; k++) {	// search for brackets
    if (b.y < c.y) break;		// ... succeeded!
    p.x = min_parabola(a, b, c);	// otherwise extrapolate
    pxlim = b.x + glimit * (c.x - b.x);	// limit of safe extrapolation

    // Now there are several possibilities to investigate:
    if ((b.x - p.x) * (p.x - c.x) > 0.0) {// p between b and c
      p.y = wrap_f(p.x, s);
      if (p.y < c.y) {			// minimum detected?
	a = b;				// ... between b and c
	b = p;
	break;
      }
      else if (p.y > b.y) {
	c = p;				// ... between a and p
	break;
      }
    }
    else if ((c.x - p.x) * (p.x - pxlim) > 0.0) {// p between c and limit
      p.y = wrap_f(p.x, s);
      if (p.y < c.y) {
	b = c;
	c = p;
      }
    }
    p.x = ((p.x - pxlim) * (pxlim - c.x) >= 0.0) ? pxlim
      : c.x + gold * (c.x - b.x);
    while (true) {			// safety catch again
      try {				// cannot be evaluated
	p.y = wrap_f(p.x, s);
	break;
      }
      catch (...) {
	p.x = 0.5 * (p.x + c.x);
      }
    }
    a = b;				// discard oldest point
    b = c;
    c = p;
  }
  if (b.y >= a.y || b.y >= c.y)
    throw MthException(MthException::failure);

  /* Now that it is certain that the points a, b, and c bracket a
     minimum, this can be located by Brent's method.
  */
  p.y = min_brent(wrap_f, s, a.x, b.x, c.x, mth_eps1 * fabs(b.x), p.x);
  x += u * p.x;
  return p.y;
}

/*************************************************************28.12.2001*/
/* fit_powell: nonlinear parameter fitting by Powell's method 		*/
/*======================================================================*/
/* y=obj(x, s)	(i)  object function, calculates residual vector y[n] for
		     parameter array x[m], returns error flag; y*y will
		     be minimized.
   x		(io) parameter vector
   n		(i)  number of equations
   s		(i)  structure containing auxiliary data for the object
		     function (just passed through)
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
				    1 = print intermediate results
   RESULT	(o)  rms deviation at the minimum
   EXCEPTIONS	     n_data: more unknowns than equations
   		     %_fail: failure of internal minimizer
   		     exceptions thrown by obj()				*/
/************************************************************************/
double fit_powell(function<Vector<double>(const Vector<double>&,
void *)> obj, Vector<double>& x, int n, void *s, int maxit, int iprot) {
  if (x.size() > n) throw MthException(MthException::n_data);
  return sqrt(min_powell([&](const Vector<double>& x, void *sv)
  -> double {
    return (obj(x, s)).sum2();
  }, x, s, maxit, iprot) / n);
}
