/*************************************************************12.11.2006*/
/* fit_simplex: multidimensional parameter fitting with the simplex	*/
/*		method of Nelder and Mead				*/
/*									*/
/* J.A. Nelder, R. Mead, The Computer Journal ?? (????) 308-313.	*/
/*									*/
/* W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,		*/
/* Numerical Recipes, Cambridge University Press, Cambridge 1986,	*/
/* section 7.1								*/
/*									*/
/* This algorithm finds minima where many other methods fail, but	*/
/* convergence is AWFULLY slow, of 1st order at best. It should never	*/
/* be used to determine final values of model parameters, but merely	*/
/* to generate initial values for the Marquardt-Levenberg or Powell	*/
/* methods								*/
/*======================================================================*/
/* f(x,aux)	(i)  object function, calculates residual vector y(n) for
		     parameter vector x(m), returns error flag; y*y will
		     be minimized
   x		(io) parameter vector
   dx		(i)  initial step size of x
   n		(i)  number of equations
   aux		(i)  structure containing auxiliary data for f(), i.e.,
   		     data simply passed through
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
   				    1 = print intermediate results
   RESULT	(o)  rms deviation at the minimum
   exceptions	     n_data: too many unknowns
		     increment:   dx == 0
   		     convergence: no convergence
   		     exceptions thrown by f()				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  f1 = 1.0,				// reflection factor
  f2 = 0.5,				// contraction factor
  f3 = 2.0,				// expansion factor
  eps = 1.0e-2;				// relative accuracy


double fit_simplex(function<Vector<double>(const Vector<double>&,
void *)> f, Vector<double>& x, const Vector<double>& dx, int n,
void *aux, int maxit, int iprot) {
  int i, j, ilo, ihi, ihi2, k,
    m = x.size();
  double sh, shh;
  Vector<double> s(m + 1), y(n), uh(m), uhh(m), uc(m);
  Matrix<double> u(m + 1, m);

  if (m > n) throw MthException(MthException::n_data);
  for (i = 0; i < m; i++) {
    if (dx[i] == 0.0) throw MthException(MthException::increment);
  }

  ilo = ihi = 0;
  for (i = 0; i <= m; i++) {		// set up the 1st simplex
    u[i] = x;
    if (i < m) u[i][i] += dx[i];	// u[m] contains original x vector
    y = f(u[i], aux);
    s[i] = y.sum2();
    if (s[i] < s[ilo]) ilo = i;
    if (s[i] > s[ihi]) ihi = i;
  }
  for (i = 0, ihi2 = ilo; i <= m; i++) {
    if (i == ihi || i == ilo) continue;
    if (s[i] > s[ihi2]) ihi2 = i;
  }
      
  for (k = 1; k < maxit; k++) {		// main iteration loop
    if (iprot) {
      printf("%d smin = %e\n", k, s[ilo]);
      for (i = 0; i < m; i++) printf("%14.5e", u[ilo][i]);
      printf("\n");
    }

    if (ilo == ihi) throw MthException(MthException::convergence);
    for (j = i = 0; j < m; j++) {	// convergence?
      if (fabs(u[ihi][j] - u[ilo][j]) > 0.5 * eps
      * fabs(u[ihi][j] + u[ilo][j])) {
        i = 1;
	break;
      }
    }
    if (i == 0) {
      x = u[ilo];
      return sqrt(s[ilo] / n);
    }

    for (j = 0; j < m; j++) {		// reflection away from u[ihi]
      for (i = 0, uc[j] = 0.0; i <= m; i++) {
        if (i != ihi) uc[j] += u[i][j];
      }
      uc[j] /= m;			// center of plane opposite u[ihi]

      uh[j] = (1.0 + f1) * uc[j] - f1 * u[ihi][j];
    }
    try {
      y = f(uh, aux);
      sh = y.sum2();
    }
    catch (...) {
      sh = mth_high;
    }

    if (sh < s[ilo]) {			// successful move?
      uhh = (1.0 - f3) * uc + f3 * uh;	// expansion step
      try {
	y = f(uhh, aux);
	shh = y.sum2();
      }
      catch (...) {
	shh = mth_high;
      }
      if (shh < sh) {			// 2nd move successful?
        u[ihi] = uhh;			// ... yes, replace high point
	s[ihi] = shh;
      }
      else {
        u[ihi] = uh;			// ... no, use uh as new high
	s[ihi] = sh;
      }
      ilo = ihi;
      ihi = ihi2;
      for (i = 0, ihi2 = ilo; i <= m; i++) {
        if (i == ihi || i == ilo) continue;
        if (s[i] > s[ihi2]) ihi2 = i;
      }
    }
    else if (sh <= s[ihi2]) {		// new point better than 2nd-highest?
      u[ihi] = uh;			// ... yes, discard previous high
      s[ihi] = sh;
      ihi = ihi2;
      for (i = 0, ihi2 = ilo; i <= m; i++) {
        if (i == ihi || i == ilo) continue;
        if (s[i] > s[ihi2]) ihi2 = i;
      }
    }
    else {
      if (sh <= s[ihi]) {		// new point better than high one?
	u[ihi] = uh;			// ... yes, discard previous high
	s[ihi] = sh;
      }
      uhh = f2 * u[ihi] + (1.0 - f2) * uc; // contraction
      try {
	y = f(uhh, aux);
	shh = y.sum2();
      }
      catch (...) {
	shh = mth_high;
      }
      if (shh <= s[ihi]) {		// improvement?
	u[ihi] = uhh;			// ... yes; use new point
	s[ihi] = shh;
      }
      else {				// if all fails:
        for (i = 0; i <= m; i++) {	// contract towards low point
	  if (i == ilo) continue;
	  u[i] = 0.5 * (u[i] + u[ilo]);
	  y = f(u[i], aux);
	  s[i] = y.sum2();
	}
      }
      for (i = 1, ilo = ihi = 0; i <= m; i++) {
        if (s[i] > s[ihi]) ihi = i;
        if (s[i] < s[ilo]) ilo = i;
      }
      for (i = 0, ihi2 = ilo; i <= m; i++) {
        if (i == ihi || i == ilo) continue;
        if (s[i] > s[ihi2]) ihi2 = i;
      }
    }
  }
  x = u[ilo];
  throw MthException(MthException::convergence);
}
