/*************************************************************07.06.2007*/
/* int_romberg: integration of a function with Romberg's method		*/
/*======================================================================*/
/* f(x, u)	(i)  integrand function (x = scalar)
   u		(i)  structure for passing additional information
   a, b		(i)  interval limits
   eps		(i)  desired accuracy: >0 = absolute
				       <0 = relative
		     Note: If the integral is zero and eps < 0,
		     the subroutine will throw a "conv" exception.
   RESULT	(o)  integral value
   exceptions	     convergence: integral not precise
   		     exceptions thrown by f()				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int mmax = 20;			// max. order of integration
					// (i.e., max. 2^20 strips!)

double int_romberg(function<double(double, void *)> f, void *u, double a,
double b, double eps) {
  int i, j, n;
  double s[2][mmax + 1], acteps, h, acc, k, q;

  if (eps >= 0.0) acteps = eps;

  h = b - a;				// zeroth approximation:
  n = 1;				// a single trapezoid
  s[1][0] = (f(a, u) + f(b, u)) * 0.5 * h;

  for (j = 1; j <= mmax; j++) {		// successive approximation schema
    for (i = 0; i < j; i++) s[0][i] = s[1][i];	// save old approx.
    h *= 0.5;				// new trapezoidal integration
    n *= 2;				// with double panel number
    for (i = 1, acc = 0.0; i < n; i += 2) acc += f(a + i * h, u);
    s[1][0] = 0.5 * s[1][0] + h * acc;
    for (i = 1, k = 1.0; i <= j; i++) {	// Romberg's extrapolation
      k *= 4.0;
      s[1][i] = (k * s[1][i - 1] - s[0][i - 1]) / (k - 1.0);
    }
    q = s[1][j];
    if (eps < 0.0) acteps = fabs(eps * q);
    if (fabs(s[1][j] - s[0][j - 1]) <= acteps) return q;
  }
  throw MthException(MthException::convergence);
}

double int_romberg(function<double(double)> f, double a, double b,
double eps) {
  return int_romberg([&](double x, void *u) -> double {
    return f(x);
  }, NULL, a, b, eps);
}
