/************************************************************************/
/* Integration by means of Simpson's method				*/
/*									*/
/* If the number of panels is not even (a requirement for Simpson's	*/
/* method), the last three are handled with Newton's 3/8 method.	*/
/************************************************************************/

#include <mathC.h>

using namespace std;


/*************************************************************26.03.2008*/
/* int_simpson: integration of a function by means of Simpson's method	*/
/*======================================================================*/
/* f(x, u)	(i)  integrand function (x: scalar)
   u		(i)  pointer to additional data needed by f()
   a, b		(i)  integration range for x
   n		(i)  number of panels
   RESULT	(o)  integral value					*/
/************************************************************************/
double int_simpson(function<double(double, void *)> f, void *u, double a,
double b, int n) {
  int i, nh;
  double h, sum, yh;

  if (n <= 0) return 0.0;

  h = (b - a) / n;			// panel width
  if (n == 1)
    sum = (f(a, u) + f(b, u)) * h * 0.5;
  else {
    nh = (n % 2 == 0) ? n : n - 3;	// number of Simpson panels
    yh = f(a + nh * h, u);
    sum = 0.0;
    if (nh > 0) {
      for (i = 1; i <= nh - 1; i += 2) sum += f(a + i * h, u);
      sum *= 2.0;
      for (i = 2; i <= nh - 2; i += 2) sum += f(a + i * h, u);
      sum *= 2.0;
      sum = (f(a, u) + sum + yh) * h / 3.0;
    }
    if (n != nh) sum += (yh + 3.0 * f(a + (nh + 1) * h, u)
    + 3.0 * f(a + (nh + 2) * h, u) + f(b, u)) * h * 0.375;
  }
  return sum;
}

double int_simpson(function<double(double)> f, double a, double b, int n) {
  return int_simpson([&](double x, void *u) -> double {
    return f(x);
  }, NULL, a, b, n);
}

/*************************************************************27.02.2013*/
/* int_simpson: integration of a table of function values		*/
/*									*/
/* It is assumed that the abscissa values are equidistant!		*/
/*======================================================================*/
/* y		(i)  integrand table (vector of function values f(x))
   a, b		(i)  integration range of x
   RESULT	(o)  integral value					*/
/************************************************************************/
double int_simpson(const Vector<double>& y, double a, double b) {
  int i, nh,
    n = y.size() - 1;			// number of panels
  double h, sum, yh;

  if (n <= 0) return 0.0;

  h = (b - a) / n;			// panel width
  if (n == 1)
    sum = (y[0] + y[1]) * h * 0.5;
  else {
    nh = (n % 2 == 0) ? n : n - 3;	// number of Simpson panels
    yh = y[nh];
    sum = 0.0;
    if (nh > 0) {
      for (i = 1; i <= nh - 1; i += 2) sum += y[i];
      sum *= 2.0;
      for (i = 2; i <= nh - 2; i += 2) sum += y[i];
      sum *= 2.0;
      sum = (y[0] + sum + yh) * h / 3.0;
    }
    if (n != nh) sum += (yh + 3.0 * y[nh + 1] + 3.0 * y[nh + 2] + y[n])
    * h * 0.375;
  }
  return sum;
}
