/************************************************************************/
/* Integration by means of the trapezium method				*/
/*									*/
/* This is the simplest of the Newton-Cotes methods. For a given number	*/
/* of nodes, Simpson's methods usually gives better results. But the	*/
/* trapezoidal method is better for the integration of periodic func-	*/
/* tions over a full period.						*/
/************************************************************************/

#include <mathC.h>

using namespace std;


/*************************************************************31.01.2015*/
/* int_trapez: integration of a function by means of trapezium method	*/
/*======================================================================*/
/* f(x, u)	(i)  integrand function (x: scalar)
   u		(i)  pointer to additional data needed by f()
   a, b		(i)  integration range for x
   n		(i)  number of panels
   RESULT	(o)  integral value					*/
/************************************************************************/
double int_trapez(function<double(double, void *)> f, void *u, double a,
double b, int n) {
  int i;
  double h, sum;

  if (n <= 0) return 0.0;

  h = (b - a) / n;			// panel width
  sum = (f(a, u) + f(b, u)) * 0.5;
  for (i = 1; i < n; i++) sum += f(a + i * h, u);
  return sum * h;
}

double int_trapez(function<double(double)> f, double a, double b, int n) {
  return int_trapez([&](double x, void *u) -> double {
    return f(x);
  }, NULL, a, b, n);
}

/*************************************************************31.01.2015*/
/* int_trapez: integration of a table of function values		*/
/*									*/
/* It is assumed that the abscissa values are equidistant!		*/
/*======================================================================*/
/* y		(i)  integrand table (vector of function values f(x))
   a, b		(i)  integration range of x
   RESULT	(o)  integral value					*/
/************************************************************************/
double int_trapez(const Vector<double>& y, double a, double b) {
  int n = y.size() - 1;			// number of panels
  double h, sum;

  if (n <= 0) return 0.0;

  h = (b - a) / n;			// panel width
  sum = y.sum() - (y[0] + y[1]) * 0.5;
  return sum * h;
}
