/*************************************************************14.08.1998*/
/* ipl_akima: interpolation by Akima's method				*/
/* Ref.: c't 1989, 206-214						*/
/*									*/
/* If x[i]<x<x[i+1], yinterpol = y[i] + b[i]*h + c[i]*h^2 + d[i]*h^3	*/
/* with h = x[i+1] - x[i].						*/
/* Akima calculates the polynomial coefficients. Use ipl_eval to obtain	*/
/* interpolated values.							*/
/*======================================================================*/
/* x, y		(i)  vectors of abscissa and ordinate values
   c		(o)  matrix of Akima polynomial coefficients (c[0][i]:
   		     linear coefficient of interval i)
   exceptions	     dimensions: vector/matrix mismatch
   		     n_data:    too few data points (n < 3)
		     ascending: abscissa values not monotonous		*/
/************************************************************************/
#include <mathC.h>

using namespace std;


static double akslope(double s1, double s2, double s3, double s4);


void ipl_akima(const Vector<double>& x, const Vector<double>& y,
Matrix<double>& c) {
  int i, n1,
    n = x.size();
  double f, yl1, sl1, yl2, sl2, yr1, sr1, yr2;

  if (n != y.size() || n != c.cols() || c.rows() != 3)
    throw MthException(MthException::dimensions);
  if (n < 3) throw MthException(MthException::n_data);
  n1 = n - 1;

  for (i = 0; i < n1; i++) {		// abscissae monotonous?
    c[2][i] = x[i + 1] - x[i];
    if (c[2][i] <= 0.0) throw MthException(MthException::ascending);
  }

  for (i = 0; i < n1; i++) c[1][i] = (y[i + 1] - y[i]) / c[2][i];
  yl1 = c[2][1] * (c[1][1] - 2 * c[1][0]) + y[0];// extrapolate 2 values left
  sl1 = (y[0] - yl1) / c[2][1];
  yl2 = c[2][0] * (c[1][0] - 2 * sl1) + yl1;
  sl2 = (yl1 - yl2) / c[2][0];

  yr1 = c[2][n - 3] * (2 * c[1][n - 2] - c[1][n - 3]) + y[n1];// ...and right
  c[1][n1] = (yr1 - y[n1]) / c[2][n - 3];
  yr2 = c[2][n - 2] * (2 * c[1][n1] - c[1][n - 2]) + yr1;
  sr1 = (yr2 - yr1) / c[2][n - 2];

  c[0][0] = akslope(sl2, sl1, c[1][0], c[1][1]);// calculate smoothed slopes
  c[0][1] = akslope(sl1, c[1][0], c[1][1], c[1][2]);
  for (i = 2; i < n1; i++) c[0][i] = akslope(c[1][i - 2], c[1][i - 1],
  c[1][i], c[1][i + 1]);
  c[0][n1] = akslope(c[1][n - 3], c[1][n - 2], c[1][n1], sr1);

  for (i = 0; i < n1; i++) {		// calc. the other coefficients
    f = c[2][i];
    c[2][i] = (c[0][i + 1] + c[0][i] - 2 * c[1][i]) / sq(f);
    c[1][i] = (3 * c[1][i] - 2 * c[0][i] - c[0][i + 1]) / f;
  }
  c[1][n1] = c[2][n1] = 0.0;
}

/*************************************************************21.01.1993*/
/* akslope: Akima's slope formula					*/
/*======================================================================*/
/* s1, ..., s4	(i)  four slopes
   RESULT	(i)  averaged slope					*/
/************************************************************************/
static double akslope(double s1, double s2, double s3, double s4) {
  double sla = fabs(s2 - s1),
    sra = fabs(s4 - s3);
  return (sla == 0.0 && sra == 0.0) ? 0.5 * (s2 + s3)
  : (sra * s2 + sla * s3) / (sla + sra);
}
