/*************************************************************17.04.2003*/
/* ipl_pspline: interpolation by a periodic cubic spline function:	*/
/* if x[i] < x < x[i+1], yspline = y[i] + b[i]*h + c[i]*h^2 + d[i]*h^3	*/
/* with h = x[i+1] - x[i].						*/
/*									*/
/* This function calculates the polynomial coefficients only. Use	*/
/* ipl_eval to generate interpolated values.				*/
/*									*/
/* If less than 3 data points are given, ipl_pspline sets the polynomial*/
/* coefficients to zero.						*/
/*======================================================================*/
/* x, y		(i)  vectors of abscissa and ordinate values; the last
		     point is regarded as periodic image of the first one
   c		(o)  spline coefficient matrix (c[0] = linear, c[1] =
   		     quadratic, c[2] = cubic coefficients)
   exceptions	     dimensions: vector/matrix size mismatch
		     ascending:  abscissa values not monotonous		*/
/************************************************************************/
#include <mathC.h>

using namespace std;


void ipl_pspline(const SortedVector<double>& x, const Vector<double>& y,
Matrix<double>& c) {
  int i, i1, n1, n2,
    n = x.size();
  double f, h;

  if (n != y.size() || n != c.cols() || c.rows() != 3)
    throw MthException(MthException::dimensions);

  if (n == 1 || n == 2)			// enough data points?
    c = 0.0;				// ... no, just clear coefficients
  else {				// ... yes, calc. cubic spline
    n1 = n - 1;
    n2 = n1 - 1;
    for (i = 0; i < n1; i++) {		// abscissa values monotonous?
      if ((c[1][i] = x[i+1] - x[i]) <= 0.0)
	throw MthException(MthException::ascending);
    }
    c[1][n1] = c[1][0];

    /* The spline coefficients c are obtained from a system of linear
       equations. The matrix representing the lhs. is symmetric and
       cyclic tridiagonal (diagonal elements: c[0], off-diagonal
       elements: c[1]). The right hand side is stored in by c[2]. */

    for (i = 1; i < n1; i++) {		// set up system of equations
      i1 = i - 1;
      c[2][i] = 3 * ((y[i + 1] - y[i]) / c[1][i] - (y[i] - y[i1]) / c[1][i1]);
      c[0][i] = 2 * (c[1][i1] + c[1][i]);
    }
    c[2][n1] = 3 * ((y[1] - y[0]) / c[1][0] - (y[n1] - y[n2]) / c[1][n2]);
    c[0][n1] = 2 * (c[1][n2] + c[1][n1]);

    h = c[1][1];			// simplified Gauss elimination
    for (i = 2; i < n1; i++) {
      f = -h / c[0][i - 1];
      c[0][i] += f * h;
      c[2][i] += f * c[2][i - 1];
      h = c[1][i];
      if (i == n2)
	c[1][i] += f * c[1][i - 1];
      else
	c[1][i] = f * c[1][i - 1];
    }
    f = -h / c[0][n2];
    c[0][n1] += f * c[1][n2];
    c[2][n1] += f * c[2][n2];

    for (i = n - 3; i >= 1; i--) {
      f = -(x[i + 1] -  x[i]) / c[0][i + 1];
      c[1][i] += f * c[1][i + 1];
      c[2][i] += f * c[2][i + 1];
    }

    f = -c[1][0] / c[0][1];
    c[0][n1] += f * c[1][1];
    c[2][n1] += f * c[2][1];

    c[1][0] = c[1][n1] = c[2][n1] / c[0][n1];
    for (i = n2; i >= 1; i--) {
      c[1][i] = (c[2][i] - c[1][i] * c[1][n1]) / c[0][i];
      f = x[i + 1] - x[i];
      c[0][i] = (y[i + 1] - y[i]) / f - (c[1][i + 1] + 2 * c[1][i]) * f / 3.0;
      c[2][i] = (c[1][i + 1] - c[1][i]) / (3 * f);
    }
    f = x[1] - x[0];			// deal with outer intervals
    c[0][0] = c[0][n1] = (y[1] - y[0]) / f - (c[1][1] + 2 * c[1][0]) * f / 3.0;
    c[2][0] = c[2][n1] = (c[1][1] - c[1][0]) / (3 * f);
  }
}
