/*************************************************************04.02.2009*/
/* ipl_reinsch: interpolation by smoothing natural cubic splines	*/
/*									*/
/* C. H. Reinsch, Numer. Math. 10 (1967) 177.				*/
/*======================================================================*/
/* x, y		(i)  abscissa and ordinate vectors
   dy		(i)  uncertainty of y
   s		(i)  smoothing parameter, recommended values are in the
		     range [n - sqrt(2n); n + sqrt(2n)];
		     s = 0 yields the normal (non-smoothing) spline
   yc		(o)  modified (smoothed) y values
   c		(o)  matrix of spline coefficients (c[0][i] is the
   		     linear term of the i-th interval, c[2][i] the
		     cubic term)
   EXCEPTIONS	(o)  MthException
   		       dimension: vector/matrix size mismatch
		       ascending: abscissa vales not monotonous		*/
/************************************************************************/
#include <mathC.h>

using namespace std;


void ipl_reinsch(const SortedVector<double>&x, const Vector<double>& y,
const Vector<double>& dy, double s, Vector<double>& yc, Matrix<double>& c) {
  int i, ih,
    n = x.size(),
    m = n + 1;
  double fa, f, ha, h, p, f2;
  Vector<double> r(m), r1(m), r2(m), t(m), t1(m), u(m), v(m);
  
  if (n != y.size() || n != dy.size() || n != c.cols() || c.rows() != 3)
    throw MthException(MthException::dimensions);

  r[0] = 0.0;
  r1[n - 1] = 0.0;
  r2[n - 1] = r2[n] = 0.0;
  u[0] = u[n - 1] = u[n] = 0.0;
  p = 0.0;

  if ((h = x[1] - x[0]) <= 0.0) throw MthException(MthException::ascending);
  f = (y[1] - y[0]) / h;
  for (i = 1; i <= n - 2; i++) {
    ih = i + 1;
    ha = h;
    if ((h = x[ih] - x[i]) <= 0.0)
      throw MthException(MthException::ascending);
    fa = f;
    f = (y[ih] - y[i]) / h;
    yc[i] = f - fa;
    t[i] = 2 * (h + ha) / 3.0;
    t1[i] = h / 3.0;
    r2[i] = dy[i - 1] / ha;
    r[i] = dy[ih] / h;
    r1[i] = -dy[i] * (1.0 / ha + 1.0 / h);
  }
  for (i = 1; i <= n - 2; i++) {
    ih = i + 1;
    c[0][i] = sq(r[i]) + sq(r1[i]) + sq(r2[i]);
    c[1][i] = r[i] * r1[ih] + r1[i] * r2[ih];
    c[2][i] = r[i] * r2[i + 2];
  }
  f2 = -s;

  while (1) {
    r[1] = 1.0 / (p * c[0][1] + t[1]);
    u[1] = yc[1];
    f = p * c[1][1] + t1[1];
    ha = 0.0;
    h = c[2][1] * p;
    for (i = 2; i <= n - 2; i++) {
      ih = i - 1;
      r1[ih] = f * r[ih];
      r2[i - 2] = ha * r[i - 2];
      r[i] = 1.0 / (p * c[0][i] + t[i] - f * r1[ih] - ha * r2[i - 2]);
      u[i] = yc[i] - r1[ih] * u[ih] - r2[i - 2] * u[i - 2];
      f = p * c[1][i] + t1[i] - h * r1[ih];
      ha = h;
      h = c[2][i] * p;
    }
    for (i = n - 2; i >= 1; i--) u[i] = r[i] * u[i] - r1[i] * u[i + 1]
    - r2[i] * u[i + 2];
    for (i = 0, fa = h = 0.0; i <= n - 2; i++) {
      ha = h;
      h = (u[i + 1] - u[i]) / (x[i + 1] - x[i]);
      v[i] = (h - ha) * sq(dy[i]);
      fa += v[i] * (h - ha);
    }
    ha = v[n - 1] = -h * sq(dy[n - 1]);
    fa -= ha * h;
    ha = f2;
    f2 = fa * sq(p);
    if (f2 >= s || f2 <= ha) break;
    f = 0.0;
    ha = (v[1] - v[0]) / (x[1] - x[0]);
    h = (v[2] - v[1]) / (x[2] - x[1]);
    ha = h - ha;
    f += sq(ha) * r[1];
    r[1] = ha;
    for (i = 2; i <= n - 2; i++) {
      ha = h;
      h = (v[i + 1] - v[i]) / (x[i + 1] - x[i]);
      ha = h - ha - sq(r1[i - 1]) - r2[i - 2] * r[i - 2];
      f += sq(ha) * r[i];
      r[i] = ha;
    }
    if ((h = fa - p * f) <= 0.0) break;
    p += (s - f2) / ((sqrt(s / fa) + p) * h);
  }

  for (i = 0; i <= n - 1; i++) {
    yc[i] = y[i] - p * v[i];
    c[1][i] = u[i];
  }
  for (i = 0; i <= n - 2; i++) {
    ih = i + 1;
    h = x[ih] - x[i];
    c[2][i] = (c[1][ih] - c[1][i]) / (3 * h);
    c[0][i] = (yc[ih] - yc[i]) / h - (h * c[2][i] + c[1][i]) * h;
  }

  h = x[n - 1] - x[n - 2];		// deal with "tail"
  c[0][n - 1] = c[0][n - 2] + h * (2 * c[1][n - 2] + 3 * h * c[2][n - 2]);
  c[2][n - 1] = 0.0;
}
