/*************************************************************16.01.2000*/
/* ipl_spline: interpolation with a natural cubic spline function:	*/
/* if x[i] < x < x[i+1], yspline = y[i] + b[i]*h + c[i]*h^2 + d[i]*h^3	*/
/* with h = x[i+1] - x[i].						*/
/*									*/
/* This function calculates the polynomial coefficients only. Use	*/
/* ipl_eval to generate interpolated values.				*/
/*									*/
/* If less than 3 data points are given, ipl_spline calculates a low-	*/
/* order interpolation or sets the polynomial coefficients to zero.	*/
/************************************************************************/
#include <mathC.h>

using namespace std;


/************************************************************************/
/* ipl_spline classical version (initial and final slope not fixed	*/
/************************************************************************/
void ipl_spline(const SortedVector<double>& x, const Vector<double>& y,
Matrix<double>& c) {
  ipl_spline(x, y, 0, c);
}

/*************************************************************29.05.2012*/
/* ipl_spline: version allowing to set initial or final slope		*/
/*======================================================================*/
/* x, y		(i)  vectors of abscissa and ordinate values
   sflag	(i)  interpolation mode:
		       0 = no restriction of the slopes
		       1 = initial slope (c[0][0]) fixed
		       2 = final slope (c[0][n-1]) fixed
		       3 = both fixed
   c(3, n)	(io) matrix of spline coefficients (i.e., c[0][i] is the
   		     linear coefficient of the i-th interval);
		     if sflag > 0, the first and/or last element of c[0]
		     must contain the fixed slopes
   EXCEPTIONS	     dimensions: vector/matrix size mismatch
   		     opion: wrong sflag
   		     ascending: abscissa values not monotonous		*/
/************************************************************************/
void ipl_spline(const SortedVector<double>& x, const Vector<double>& y,
int sflag, Matrix<double>& c) {
  int i,
    n = x.size(),
    n1 = n - 1,
    n2 = n - 2;
  double f, sinit, sfinal;

  if (sflag < 0 || sflag > 3) throw MthException(MthException::option);
  if (n != y.size() || n != c.cols() || c.rows() != 3)
    throw MthException(MthException::dimensions);

  if (n == 1)				// enough data points?
    c[1][0] = c[2][0] = 0.0;		// ... no, just clear coefficients
  else if (n == 2) {			// ... no, try low-order interpolation
    f = x[1] - x[0];
    if (f <= 0.0) throw MthException(MthException::ascending);
    switch (sflag) {
      case 0:
	c[0][0] = c[0][1] = (y[1] - y[0]) / f;
	c[1][0] = c[1][1] = c[2][0] = c[2][1] = 0.0;
	break;
      case 1:
	c[1][0] = (y[1] - y[0] - f * c[0][0]) / sq(f);
	c[2][0] = 0.0;
	c[0][1] = c[0][0] + 2 * f * c[1][0];
	c[1][1] = c[2][1] = 0.0;
      case 2:
        c[1][1] = c[2][1] = 0.0;
	c[1][0] = -(y[1] - y[0] - f * c[0][1]) / sq(f);
	c[0][0] = c[0][1] - 2 * f * c[1][0];
	c[2][0] = 0.0;
      case 3:
        c[2][0] = (-2 * (y[1] - y[0]) + f * (c[0][0] + c[0][1])) / cb(f);
	c[1][0] = (c[0][1] - c[0][0] - 3 * c[2][0] * sq(f)) / (2 * f);
	c[1][1] = c[2][1] = 0.0;
    }
  }
  else {				// ... yes, calc. cubic spline
    if (sflag & 0x01) sinit = c[0][0];
    if (sflag & 0x02) sfinal = c[0][n1];

    for (i = 0; i < n1; i++) {		// abscissa values monotonous?
      if ((c[1][i] = x[i + 1] - x[i]) <= 0.0)
	throw MthException(MthException::ascending);
    }

    /* The quadratic coefficients are obtained from a set of linear
       equations with a symmetric, tridiagonal system matrix. We store
       the diagonal elements in c[0], the off-diagonal elements in c[1],
       and the right-hand side in c[2].	*/
    for (i = 1; i < n1; i++) {
      c[2][i] = 3 * ((y[i + 1] - y[i]) / c[1][i]
      - (y[i] - y[i - 1]) / c[1][i - 1]);
      c[0][i] = 2 * (c[1][i - 1] + c[1][i]);
    }
    c[1][n1] = 0.0;

    if (sflag & 0x01) {
      c[0][1] -= 0.5 * c[1][0];
      c[2][1] -= 1.5 * ((y[1] - y[0]) / c[1][0] - sinit);
    }
    if (sflag & 0x02) {
      c[0][n2] -= 0.5 * c[1][n2];
      c[2][n2] += 1.5 * ((y[n1] - y[n2]) / c[1][n2] - sfinal);
    }

    for (i = 1; i < n2; i++) {		// simplified Gauss elimination
      f = -c[1][i] / c[0][i];
      c[0][i + 1] += f * c[1][i];
      c[2][i + 1] += f * c[2][i];
    }
    for (i = n2; i >= 1; i--) {
      f = c[1][i];
      c[1][i] = (c[2][i] - c[1][i + 1] * f) / c[0][i];
      c[0][i] = (y[i + 1] - y[i]) / f - (c[1][i + 1] + 2 * c[1][i]) * f / 3.0;
      c[2][i] = (c[1][i + 1] - c[1][i]) / (3 * f);
    }

    f = x[1] - x[0];			// deal with 1st interval
    switch (sflag) {
      case 0:
      case 2:
	c[0][0] = (y[1] - y[0]) / f - f * c[1][1] / 3.0;
	c[1][0] = 0.0;
	c[2][0] = c[1][1] / (3 * f);
	break;
      case 1:
      case 3:
	c[0][0] = sinit;
	c[1][0] = 1.5 * ((y[1] - y[0]) / f - sinit) / f - 0.5 * c[1][1]; 
	c[2][0] = (c[1][1] - c[1][0]) / (3 * f);
	break;
    }
    f = x[n1] - x[n2];			// ... and with last one
    switch (sflag) {
      case 0:
      case 1:
	c[0][n1] = c[0][n2] + f * (2 * c[1][n2] + 3 * f * c[2][n2]);
	c[2][n1] = 0.0;
        break;
      case 2:
      case 3:
	c[0][n1] = sfinal;
        c[1][n1] = -1.5 * ((y[n1] - y[n2]) / f - sfinal) / f - 0.5 * c[1][n2];
	c[2][n1] = 0.0;
	break;
    }
  }
}
