/*************************************************************10.10.1994*/
/* ipl_track: generate initial values for iterative procedures by extra-*/
/*	  polation from previous results (in other words, store arrays	*/
/*	  of results, y[], for various values of the control variable	*/
/*	  x, then extrapolate for a new x value)			*/
/*									*/
/* Track stores up to vmax previous vectors. If more vectors are stored,*/
/* the oldest entry is overwritten.					*/
/*======================================================================*/
/* x		(i)  independent variable
   y		(io) vector of dependent variables
   mode		(i)  mode control:
   		       MthKey::init = initialize (set slot)
			       store = store an array
			       estimate = extrapolate
			       reset = purge extrapolation table
   restr	(i)  restriction of y values:
		       0 = no restriction
		       1 = y value positive
		       2 = y value < 1
		       3 = y between 0 and 1
   slot		(io) storage slot number; used to distinguish between
   		     interpolations of different (x, y) sets in the
		     same program; an output parameter in "init" mode,
		     otherwise an input parameter
   EXCEPTIONS	     option: mode or restr[] too large
   		     range: slot undefined
		     dimension: wrong size of y, most probably because
		       of wrong initialization				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int vmax = 3;			// number of entries stored
const double lhigh = log(mth_high);

struct data {				// data set:
  double x;				//   abscissa
  Vector<double> y;			//   array of ordinate values
  Vector<bool> f;			//   flag: value out of range
};
struct table {				// interpolation table
  Vector<struct data> d;
  int k;				//  number of valid entries
};


void ipl_track(double x, Vector<double>& y, int mode,
const Vector<int>& restr, int& slot) {
  int i,
    n = y.size();
  double h, d12, d13, d23;
  static int k = 0,
    ntbl = 0;
  static Vector<struct table> tbl;

  for (i = 0; i < n; i++) {
    if (restr[i] < 0 || restr[i] > 3)
      throw MthException(MthException::option);
  }
  if (mode != MthKey::init) {
    if (slot > ntbl - 1) throw MthException(MthException::range);
    if (tbl[slot].d[0].y.size() != n)
      throw MthException(MthException::dimensions);
  }

  switch (mode) {
    case MthKey::init:			// set up a new table
      slot = ntbl;
      ntbl++;
      tbl.size(ntbl);
      tbl[slot].d.size(vmax);
      for (i = 0; i < vmax; i++) {
	tbl[slot].d[i].y.size(n);
	tbl[slot].d[i].f.size(n);
      }
      tbl[slot].k = 0;
      break;
    case MthKey::reset:			// purge table
      tbl[slot].k = 0;
      break;
    case MthKey::store:			// enter a new data set
      if (tbl[slot].k == vmax) {
	for (i = 1; i < vmax; i++) {
	  tbl[slot].d[i - 1].x = tbl[slot].d[i].x;
	  tbl[slot].d[i - 1].y = tbl[slot].d[i].y;
	  tbl[slot].d[i - 1].f = tbl[slot].d[i].f;
	}
      }
      else
	tbl[slot].k++;
      k = tbl[slot].k;
      tbl[slot].d[k - 1].x = x;
      for (i = 0; i < n; i++) {
	h = y[i];
	if (restr[i] == 1)
	  h = (tbl[slot].d[k - 1].f[i] = (h <= 0.0)) ? -lhigh : log(h);
	else if (restr[i] == 2)
	  h = (tbl[slot].d[k - 1].f[i] = (h >= 1.0)) ? +lhigh : -log(1.0 - h);
	else if (restr[i] == 3) {
	  tbl[slot].d[k - 1].f[i] = (h <= 0.0 || h >= 1.0);
	  h = (h > 0.0) ? ((h < 1.0) ? -log(1.0 / h - 1.0) : +lhigh) : -lhigh;
	}
	tbl[slot].d[k - 1].y[i] = h;
      }
      break;
    case MthKey::estimate:		// extrapolate
      k = tbl[slot].k;
      if (k == 0) y = 0.0;
      if (k >= 1) y = tbl[slot].d[k - 1].y;
      if (k >= 2) {			// ... linear
	d12 = tbl[slot].d[k - 1].x - tbl[slot].d[k - 2].x;
	h = (x - tbl[slot].d[k - 1].x) / d12;
	for (i = 0; i < n; i++) {
	  if (!tbl[slot].d[k - 1].f[i] && !tbl[slot].d[k - 2].f[i])
	    y[i] += (tbl[slot].d[k - 1].y[i] - tbl[slot].d[k - 2].y[i]) * h;
	}
      }
      if (k >= 3) {			// ... quadratic
	d23 = tbl[slot].d[k - 2].x - tbl[slot].d[k - 3].x;
	d13 = tbl[slot].d[k - 1].x - tbl[slot].d[k - 3].x;
	h = -(x - tbl[slot].d[k - 1].x) * (x - tbl[slot].d[k - 2].x)
	/ (d13 * d12 * d23);
	for (i = 0; i < n; i++) {
	  if (!tbl[slot].d[k - 1].f[i] && !tbl[slot].d[k - 2].f[i]
	  && !tbl[slot].d[k - 3].f[i]) y[i] -= (tbl[slot].d[k - 3].y[i] * d12
	  - tbl[slot].d[k - 2].y[i] * d13 + tbl[slot].d[k - 1].y[i] * d23) * h;
	}
      }
      for (i = 0; i < n; i++) {
	if (restr[i] == 1)
	  y[i] = (y[i] <= -lhigh) ? 0.0 : exp(y[i]);
	else if (restr[i] == 2)
	  y[i] = (y[i] >= lhigh) ? 1.0 : 1.0 - exp(-y[i]);
	else if (restr[i] == 3)
	  y[i] = (y[i] <= -lhigh) ? 0.0 :
	  ((y[i] >= lhigh) ? 1.0 : 1.0 / (1.0 + exp(-y[i])));
      }
      break;
    default:
      throw MthException(MthException::option);
  }
}

void ipl_track(double x, double& y, int mode, int restr, int& slot) {
  Vector<int> restr_vec(1);
  Vector<double> y_vec(1);

  restr_vec = restr;
  y_vec = y;
  ipl_track(x, y_vec, mode, restr_vec, slot);
  if (mode == MthKey::estimate) y = y_vec[0];
}
