/*************************************************************08.05.2008*/
/* lin_jacobi: eigenvalues and -vectors of a symmetric matrix by	*/
/*	       Jacobi's method						*/
/*									*/
/* adapted from "Numerical Recipes"					*/
/*======================================================================*/
/* a		(i)  symmetric matrix for which eigenvalues are sought
		     NOTE: We do not check for symmetry. The subdiagonal
		     elements are ignored.
   d		(o)  vector of eigenvalues, sorted (ascending order)
   v		(o)  matrix of which each row (v[i]), is an eigenvector
   exceptions	     square:      not a square matrix
		     convergence: no convergence			*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int niter = 50;			// max. number of sweeps


void lin_jacobi(Matrix<double> a, Vector<double>& d, Matrix<double>& v) {
  int i, j, k, l,
    n = a.rows();
  double thresh, theta, tau, t, sum, s, c, g, h;
  Vector<double> b(n), z(n);

  if (n != a.cols()) throw MthException(MthException::square);

  if (n <= 2) {				// use analytic method
    lin_eigen2(a, d, v);
    return;
  }

  d.size(n);
  v.size(n, n);
  v.I();
  for (i = 0; i < n; i++) b[i] = d[i] = a[i][i];

  for (k = 0; k < niter; k++) {		// main loop
    z = 0.0;
    sum = 0.0;				// sum of off-diagonal elements
    for (i = 0; i < n - 1; i++) {
      for (j = i + 1; j < n; j++) sum += fabs(a[i][j]);
    }
    if (sum <= 0.0) {			// matrix diagonalized?
      Vector<int> sk(n);		// sort eigenvalues, ascending
      sort_index(d, sk);
      Matrix<double> vh = v;
      b = d;
      for (i = 0; i < n; i++) {
        d[i] = b[sk[i]];
        v[i] = vh[sk[i]];
	if (v[i][i] < 0.0) v[i] *= (-1.0);
      }
      return;
    }

    thresh = (k < 4) ? 0.2 * sum / sq(n) : 0.0;
    for (i = 0; i < n - 1; i++) {
      for (j = i + 1; j < n; j++) {
        g = 100.0 * fabs(a[i][j]);
	// skip rotation for small off-diagonal elements after 4 sweeps
	if (k > 4 && fabs(d[i] + g) == fabs(d[i])
	&& fabs(d[j] + g) == fabs(d[j]))
	  a[i][j] = 0.0;
	else if (fabs(a[i][j]) > thresh) {
	  h = d[j] - d[i];
	  if (fabs(h) + g == fabs(h))
	    t = a[i][j] / h;
	  else {
	    theta = 0.5 * h / a[i][j];
	    t = 1.0 / (fabs(theta) + sqrt(1.0 + sq(theta)));
	    if (theta < 0.0) t = -t;
	  }
          c = 1.0 / sqrt(1 + sq(t));	// cos
	  s = t * c;			// sin
	  tau = s / (1.0 + c);
	  h = t * a[i][j];
	  z[i] -= h;
	  z[j] += h;
	  d[i] -= h;
	  d[j] += h;
	  a[i][j] = 0.0;
	  for (l = 0; l < i; l++) {	// rotation for 0 <= l < i
	    g = a[l][i];
	    h = a[l][j];
	    a[l][i] = g - s * (h + g * tau);
	    a[l][j] = h + s * (g - h * tau);
	  }
	  for (l = i + 1; l < j; l++) {	// rotation for i < l < j
	    g = a[i][l];
	    h = a[l][j];
	    a[i][l] = g - s * (h + g * tau);
	    a[l][j] = h + s * (g - h * tau);
	  }
	  for (l = j + 1; l < n; l++) {	// rotation for j < l < n
	    g = a[i][l];
	    h = a[j][l];
	    a[i][l] = g - s * (h + g * tau);
	    a[j][l] = h + s * (g - h * tau);
	  }
	  for (l = 0; l < n; l++) {	// update eigenvectors
	    g = v[i][l];
	    h = v[j][l];
	    v[i][l] = g - s * (h + g * tau);
	    v[j][l] = h + s * (g - h * tau);
	  }
	}
      }
    }
    b += z;
    d = b;
  }
  throw MthException(MthException::convergence);
}
