/*************************************************************22.08.2015*/
/* lin_rayleigh: nearest eigenvalue and associated eigenvector by	*/
/* 		 Rayleigh quotient iteration				*/
/*									*/
/* [1] R. Kress, Numerical Analysis, Springer, Berlin 1991.		*/
/* [2] G. H. Golub and Ch. F. van Loan, Matrix Computations, 3rd ed.,	*/
/*     Johns Hopkins University Press, Baltimore, 1996.			*/
/* [3] https://en.wikipedia.org/wiki/Rayleigh_quotient_iteration	*/
/*======================================================================*/
/* a		(i)  symmetric matrix for which eigenvalues are sought
		     NOTE: We do not check for symmetry.
   e		(io) eigenvalue estimate/result
   		     NOTE: If the estimate is far off any eigenvalue, it
		     is difficult to predict to which eigenvalue the
		     algorithm will converge.
   v		(o)  normalized eigenvector
   exceptions	     dimensions: dimensions of a and u incompatible 
		     square:     a not a square matrix
		     singular:	 matrix rank too low			*/
/************************************************************************/
#include <mathC.h>

using namespace std;


void lin_rayleigh(const Matrix<double>& a, double& e, Vector<double>& u) {
  int i,
    n = a.rows();
  Vector<double> uh(n); 
  Matrix<double> b(n, n); 

  if (n != a.cols()) throw MthException(MthException::square);

  u.size(n);
  if (n <= 2) {				// use analytic method
    if (n == 1) {
      e = a[0][0];
      u[0] = 1.0;
    }
    else {
      lin_eigen2(a, uh, b);
      i = (fabs(uh[0] - e) > fabs(uh[1] - e));	// eigenvalue nearest to e
      e = uh[i];
      u = b[i];
    }
    return;
  }

  u = sqrt(1.0 / n);			// initial estimate of u
  do {
    b = a;
    for (i = 0; i < n; i++) b[i][i] -= e;	// B = A - e * I
    if ((b * u).sum2() < sq(mth_eps1)) break;
    try {
      uh = lin_gauss(b, u);
    }
    catch (...) {			// if e is accidentally correct,
      int ih, j, jh, k;			// b is singular, and we land here
      Vector<double> d(n - 1), dh(n - 1);
      Matrix<double> c(n - 1, n - 1);

      for (k = 0; k < n; k++) {		// solve as linear problem,
        u[k] = 1.0;			// keeping u[k] fixed
	for (j = 0, jh = 0; j < n; j++) {
	  if (j == k) continue;
	  for (i = 0, ih = 0; i < n; i++) {
	    if (i != k) c[jh][ih++] = b[j][i];
	  }
	  d[jh++] = -b[j][k];
	}
	try {
	  dh = lin_gauss(c, d);
	}
	catch (...) {
	  continue;
	}
	for (j = 0, jh = 0; j < n; j++) {
	  if (j != k) u[j] = dh[jh++];
	}
	u *= sqrt(1.0 / u.sum2());
	return;
      }
      throw MthException(MthException::singular);
    }
    u = uh * sqrt(1.0 / uh.sum2());	// normalized eigenvector
    e = u * (a * u);			// Rayleigh quotient -> eigenvalue
  } while (true);
}
