/*************************************************************28.12.2001*/
/* min_descent: minimum of a function with the steepest-descent method	*/
/*======================================================================*/
/* y=obj(x,aux)	(i)  object function, calculates scalar y for parameter
		     array x[m]
   x		(io) parameter vector
   aux		(i)  structure containing auxiliary data for the object
		     function (just passed through)
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
				    1 = print intermediate results
   RESULT	(o)  rms deviation at the minimum
   EXCEPTIONS	     exceptions thrown by obj()				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  dinc = 1.0e-5,			// differentiation increment
  xinc = 1.0e-2,			// initial step factor
  xf = 5.0,
  eps = 1.0e-6,				// desired accuracy
  epsaux = 1.0e-4;


double min_descent(function<double(const Vector<double>&, void *)> f,
Vector<double>& x, void *aux, int maxit, int iprot) {
  bool converged;
  int i, grad_state, k,
    m = x.size();
  double y, ynew, xs, gsize[2];
  Vector<double> delta(m), dx(m), grad(m), xnew(m), xold(m);
  enum age {fresh, acceptable, old};

  y = f(x, aux);			// see if this works

  xs = sqrt(x.sum2()) * xinc;
  if (xs == 0.0) xs = xinc;
  grad_state = old;

  for (k = 0; k < maxit; k++) {
    if (grad_state == old) {
      for (i = 0; i < m; i++) dx[i] = (x[i] == 0.0) ? dinc : dinc * x[i];
      grad = diff_gradient(f, aux, x, 0, dx);
      gsize[0] = gsize[1];
      gsize[1] = sqrt(grad.sum2());
      grad /= gsize[1];
      grad_state = fresh;
      if (k > 0)
	xs = gsize[1] / fabs(gsize[1] - gsize[0]) * sqrt((x - xold).sum2());
      xold = x;
    }
    delta = xs * grad;
    xnew = x - delta;
    try {
      ynew = f(xnew, aux);
    }
    catch (...) {
      xs /= xf;
      continue;
    }
    if (ynew < y) {			// successful step?
      x = xnew;
      y = ynew;
      xs *= xf;
      grad_state = acceptable;
      if (iprot) {
	printf("%d: y = %e\n", k, ynew);
	x.print(stdout, "", " %12.5e", "\n");
      }
    }
    else {				// no ...
      if (grad_state != fresh)		// ... gradient too old?
	grad_state = old;
      else {				// ... convergence?
	for (i = 0, converged = true; i < m; i++) {
	  if (fabs(delta[i])/(epsaux + fabs(x[i])) > eps) {
	    converged = false;
	    break;
	  }
	}
	if (converged) return y;
	xs /= xf;
      }
    }
  }
  throw MthException(MthException::convergence);
}
