/************************************************************************/
/* multicomplex.C: class multicomplex (multicomplex arithmetics)	*/
/************************************************************************/
#include <mathC.h>

/* NOTE:
   The multicomplex log() function is problematic because of the existence
   of multiple branches.
   A reasonably safe evaluation is possible only if (a) the "most real"
   component (= r_idx(0)) outweighs the imaginary components and (b) the
   imaginary components are smaller in size than 2 pi.
   There are several options:
   1: recursively using the formula of the complex log_e.
   2: using a multicomplex polynomial based on a real Taylor series for
      log_e.
   3: using a multicomplex polynomial based on a real Chebyshev series
      for log_2 by Bradley Alpert (NIST, Boulder, CO, USA).
      dominant-real cases]
   4: iteratively inverting the multicomplex e^x.

   The default setting can be overridden at compile time with the
   compiler option -DMCXLOG=n.

   A similar choice of options exists for the atan() function.
*/
#ifndef MCXLOG
  #define MCXLOG	4
  #define MCXATAN	4
#endif

using namespace std;


const double thr = 1.0e-12;		// threshold for iterations


/************************************************************************/
/* unary operator overloads						*/
/************************************************************************/
multicomplex operator+(const multicomplex& m) noexcept {
  multicomplex mh(m);
  return mh;
}

multicomplex operator-(const multicomplex& m) noexcept {
  multicomplex mh(m);
  mh *= (-1.0);
  return mh;
}

/************************************************************************/
/* binary operator overloads						*/
/************************************************************************/
multicomplex operator+(const multicomplex& m, double d) noexcept {
  multicomplex mh(m);
  mh += d;
  return mh;
}
multicomplex operator+(double d, const multicomplex& m) noexcept {
  return m + d;
}
multicomplex operator+(const multicomplex& ma, const multicomplex& mb) {
  if (ma.level() >= mb.level()) {
    multicomplex mh(ma);
    mh += mb;
    return mh;
  }
  else {
    multicomplex mh(mb);
    mh += ma;
    return mh;
  }
}
  
multicomplex operator-(const multicomplex& m, double d) noexcept {
  multicomplex mh(m);
  mh -= d;
  return mh;
}
multicomplex operator-(double d, const multicomplex& m) noexcept {
  return m * (-1.0) + d;
}
multicomplex operator-(const multicomplex& ma, const multicomplex& mb) {
  if (ma.level() >= mb.level()) {
    multicomplex mh(ma);
    mh -= mb;
    return mh;
  }
  else {
    multicomplex mh(mb);
    mh *= (-1.0);
    mh += ma;
    return mh;
  }
}
  
multicomplex operator*(const multicomplex& m, double d) noexcept {
  multicomplex mh(m);
  mh *= d;
  return mh;
}
multicomplex operator*(double d, const multicomplex& m) noexcept {
  return m * d;
}
multicomplex operator*(const multicomplex& ma, const multicomplex& mb) {
  if (ma.level() >= mb.level()) {
    multicomplex mh(ma);
    mh *= mb;
    return mh;
  }
  else {
    multicomplex mh(mb);
    mh *= ma;
    return mh;
  }
}

multicomplex operator/(const multicomplex& m, double d) noexcept {
  multicomplex mh(m);
  mh *= (1.0 / d);
  return mh;
}
multicomplex operator/(double d, const multicomplex& m) noexcept {
  multicomplex mh;
  mh.init(m.level());
  mh = d;
  mh /= m;
  return mh;
}
multicomplex operator/(const multicomplex& ma, const multicomplex& mb) {
  if (ma.level() >= mb.level()) {
    multicomplex mh(ma);
    mh /= mb;
    return mh;
  }
  else {
    multicomplex mh;
    mh.init(mb.level());
    mh = ma;
    mh /= mb;
    return mh;
  }
}

bool operator==(const multicomplex& m, double d) noexcept {
  unsigned int l = m.level();
  if (l > 1)
    return (*m.p0() == d) && (*m.p1() == 0.0);
  else if (l == 1)
    return (m.c_idx(0) == d);
  else
    return false;
}
bool operator==(double d, const multicomplex& m) noexcept {
  return (m == d);
}
bool operator==(const multicomplex& ma, const multicomplex& mb) noexcept {
  unsigned int l = ma.level();
  if (l != mb.level()) return false;
  if (l > 1)
    return (*ma.p0() == *mb.p0()) && (*ma.p1() == *mb.p0());
  else if (l == 1)
    return (ma.c_idx(0) == mb.c_idx(0));
  else
    return false;
}

bool operator!=(const multicomplex& m, double d) noexcept {
  unsigned int l = m.level();
  if (l > 1)
    return (*m.p0() != d) || (*m.p1() != 0.0);
  else if (l == 1)
    return (m.c_idx(0) != d);
  else
    return true;
}
bool operator!=(double d, const multicomplex& m) noexcept {
  return (m != d);
}
bool operator!=(const multicomplex& ma, const multicomplex& mb) noexcept {
  unsigned int l = ma.level();
  if (l != mb.level()) return true;
  if (l > 1)
    return (*ma.p0() != *mb.p0()) || (*ma.p1() != *mb.p0());
  else if (l == 1)
    return (ma.c_idx(0) != mb.c_idx(0));
  else
    return true;
}


/************************************************************************/
/* special functions							*/
/************************************************************************/
double real(const multicomplex& m) noexcept {
  return m.r_idx(0);
}

multicomplex sin(const multicomplex& m) noexcept {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    *mh.p0() = sin(*m.p0()) * cosh(*m.p1());
    *mh.p1() = cos(*m.p0()) * sinh(*m.p1());
  }
  else
    mh.c_idx(0, sin(m.c_idx(0)));
  return mh;
}

multicomplex cos(const multicomplex& m) noexcept {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    *mh.p0() = cos(*m.p0()) * cosh(*m.p1());
    *mh.p1() = sin(*m.p0()) * sinh(*m.p1()) *(-1.0);
  }
  else
    mh.c_idx(0, cos(m.c_idx(0)));
  return mh;
}

multicomplex tan(const multicomplex& m) noexcept {
  return sin(m) / cos(m);
}

multicomplex sinh(const multicomplex& m) noexcept {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    *mh.p0() = sinh(*m.p0()) * cos(*m.p1());
    *mh.p1() = cosh(*m.p0()) * sin(*m.p1());
  }
  else
    mh.c_idx(0, sinh(m.c_idx(0)));
  return mh;
}

multicomplex cosh(const multicomplex& m) noexcept {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    *mh.p0() = cosh(*m.p0()) * cos(*m.p1());
    *mh.p1() = sinh(*m.p0()) * sin(*m.p1());
  }
  else
    mh.c_idx(0, cosh(m.c_idx(0)));
  return mh;
}
  
multicomplex tanh(const multicomplex& m) noexcept {
  return sinh(m) / cosh(m);
}

multicomplex exp(const multicomplex& m) noexcept {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    *mh.p0() = exp(*m.p0()) * cos(*m.p1());
    *mh.p1() = exp(*m.p0()) * sin(*m.p1());
  }
  else
    mh.c_idx(0, exp(m.c_idx(0)));
  return mh;
}

#if MCXLOG == 1
multicomplex log(const multicomplex& m) {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    multicomplex q, r, iu;
    q.init(l - 1);
    r.init(l - 1);
    iu.init(l - 1);
    iu = 0.0;
    iu.r_idx(0x01 << (l - 2), 1.0);	// imaginary unit, level {l-1}
    q = 2.0 * *m.p1() / (*m.p0() * iu - *m.p1());
    *mh.p0() = 0.5 * (log(sq(*m.p0())) + log1p(sq(*m.p1() / *m.p0())));
    *mh.p1() = 0.5 * iu * log1p(q);
  }
  else
    mh.c_idx(0, log(m.c_idx(0)));
  return mh;
}
#elif MCXLOG == 2
multicomplex log(const multicomplex& m) {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  int i, k;
  const int n = 19;			// max. order of Taylor series
  double xfr, xsc,
    x = m.r_idx(0);
  const double ln2 = log(2.0),
    sqrt2 = sqrt(2.0);
  multicomplex y(m);
  xfr = frexp(x, &k);			// mantissa in [0.5; 1.0], exponent
  xsc = ldexp(sqrt2, -k);		// scaling factor
  y *= xsc;
  y = (y - 1.0) / (y + 1.0);
  for (i = n - 2, mh = 1.0 / static_cast<double>(n); i >= 1; i -= 2) {
    mh *= sq(y);
    mh += 1.0 / static_cast<double>(i);
  }
  mh *= 2.0 * y;
  mh += (static_cast<double>(k) - 0.5) * ln2;
  return mh;
}
#elif MCXLOG == 3
multicomplex log(const multicomplex& m) {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  int k;
  double xfr, xsc,
    x = m.r_idx(0);
  const double a[] = {
    -4.1503749927884376e-01,  4.8089834696299577e-01,
    -8.0149724493844650e-02,  1.7811049887042410e-02,
    -4.4527624714398644e-03,  1.1874033341409927e-03,
    -3.2983426277655650e-04,  9.4238293404332160e-05,
    -2.7486151974305394e-05,  8.1443370460923350e-06,
    -2.4433500383312513e-06,  7.3966579219588530e-07,
    -2.2592596818005443e-07,  7.0659230401327330e-08,
    -2.1952826114457824e-08,  5.7842438255313840e-09,
    -1.7637829658648360e-09,  1.0761035247909333e-09,
    -3.4850519876073720e-10
  },
  ln2 = log(2.0);
  multicomplex y(m);
  xfr = frexp(x, &k);
  xsc = ldexp(1.0, -k);
  y = 4.0 * xsc * y - 3.0;
  mh = (((((((((((((((((a[18] * y + a[17]) * y + a[16]) * y + a[15]) * y
  + a[14]) * y + a[13]) * y + a[12]) * y + a[11]) * y + a[10]) * y
  + a[9]) * y + a[8]) * y + a[7]) * y + a[6]) * y + a[5]) * y + a[4]) * y
  + a[3]) * y + a[2]) * y + a[1]) * y + a[0] + k;
  mh *= ln2;
  return mh;
}
#else
multicomplex log(const multicomplex& m) {
  unsigned int l = m.level();
  multicomplex mh, me;
  mh.init(l);
  me.init(l);
  if (l > 1) {
    unsigned int i, k;
    int count;
    double s, sold;
    k = 0x01 << l;
    for (i = 0, s = 0.0; i < k; i++) s += fabs(m.r_idx(i));
    if (s == 0.0) throw MthException(MthException::range);
    mh.r_idx(0, log(m.r_idx(0)));
    s = mth_high;
    count = 0;
    do {
      sold = s;
      me = m * exp(-mh);
      mh -= 2.0 * (1.0 - me) / (1.0 + me);
      s = mh.r_idx(k - 1);
      if (++count > 10) throw MthException(MthException::convergence);
    } while (fabs(s - sold) > fabs(s) * thr);
  }
  else
    mh.c_idx(0, log(m.c_idx(0)));
  return mh;
}
#endif

#if 0	// slow version: series expansion
multicomplex log1p(const multicomplex& m) {
  unsigned int l = m.level();
  int i, k;
  const int n = 19;			// max. order of Taylor series
  const double ln2 = log(2.0),
    sqrt2 = sqrt(2.0);
  double xfr, xsc, x;
  multicomplex mh, q(m);
  mh.init(l);
  x = (1.0 + q.r_idx(0)) / sqrt2;
  xfr = frexp(x, &k);			// mantissa in [0.5; 1.0], exponent
  xsc = ldexp(sqrt2, -k);		// scaling factor
  if (k != 0) q = (1.0 + q) * xsc - 1.0;
  q /= (q + 2.0);
  for (i = n - 2, mh = 1.0 / static_cast<double>(n); i >= 1; i -= 2) {
    mh *= sq(q);
    mh += 1.0 / static_cast<double>(i);
  }
  mh *= 2.0 * q;
  if (k != 0) mh += (static_cast<double>(k) - 0.5) * ln2;
  return mh;
}
#else	// Brad's trick
multicomplex log1p(const multicomplex& m) {
  unsigned int l = m.level();
  multicomplex mh, u;
  mh.init(l);
  u.init(l);
  u = 1.0 + m;
  mh = (u == 1.0) ? m : log(u) * m / (u - 1.0);
  return mh;
}
#endif

#if MCXATAN == 1	// using multicomplex log1p(): unsafe
multicomplex atan(const multicomplex& m) {
  unsigned int l = m.level();
  if (l == 0) throw MthException(MthException::initialization);
  if (m.r_idx(0) > 1.0) return 0.5 * mth_pi - atan(1.0 / m);
  if (m.r_idx(0) < -1.0) return -0.5 * mth_pi - atan(1.0 / m);
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    multicomplex q, iu;
    q.init(l);
    iu.init(l);
    iu = 0.0;
    iu.r_idx(0x01 << (l - 1), 1.0);	// imaginary unit, level {l}
    q = -0.5 * log1p(-2.0 * m / (iu + m));
//    *mh.p1() = *q.p0();
//    *mh.p0() = (-1.0) * *q.p1();
    mh = q * iu;
  }
  else
    mh.c_idx(0, atan(m.c_idx(0)));
  return mh;
}
#elif MCXATAN == 2	// using Taylor expansions around 0 and 1
multicomplex atan(const multicomplex& m) {
  unsigned int l = m.level();
  if (l == 0) throw MthException(MthException::initialization);
  if (m.r_idx(0) > 1.0) return 0.5 * mth_pi - atan(1.0 / m);
  if (m.r_idx(0) < -1.0) return -0.5 * mth_pi + atan(-1.0 / m);
  if (m.r_idx(0) < 0.0) return -atan((-1.0) * m);
  const double
    c0[] = {1.0, -1.0 / 3.0, 1.0 / 5.0, -1.0 / 7.0, 1.0 / 9.0,	// 1-9
      -1.0 / 11.0, 1.0 / 13.0, -1.0 / 15.0, 1.0 / 17.0,		// 11-17
      -1.0 / 19.0, 1.0 / 21.0, -1.0 / 23.0, 1.0 / 25.0,		// 19-25
      -1.0 / 27.0, 1.0 / 29.0, -1.0 / 31.0, 1.0 / 33.0,		// 27-33
      -1.0 / 35.0, 1.0 / 37.0, -1.0 / 39.0, 1.0 / 41.0},	// 35-41
    c1[] = {mth_pi / 4.0, 1.0 / 2.0, -1.0 / 4.0, 1.0 / 12.0,	// 0-3
      0.0, -1.0 / 40.0, 1.0 / 48.0, -1.0 / 112.0, 0.0,		// 4-8
      1.0 / 288.0, -1.0 / 320.0, 1.0 / 704.0, 0.0,		// 9-12
      -1.0 / 1664.0, 1.0 / 1792.0, -1.0 / 3840.0, 0.0,		// 13-16
      1.0 / 8704.0, -1.0 / 9216.0, 1.0 / 19456.0, 0.0,		// 17-20
      -1.0 / 43008.0, 1.0 / 45056.0, -1.0 / 94208.0, 0.0,	// 21-24
      1.0 / 204800.0, -1.0 / 212992.0, 1.0 / 442368.0, 0.0,	// 25-28
      -1.0 / 950272.0};						// 29
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    int i;
    multicomplex q(m);
    if (fabs(m.r_idx(0)) < 0.5) {	// Taylor expansion at 0.0
      q *= q;
      for (i = 19, mh = c0[20]; i >= 0; i --) {
	mh *= q;
	mh += c0[i];
      }
      mh *= m;
    }
    else {				// Taylor expansion at + 1.0
      q = m - 1.0;
      for (i = 28, mh = c1[29]; i >= 0; i --) {
	mh *= q;
	mh += c1[i];
      }
    }
  }
  else
    mh.c_idx(0, atan(m.c_idx(0)));
  return mh;
}
#elif MCXATAN == 3	// rational approximation
/* Rational approximation on the interval [0,∞), plus reflection, from an
   adaption of the techniques developed for nonreflecting boundary kernels
   (B. Alpert, L. Greengard, L. Hagstrom, SIAM J. Numer. Anal. 37 (2000)
   1138–1164).
*/
multicomplex atan(const multicomplex& m) {
  unsigned int l = m.level();
  if (l == 0) throw MthException(MthException::initialization);
  const double
    a[] = {M_PI_2,            7.9936890512667005e+00, 2.2874345836543414e+01,
      4.5807301602462309e+01, 6.9833957904115201e+01, 8.4136470685122290e+01,
      8.1312230843820890e+01, 6.2974751585413320e+01, 3.8426127583909210e+01,
      1.7740175044448743e+01, 5.6693212747372805e+00, 9.8749170487742506e-01},
    b[] = {1.00000000000,     5.7255602765621789e+00, 1.8207265719463931e+01,
      4.0540732685671591e+01, 6.9051708778393544e+01, 9.3786245991658207e+01,
      1.0359714413690905e+02, 9.3616998654801008e+01, 6.8800366354395384e+01,
      4.0315901336180532e+01, 1.8069338946228228e+01, 5.6693212747355582e+00,
      9.8749170487742980e-01};
  if (m.r_idx(0) > 0) {
    multicomplex p, q;
    p.init(l);
    q.init(l);
    p = -(((((((((((a[11] * m + a[10]) * m + a[9]) * m + a[8]) * m
    + a[7]) * m + a[6]) * m + a[5]) * m + a[4]) * m + a[3]) * m
    + a[2]) * m + a[1]) * m + a[0]);
    q = (((((((((((b[12] * m + b[11]) * m + b[10]) * m + b[9]) * m
    + b[8]) * m + b[7]) * m + b[6]) * m + b[5]) * m + b[4]) * m
    + b[3]) * m + b[2]) * m + b[1]) * m + b[0];
    return p / q + M_PI_2;
  }
  else
    return -atan(-m);
}
#else		// inversion
multicomplex atan(const multicomplex& m) {
  unsigned int l = m.level();
  if (l == 0) throw MthException(MthException::initialization);
  if (m.r_idx(0) > 1.0) return 0.5 * mth_pi - atan(1.0 / m);
  if (m.r_idx(0) < -1.0) return -0.5 * mth_pi - atan(1.0 / m);
  multicomplex mh, me;
  mh.init(l);
  me.init(l);
  if (l > 1) {
    unsigned int i, k;
    int count; 
    double s, sold;
    k = 0x01 << l;
    mh.r_idx(0, atan(m.r_idx(0)));
    s = mth_high;
    count = 0;
    do {
      sold = s;
      me = cos(mh);
      mh -= (sin(mh) - m * me) * me; 
      s = mh.r_idx(k - 1);
      if (++count > 10) throw MthException(MthException::convergence);
    } while (fabs(s - sold) > fabs(s) * thr);
  }
  else
    mh.c_idx(0, atan(m.c_idx(0)));
  return mh;
}
#endif

multicomplex atanh(const multicomplex& m) {
  return 0.5 * log((1.0 + m) / (1.0 - m));
}

multicomplex pow(const multicomplex& m, double d) {
  unsigned int l = m.level();
  if (l == 0)
    throw MthException(MthException::initialization);
  else if (l == 1) {
    multicomplex mh;
    mh.init(l);
    mh.c_idx(0, pow(m.c_idx(0), d));
    return mh;
  }
  else
    return pow(m.r_idx(0), d) * exp(log(m / m.r_idx(0)) * d);
}
multicomplex pow(const multicomplex& m, const multicomplex& e) {
  unsigned int l = max(m.level(), e.level());
  if (m.level() == 0 || e.level() == 0)
    throw MthException(MthException::initialization);
  if (l == 1) {
    multicomplex mh;
    mh.init(l);
    mh.c_idx(0, pow(m.c_idx(0), e.c_idx(0)));
    return mh;
  }
  else
    return exp(log(m) * e);
}

multicomplex sqrt(const multicomplex& m) {
  unsigned int l = m.level();
  multicomplex mh;
  mh.init(l);
  if (l > 1) {
    unsigned int i, k;
    int count;
    double s, sold;
    k = 0x01 << l;
    for (i = 0, s = 0.0; i < k; i++) s += fabs(m.r_idx(i));
    if (s == 0.0) return m;
    mh.r_idx(0, sqrt(m.r_idx(0)));
    s = mth_high;
    count = 0;
    do {
      sold = s;
      mh = 0.5 * (mh + m / mh);
      s = mh.r_idx(k - 1);
      if (++count > 10) throw MthException(MthException::convergence);
    } while (fabs(s - sold) > fabs(s) * thr);
  }
  else
    mh.c_idx(0, sqrt(m.c_idx(0)));
  return mh;
}

multicomplex nthrt(const multicomplex& m, int n) {
  unsigned int l = m.level();
  multicomplex mh, mpwr;
  mh.init(l);
  mpwr.init(l);
  if (l > 1) {
    unsigned int i, k;
    int count;
    double s, sold,
      nh = static_cast<double>(abs(n));;
    if (n == 0) throw MthException(MthException::range);
    if (n == 1) return m;
    if (n == -1) {
      mh = 1.0 / m;
      return mh;
    }
    k = 0x01 << l;
    for (i = 0, s = 0.0; i < k; i++) s += fabs(m.r_idx(i));
    if (s == 0.0) {
      if (n > 0)
	return m;
      else
	throw MthException(MthException::range);
    }
    mh.r_idx(0, pow(m.r_idx(0), 1.0 / nh));
    s = mth_high;
    count = 0;
    do {
      sold = s;
      for (i = 2, mpwr = mh; i < abs(n); i++) mpwr *= mh;
      mh = mh * (nh - 1.0) / nh + m / (nh * mpwr);
      s = mh.r_idx(k - 1);
      if (++count > 10) throw MthException(MthException::convergence);
    } while (fabs(s - sold) > fabs(s) * thr);
    if (n < 0) mh = 1.0 / mh;
  }
  else {
    if (n == 0) throw MthException(MthException::range);
    mh.c_idx(0, pow(m.c_idx(0), 1.0 / static_cast<double>(n)));
  }
  return mh;
}
