/*************************************************************18.02.2016*/
/* ode_ck: ordinary differential equation solver using the method of	*/
/*	   Cash and Karp						*/
/*	   with automatic control of the step size			*/
/*									*/
/* [1] J. R. Cash, A. H. Karp, ACM Transactions on Math. Software 16	*/
/*     (1990) 201-222 (DOI: 10.1145/79505.79507).			*/
/* [2] Wikipedia, https://en.wikipedia.org/wiki/Cash-Karp_method.	*/
/*======================================================================*/
/* f(x, y, s)	(i)  function defining the ODE, i.e. dy/dx = f(x, y, s).
   s		(i)  structure containing pass-through arguments for f()
   x0, y0	(i)  initial values
   xend		(i)  end of integration interval
   dxmin	(i)  minimal (abs.) step size (protection against infinite
   		     loop at a pole of f(); dxmin <= 0 turns feature off)
   RESULT	(o)  function value at xend
   exceptions	     increment = internal step size less than dxmin	*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  eps = 1.0e-6,				// desired relative precision
  hfact = 0.9,				// stepsize safety factor
  hmax = 5.0,				// max. stepsize shrink/grow
  pdown = 0.2,				// error exponent for shrink
  pup = 0.25,				// errorm exponent for grow
  thr = pow(hfact / hmax, 1.0 / pdown),
  a[5][5] = {{0.2, 0.0, 0.0, 0.0, 0.0},
	     {0.075, 0.225, 0.0, 0.0, 0.0},
	     {0.3, -0.9, 1.2, 0.0, 0.0},
	     {-11.0 / 54.0, 2.5, -70.0 / 27.0, 35.0 / 27.0, 0.0},
	     {1631.0 / 55296.0, 175.0 / 512.0, 575.0 / 13824.0,
	      44275.0 / 110592.0, 253.0 / 4096.0}
	    },
  b4[6] = {2825.0 / 27648.0, 0.0, 18575.0 / 48384.0, 13525.0 / 55296.0,
	  277.0 / 14336.0, 0.25},
  b5[6] = {37.0 / 378.0, 0.0, 250.0 / 621.0, 125.0 / 594.0, 0.0,
	  512.0 / 1771.0},
  c[6] = {0.0, 0.2, 0.3, 0.6, 1.0, 0.875};


double ode_ck(function<double(double, double, void *)> f, void *s,
double x0, double y0, double xend, double dxmin) {
  bool at_max, allow_up;
  int i, k;
  double h, crit, y, yt, y4, y5, yh[6];

  h = xend - x0;			// largest possible step size
  y = y0;
  at_max = true;
  while (h * (x0 - xend) < 0.0) {
    allow_up = true;
    while (true) {			// repeat until precision OK
      if ((x0 + h - xend) * h > 0.0) {	// make sure not to exceed xend
	h = xend - x0;
	at_max = true;
      }
      yh[0] = f(x0, y, s);		// calculate slopes at all nodes
      try {
	for (i = 1; i <= 5; i++) {
	  yt = y;
	  for (k = 0; k < i; k++) yt += h * a[i - 1][k] * yh[k];
	  yh[i] = f(x0 + c[i] * h, yt, s);
	}
      }
      catch (...) {
        h /= hmax;			// try again with a smaller step
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = allow_up = false;
	continue;
      }
      for (k = 0, y4 = y; k <= 5; k++) y4 += h * b4[k] * yh[k]; 
      for (k = 0, y5 = y; k <= 5; k++) y5 += h * b5[k] * yh[k]; 
      crit = fabs(y4 - y5) / ((fabs(y5) + fabs(h * yh[5]) + eps) * eps);
      if (crit > 1.0) {			// step size adjustment
	h *= fmax(hfact * pow(crit, -pdown), 1.0 / hmax);
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = allow_up = false;	// forbid h increase (danger of loop!)
      }
      else if (crit < thr && !at_max && allow_up)
	h *= hfact * pow(crit, -pup);
      else
	break;
    }
    x0 += h;
    y = y5;
  }
  return y;
}

/*************************************************************18.02.2016*/
/* alternative version where y and f() = dy/dx are vectors		*/
/************************************************************************/
Vector<double> ode_ck(function<Vector<double>(double,
const Vector<double>&, void *)> f, void *s, double x0,
const Vector<double>& y0, double xend, double dxmin) {
  bool at_max, allow_up;
  int i, k,
    n = y0.size();
  double h, crit;
  Vector<double> y(n), yt(n), y4(n), y5(n);
  Matrix<double> yh(6, n);

  h = xend - x0;			// largest possible step size
  y = y0;
  at_max = true;
  while (h * (x0 - xend) < 0.0) {
    allow_up = true;
    while (true) {			// repeat until precision OK
      if ((x0 + h - xend) * h > 0.0) {	// make sure not to exceed xend
	h = xend - x0;
	at_max = true;
      }
      yh[0] = f(x0, y, s);		// calculate slopes at all nodes
      try {
	for (i = 1; i <= 5; i++) {
	  yt = y;
	  for (k = 0; k < i; k++) yt += (h * a[i - 1][k]) * yh[k];
	  yh[i] = f(x0 + c[i] * h, yt, s);
	}
      }
      catch (...) {
        h /= hmax;			// try again with a smaller step
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = allow_up = false;	// and forbid to increase it
	continue;
      }
      for (k = 0, y4 = y; k <= 5; k++) y4 += (h * b4[k]) * yh[k]; 
      for (k = 0, y5 = y; k <= 5; k++) y5 += (h * b5[k]) * yh[k]; 
      for (i = 0, crit = -1.0; i < n; i++) crit = fmax(crit,
      fabs(y4[i] - y5[i]) / ((fabs(y5[i]) + fabs(h * yh[5][i]) + eps)
      * eps));
      if (crit > 1.0) {			// step size adjustment
	h *= fmax(hfact * pow(crit, -pdown), 1.0 / hmax);
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = allow_up = false;	// forbid h increase (danger of loop!)
      }
      else if (crit < thr && !at_max && allow_up)
	h *= fmin(hfact * pow(crit, -pup), hmax);
      else
	break;
    }
    x0 += h;
    y = y5;
  }
  return y;
}
