/*************************************************************08.03.2004*/
/* ode_runge_kutta: ordinary differential equation solver using the	*/
/*		    method of Runge and Kutta				*/
/*		    with automatic control of the step size		*/
/*									*/
/* G. Jordan-Engeln, F. Reutter, Formelsammlung zur Numerischen		*/
/* Mathematik, BI Hochschultaschenbcher Vol. 106, Bibliographisches	*/
/* Institut, Mannheim 1976.						*/
/*======================================================================*/
/* f(x, y, s)	(i)  function defining the ODE, i.e. dy/dx = f(x, y, s).
   s		(i)  structure containing pass-through arguments for f()
   x0, y0	(i)  initial values
   xend		(i)  end of integration interval
   dxmin	(i)  minimal (abs.) step width (protection against infinite
   		     loop at a pole of f(); dxmin <= 0 turns feature off)
   RESULT	(o)  function value at xend
   exceptions	     increment: internal step size less than dxmin	*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const double
  crit_hi = 0.2,			// Lipshitz criteria for step size
  crit_lo = 0.005;


double ode_runge_kutta(function<double(double, double, void *)> f,
void *s, double x0, double y0, double xend, double dxmin) {
  bool at_max, allow_up;
  double h, yh[4], crit;

  h = xend - x0;			// largest possible step size
  at_max = true;
  while (h * (x0 - xend) < 0.0) {
    allow_up = true;
    while (true) {
      if ((x0 + h - xend) * h > 0.0) {	// make sure not to exceed xend
	h = xend - x0;
	at_max = true;
      }
      yh[0] = f(x0, y0, s) * h;
      try {
	yh[1] = f(x0 + 0.5 * h, y0 + 0.5 * yh[0], s) * h;
	yh[2] = f(x0 + 0.5 * h, y0 + 0.5 * yh[1], s) * h;
	yh[3] = f(x0 + h, y0 + yh[2], s) * h;
      }
      catch (...) {
        h *= 0.5;
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = allow_up = false;
	continue;
      }
      crit = 2.0 * fabs((yh[1] - yh[2]) / (yh[0] - yh[1]));
      if (crit > crit_hi) {		// step size adjustment
	h *= 0.5;
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = false;
      }
      else if (crit < crit_lo && !at_max && allow_up)
	h *= 2.0;
      else
	break;
    }
    x0 += h;
    y0 += (yh[0] + 2.0 * yh[1] + 2.0 * yh[2] + yh[3]) / 6.0;
  }
  return y0;
}

/*************************************************************14.02.2016*/
/* alternative version where y and dy/dx are vectors			*/
/************************************************************************/
Vector<double> ode_runge_kutta(function<Vector<double>(double,
const Vector<double>&, void *)> f, void *s, double x0,
const Vector<double>& y0, double xend, double dxmin) {
  bool at_max, allow_up;
  int i,
    n = y0.size();
  double h, crit_max, crit_min, c;
  Vector<double> y = y0;
  Matrix<double> yh(4, n);

  h = xend - x0;			// largest possible step size
  at_max = true;
  while (h * (x0 - xend) < 0.0) {
    allow_up = true;
    while (true) {
      if ((x0 + h - xend) * h > 0.0) {	// make sure not to exceed xend
	h = xend - x0;
	at_max = true;
      }
      yh[0] = f(x0, y0, s) * h;
      try {
	yh[1] = f(x0 + 0.5 * h, y0 + 0.5 * yh[0], s) * h;
	yh[2] = f(x0 + 0.5 * h, y0 + 0.5 * yh[1], s) * h;
	yh[3] = f(x0 + h, y0 + yh[2], s) * h;
      }
      catch (...) {
        h *= 0.5;
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = allow_up = false;
	continue;
      }
      for (i = 0, crit_max = -1.0, crit_min = mth_high; i < n; i++) {
	c = 2.0 * fabs((yh[1][i] - yh[2][i]) / (yh[0][i] - yh[1][i]));
	crit_max = fmax(crit_max, c);
	crit_min = fmin(crit_min, c);
      }
      if (crit_min > crit_hi) {		// step size adjustment
	h *= 0.5;
	if (fabs(h) < dxmin) throw MthException(MthException::increment);
	at_max = false;
      }
      else if (crit_max < crit_lo && !at_max && allow_up)
	h *= 2.0;
      else
	break;
    }
    x0 += h;
    y += (yh[0] + 2.0 * yh[1] + 2.0 * yh[2] + yh[3]) / 6.0;
  }
  return y;
}
