/*************************************************************14.08.1998*/
/* ran_ecuyer: random number generator using l'Ecuyer's method		*/
/*									*/
/* This routine uses a double congruential algorithm described by	*/
/* l'Ecuyer. In order to destroy sequential correlation, the shuffling	*/
/* table method is used in addition.					*/
/* Ref.: W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,	*/
/*	 "Numerical Recipes", Cambridge University Press, Cambridge	*/
/*	 1986, section 7.1						*/
/*	 P. l'Ecuyer, ACM Comm. 31, 742 (1988)				*/
/************************************************************************/
#include <mathC.h>

using namespace std;


const int
  m1 = 2147483563,			// constants of 1st generator
  a1 = 40014,				// note: m1 = a1*q1 + r1
  q1 = 53668,
  r1 = 12211,
  m2 = 2147483399,			// constants of 2nd generator
  a2 = 40692,
  q2 = 52774,
  r2 = 3791,
  ntab = 32,				// shuffling table size
  ndiv = 1 + (m1 - 1) / ntab;


double ran_ecuyer() noexcept {
  int j, k;
  static int irand, i1, i2, v[ntab], linit = 0;

  if (linit == 0) {			// initialize subroutine
    linit = 1;
    i1 = ran_seed();
    for (j = 0; j < ntab; j++) {
      k = i1 / q1;
      i1 = a1 * (i1 - k * q1) - k * r1;
      if (i1 < 0) i1 += m1;
      v[j] = i1;
    }
    irand = v[ntab - 1];
  }

  k = i1 / q1;
  i1 = a1 * (i1 - k * q1) - k * r1;
  if (i1 < 0) i1 += m1;			// 1st random number
  k = i2 / q2;
  i2 = a2 * (i2 - k * q2) - k * r2;
  if (i2 < 0) i2 += m2;			// 2nd random number

  /* calculation of a random number without sequential correlation:
     use previous random number to select a table entry
  */
  k = irand / ndiv;			// table index
  irand = v[k] - i2;
  v[k] = i1;
  if (irand < 1) irand += m1 - 1;
  return static_cast<double>(irand) / static_cast<double>(m1);
}
