/*************************************************************31.08.2001*/
/* diff_2all: numerical differentiation of a 2-dimensional function - 	*/
/*	      all derivatives up to a specified order			*/
/*									*/
/* Derivatives are calculated by difference expressions using 0-2 steps	*/
/* of extrapolation by Ridder's method.					*/
/*======================================================================*/
/* f(x, y, s)	(i)  function to be differentiated
   x, y		(i)  at this location
   s		(i)  structure passed to the function
   n		(i)  order of differentiation (0 <= n <= 4)
   r		(i)  number of extrapolation steps (0 <= r <= 2)
   dx, dy	(i)  (minimal) differentiation increments
   deriv	(o)  result array (length 1, 3, 6, or 10):
		     f0x0y, f1x0y, f0x1y, f2x0y, f1x1y, f0x2y, ...
   RESULT	(o)  error code:
		       MTH_EXCEED: n or r value not available		*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

#define MAX_N		4		/* max. order of differentiation */
#define MAX_R		2		/* max. number of extrapolations */
#define NR		((MAX_N + 1)/2)
#define NRT		(2*NR + 1)
#define Q		1.6		/* ratio for expolation scheme */


/* function prototypes */
static void diff_core(double (*f)(), double x, double y, void *s, int n,
double dx, double dy, double *deriv);


int diff_2all(double (*f)(), double x, double y, void *s, int n, int r,
double dx, double dy, double *deriv) {
  int i, j, k;
  double q2, d[MAX_R + 1][((MAX_N + 2) * (MAX_N + 1)) / 2];

  if (n < 0 || n > MAX_N || r < 0 || r > MAX_R) return (MTH_EXCEED);

  for (k = 0; k <= r; k++) {
    diff_core(f, x, y, s, n, dx, dy, &d[k][0]);
    dx *= Q;
    dy *= Q;
  }
  for (i = 0; i < ((n + 2) * (n + 1))/2; i++) {
    for (j = r, q2 = SQ(Q); j > 0; j--, q2 *= SQ(Q)) {
      for (k = 0; k < j; k++) d[k][i] = (q2 * d[k][i] - d[k + 1][i])
      / (q2 - 1.0);
    }
    *(deriv++) = d[0][i];
  }
  return (MTH_GOOD); 
}

/*************************************************************30.08.2001*/
/* diff_core: numerical differentiation of a 2-dimensional function -	*/
/*	      all derivatives up to a specified order			*/
/*======================================================================*/
/* f(x, y, s)	(i)  function to be differentiated
   x, y		(i)  at this location
   s		(i)  structure passed to the function
   n		(i)  order of differentiation (0 <= n <= 4)
   dx, dy	(i)  (minimal) differentiation increments
   deriv	(o)  result array (length 1, 3, 6, or 10)		*/
/************************************************************************/
static void diff_core(double (*f)(), double x, double y, void *s, int n,
double dx, double dy, double *deriv) {
  int i, j, jmax;
  double dx2, dy2, *(p[NRT]), amat[NRT * NRT];
  dmatrix a;

  dmatrix_overlay(&a, p, amat, -NR, NR, -NR, NR);

  *(deriv++) = a.a[0][0] = (*f)(x, y, s);/* 0th order */
  if (n == 0) return;

  for (i = -1; i <= 1; i++) {		/* 1st order */
    jmax = 1 - abs(i);
    for (j = -jmax; j <= jmax; j += 2)
      a.a[i][j] = (*f)(x + i * dx, y + j * dy, s);
  }
  *(deriv++) = (a.a[1][0] - a.a[-1][0]) / (2 * dx);		/* 1x0y */
  *(deriv++) = (a.a[0][1] - a.a[0][-1]) / (2 * dy);		/* 0x1y */
  if (n == 1) return;

  for (i = -1; i <= 1; i += 2) {	/* 2nd order*/
    for (j = -1; j <= 1; j += 2) a.a[i][j] = (*f)(x + i * dx, y + j * dy, s);
  }
  dx2 = SQ(dx);
  dy2 = SQ(dy);
  *(deriv++) = (a.a[1][0] - 2 * a.a[0][0] + a.a[-1][0]) / dx2;	/* 2x0y */
  *(deriv++) = (a.a[1][1] - a.a[1][-1] - a.a[-1][1] + a.a[-1][-1])
  / (4 * dx * dy);						/* 1x1y */
  *(deriv++) = (a.a[0][1] - 2 * a.a[0][0] + a.a[0][-1]) / dy2;	/* 0x2y */
  if (n == 2) return;

  for (i = -2; i <= 2; i += 2) {	/* 3rd order */
    jmax = 2 - abs(i);
    for (j = -jmax; j <= jmax; j += 4)
      a.a[i][j] = (*f)(x + i * dx, y + j * dy, s);
  }
  *(deriv++) = (a.a[2][0] - 2 * a.a[1][0] + 2 * a.a[-1][0] - a.a[-2][0])
  / (2 * CB(dx));						/* 3x0y */
  *(deriv++) = (a.a[-1][1] - 2 * a.a[0][1] + a.a[1][1] - a.a[-1][-1]
  + 2 * a.a[0][-1] - a.a[1][-1]) / (2 * dx2 * dy);		/* 2x1y */
  *(deriv++) = (a.a[1][-1] - 2 * a.a[1][0] + a.a[1][1] - a.a[-1][-1]
  + 2 * a.a[-1][0] - a.a[-1][1]) / (2 * dx * dy2);		/* 1x2y */
  *(deriv++) = (a.a[0][2] - 2 * a.a[0][1] + 2 * a.a[0][-1] - a.a[0][-2])
  / (2 * CB(dy));						/* 0x3y */
  if (n == 3) return;

  for (i = -2; i <= 2; i += 1) {	/* 4th order */
    if (i == 0) continue;
    j = 3 - abs(i);
    a.a[i][j] = (*f)(x + i * dx, y + j * dy, s);
    a.a[i][-j] = (*f)(x + i * dx, y - j * dy, s);
  }
  *(deriv++) = (a.a[-2][0] - 4 * a.a[-1][0] + 6 * a.a[0][0] - 4 * a.a[1][0]
  + a.a[2][0]) / SQ(dx2);					/* 4x0y */
  *(deriv++) = (a.a[-2][1] - 2 * a.a[-1][1] + 2 * a.a[1][1] - a.a[2][1]
  - a.a[-2][-1] + 2 * a.a[-1][-1] - 2 * a.a[1][-1] + a.a[2][-1])
  / (-4 * CB(dx) * dy);						/* 3x1y */
  *(deriv++) = (a.a[-1][-1] - 2 * a.a[0][-1] + a.a[1][-1] - 2 * a.a[-1][0]
  + 4 * a.a[0][0] - 2 * a.a[1][0] + a.a[-1][1] - 2 * a.a[0][1] + a.a[1][1])
  / (dx2 * dy2);						/* 2x2y */
  *(deriv++) = (a.a[1][-2] - 2 * a.a[1][-1] + 2 * a.a[1][1] - a.a[1][2]
  - a.a[-1][-2] + 2 * a.a[-1][-1] - 2 * a.a[-1][1] + a.a[-1][2])
  / (-4 * dx * CB(dy));						/* 1x3y */
  *(deriv++) = (a.a[0][-2] - 4 * a.a[0][-1] + 6 * a.a[0][0] - 4 * a.a[0][1]
  + a.a[0][2]) / SQ(dy2);					/* 0x4y */
}
