/*************************************************************13.08.1995*/
/* diff_ridders: numerical differentiation of a 1-dimensional function	*/
/*		 using the algorithm of Ridders				*/
/* Ref.: W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,	*/
/* Numerical Recipes, 2nd Edition, p. 180ff, Cambridge University	*/
/* Press, 1992.								*/
/*======================================================================*/
/* f(x, aux)	(i)  function to be differentiated
   aux		(i)  structure for passing additional information
   n	 	(i)  order of derivative (0 <= n <= nmax)
   x		(i)  argument of f
   h		(i)  initial step size of differentiation
		     (a range over which f changes SIGNIFICANTLY,
		     e.g. for exp(x) at x=1: h = 0.5)
   fd		(o)  nth derivative of f at location x
   err		(o)  estimate of its error
   RESULT	(o)  error code:
		       MTH_INCR = initial step size too small
		       MTH_EXCEED = n too high (edit fdiff1!)		*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

#define	CON		1.4		/* interval contraction factor */
#define CON2		CON * CON
#define SAFE		2.0		/* termination safety factor */
#define THR		1.E-10		/* lower increment threshold */
#define NTAB		10		/* size of interpolation table */
#define NDMX		4		/* max. differentiation order */

double fdiff1(double (*f)(), void *aux, int n, double x, double h);

int diff_ridders(double (*f)(), void *aux, int n, double x, double h,
double *fd, double *err) {
  int i, j;
  double errt, q, e1, e2, a[NTAB][NTAB];

  if (n < 0 || n > NDMX) return (MTH_EXCEED);/* diff. recipe available? */
  if (fabs(h) < THR) return (MTH_INCR);	/* increment OK? */

  if (n == 0) {				/* trivial: return function value*/
    *fd = (*f)(x);
    *err = 0.0;
    return (MTH_GOOD);
  }

  a[0][0] = fdiff1(f, aux, n, x, h);	/* calculate initial estimate */
  *err = DHIGH;

  for (i = 1; i < NTAB; i++) {
    h /= CON;				/* reduce increment size */
    a[i][0] = fdiff1(f, aux, n, x, h);	/* calculate new derivative */
    for (j = 1, q = CON2; j <= i; j++) {/* extrapolate */
      a[i][j] = (a[i][j - 1] * q - a[i - 1][j - 1]) / (q - 1.0);
      q *= CON2;
      e1 = fabs(a[i][j] - a[i][j - 1]);
      e2 = fabs(a[i][j] - a[i - 1][j - 1]);
      errt = (e1 > e2) ? e1 : e2;
      if (errt <= *err) {		/* error decreased? */
	*err = errt;			/* yes: save last result */
	*fd = a[i][j];
      }
    }
    if (fabs(a[i][i] - a[i - 1][i - 1]) >= SAFE * *err) break;
  }
  return (MTH_GOOD);
}

/*************************************************************13.08.1995*/
/* fdiff1: auxiliary function; recipes for numerical differentiation	*/
/*======================================================================*/
/* f(x, aux)	(i)  function to be differentiated
   aux		(i)  structure for passing additional information
   n		(i)  order of derivative (1 <= n <= NDMX)
   x		(i)  argument of f
   h		(i)  increment of differentiation
   RESULT	(o)  raw estimate of nth derivative at location x	*/
/************************************************************************/
double fdiff1(double (*f)(), void *aux, int n, double x, double h) {
  switch (n) {
  case 1:
    return (((*f)(x + h, aux) - (*f)(x - h, aux)) / (2 * h));
    break;
  case 2:
    return (((*f)(x + 2 * h, aux) - 2 * (*f)(x, aux) + (*f)(x - 2 * h, aux))
    / SQ(2 * h));
    break;
  case 3:
    return (((*f)(x + 3 * h, aux) - 3 * (*f)(x + h, aux)
    + 3 * (*f)(x - h, aux) - (*f)(x - 3 * h, aux)) / CB(2 * h));
    break;
  case 4:
    return (((*f)(x + 4 * h, aux) - 4 * (*f)(x + 2 * h, aux)
    + 6 * (*f)(x, aux) - 4 * (*f)(x - 2 * h, aux) + (*f)(x - 4 * h, aux))
    / SQ(SQ(2 * h)));
    break;
  default:
    return (0.0);
  }
}
