/*************************************************************24.01.2003*/
/* diff_romberg: numerical differentiation of a 1-dimensional function	*/
/*		 for functions with a limited x range			*/
/*									*/
/* symmetric case:  extrapolation scheme of Romberg			*/
/*		    G. Jordan-Engeln, F. Reutter: Numerische Mathematik	*/
/*		    fr Ingenieure, BI Hochschultaschenbcher Bd. 104,	*/
/*		    Bibliographisches Institut, Mannheim 1973, pp. 242ff*/
/* asymmetric case: differentiation of Lagrange interpolation polynomial*/
/*		    ibid., pp. 198ff					*/
/*======================================================================*/
/* f(x, aux)	(i)  function to be differentiated
   aux		(i)  structure for passing additional information
   n	 	(i)  order of derivative (0 <= n <= NDMX)
   x		(i)  argument of f
   h		(i)  initial step size of differentiation
		     (subsequent steps use larger increments)
   maxord	(i)  max. order of extrapolation (maxord < NTAB,
		     internally MIN(maxord, NTAB-1) is used)
   xmin, xmax	(i)  boundaries of x
   fd		(o)  nth derivative of f at location x
   RESULT	(o)  error code:
		       MTH_LIB = error in Lagrange interpolation
		       MTH_EXCEED = n out of range (edit fdiff1!)
		       MTH_INCR = initial step size too small
		       MTH_RANGE = x range too narrow or xmax < xmin	*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

#define	CON		1.4142135	/* interval contraction factor */
#define THR		1.E-10		/* lower increment threshold */
#define NTAB		10		/* size of interpolation table */
#define NDMX		4		/* max. differentiation order */
#define XDIFF		1.0E-3
#define FREE		0
#define BOUNDED		1


static double fdiff1(double (*f)(), void *aux, int n, double x, double h);


int diff_romberg(double (*f)(), void *aux, int n, double x, double h,
int maxord, double xmin, double xmax, double *fd) {
  int i, j, ord, mode;
  double q, a[NTAB], b[NTAB], c[NTAB];

  if (n < 0 || n > NDMX) return (MTH_EXCEED);/* diff. recipe available? */
  if (h < THR * fabs(x)) return (MTH_INCR);/* increment OK? */

  if (n == 0) {				/* trivial: return function value*/
    *fd = (*f)(x, aux);
    return (MTH_GOOD);
  }

  q = pow(CON, maxord) * h;		/* x range for symmetric formula */
  mode = (x > xmin && x < xmax) ? FREE : BOUNDED;
  if (x - q < xmin) {			/* asymmetric (left bounded) */
    ord = MIN(2 * (maxord + 1), NTAB - 1);
    a[0] = 0.0;
    a[1] = h;
    for (i = 2; i <= ord; i++) {
      if (mode == FREE) {
	a[i] = (a[i - 1] > 0.0) ? -a[i - 1] : -CON * a[i - 1];
	if (x + a[i] < xmin + XDIFF * h) {
	  a[i] = xmin - x;
	  mode = BOUNDED;
	}
      }
      else 
	a[i] = a[i - 1] * ((a[i - 1] < 0.0) ? -CON : CON);
    }
    if (x + a[ord] > xmax) return (MTH_RANGE);
    for (i = 0; i <= ord; i++) b[i] = (*f)(x + a[i], aux);
    if ((i = ipl_lagrange(ord + 1, a, b, c)) != 0) return (MTH_LIB(i));
    for (i = n, *fd = c[n]; i > 1; i--) *fd *= i;
  }
  else if (x + q <= xmax) {		/* symmetric (Romberg) method */
    ord = MIN(maxord, NTAB - 1);
    a[0] = fdiff1(f, aux, n, x, h);	/* calculate initial estimate */
    for (i = 1; i <= ord; i++) {
      h *= CON;
      b[0] = fdiff1(f, aux, n, x, h);
      for (j = 1, q = 1.0 / SQ(CON); j <= i; j++) {
	b[j] = (a[j - 1] - q * b[j - 1]) / (1.0 - q);
	q /= SQ(CON);
      }
      for (j = 0; j <= i; j++) a[j] = b[j];
    }
    *fd = b[ord];
  }
  else {				/* asymmetric (right bounded) */
    ord = MIN(2 * (maxord + 1), NTAB - 1);
    a[0] = -h;
    for (i = 1; i <= ord; i++) {
      if (mode == FREE) {
	a[i] = (a[i - 1] < 0.0) ? -a[i - 1] : -CON * a[i - 1];
	if (x + a[i] > xmax - XDIFF * h) {
	  a[i] = xmax - x;
	  mode = BOUNDED;
	}
      }
      else 
	a[i] = a[i - 1] * ((a[i - 1] > 0.0) ? -CON : CON);
    }
    if (x + a[ord] < xmin) return (MTH_RANGE);
    for (i = 0; i <= ord; i++) b[i] = (*f)(x + a[i], aux);
    if ((i = ipl_lagrange(ord + 1, a, b, c)) != 0) return (MTH_LIB(i));
    for (i = n, *fd = c[n]; i > 1; i--) *fd *= i;
    *fd = n * c[n];
  }
  return (MTH_GOOD);
}

/*************************************************************13.08.1995*/
/* fdiff1: auxiliary function; recipes for numerical differentiation	*/
/*======================================================================*/
/* f(x, aux)	(i)  function to be differentiated
   aux		(i)  structure for passing additional information
   n		(i)  order of derivative (1 <= n <= NDMX)
   x		(i)  argument of f
   h		(i)  increment of differentiation
   RESULT	(o)  raw estimate of nth derivative at location x	*/
/************************************************************************/
static double fdiff1(double (*f)(), void *aux, int n, double x, double h) {
  switch(n) {
  case 1:
    return (((*f)(x + h, aux) - (*f)(x - h, aux)) / (2 * h));
    break;
  case 2:
    return (((*f)(x + 2 * h, aux) - 2 * (*f)(x, aux) + (*f)(x - 2 * h, aux))
    / SQ(2 * h));
    break;
  case 3:
    return (((*f)(x + 3 * h, aux) - 3 * (*f)(x + h, aux)
    + 3 * (*f)(x - h, aux) - (*f)(x - 3 * h, aux)) / CB(2 * h));
    break;
  case 4:
    return (((*f)(x + 4 * h, aux) - 4 * (*f)(x + 2 * h, aux)
    + 6 * (*f)(x, aux) - 4 * (*f)(x - 2 * h, aux) + (*f)(x - 4 * h, aux))
    / SQ(SQ(2 * h)));
    break;
  default:
    return (0.0);
  }
}
