/*************************************************************27.06.2005*/
/* eqn_cubic: solution of a cubic equation				*/
/*									*/
/* default version: iterative search for 1st root, then deflation	*/
/*		    and direct solution of quadratic equation		*/
/* old version:	    Cardano's formula (activate by -DCARDANO)		*/
/*									*/
/* Equations of lesser degree are solved by the appropriate formulas.	*/
/* The solutions are arranged in ascending order.			*/
/*									*/
/* Note: The iterative algorithm is faster than Cardano's if the	*/
/*	 absolute coefficients of the reduced polynomial are not much	*/
/*	 larger than 1.							*/
/*======================================================================*/
/* a		(i)  array containing the polynomial coefficients
   x		(o)  result array
   RESULT	(o)  number of valid solutions				*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

#define PREC		1.0e-6

int eqn_cubic(const double *a, double *x) {
  int i, nreal;
  int ipwr, maxpwr;

#ifdef CARDANO				/* use Cardano's method */
  double w, h, p, q, dis, phi;
#else					/* use iterative method */
  int e, emin, emax;
  double w, h, y, y1, y2, dx, b[4], c[3], p, dis, f, sc;
#endif	/* CARDANO */

  if (a[3] == 0.0) return (eqn_quadratic(a, x));

#ifdef CARDANO
  w = a[2] / (3 * a[3]);
  p = CB(a[1] / (3 * a[3]) - SQ(w));
  q = -0.5 * (2 * CB(w) - (a[1] * w - a[0]) / a[3]);
  dis = SQ(q) + p;			/* discriminant */
  if (dis < 0.0) {			/* 3 real solutions */
    h = q / sqrt(-p);
    phi = acos(MAX(-1.0, MIN(1.0, h)));
    p = 2 * pow(-p, 1.0 / 6.0);
    for (i = 0; i < 3; i++) x[i] = p * cos((phi + 2 * i * M_PI) / 3.0) - w;
    nreal = 3;
  }
  else {				/* only one real solution */
    dis = sqrt(dis);
    x[0] = CBRT(q + dis) + CBRT(q - dis) - w;
    nreal = 1;
  }

  if (nreal == 3) {			/* sort results */
    if (x[1] < x[0]) SWAP(x[0], x[1], h);
    if (x[2] < x[1]) SWAP(x[1], x[2], h);
    if (x[1] < x[0]) SWAP(x[0], x[1], h);
  }

  /* Perform 1 Newton iteration step to minimize round-off errors */
  for (i = 0; i < nreal; i++, x++) {
    if ((h = a[1] + *x * (2 * a[2] + 3 * *x * a[3])) != 0.0)
    *x -= (a[0] + *x *(a[1] + *x *(a[2] + *x *a[3]))) / h;
  }

#else	/* CARDANO */
  b[3] = 1.0;
  for (i = 0, emin = emax = 0, h = 1.0 / a[3]; i <= 2; i++) {
    b[i] = a[i] * h;			/* reduce & scale polynomial */
    f = frexp(b[i], &e);
    e /= (3 - i);
    emax = MAX(emax, e);
    emin = MIN(emin, e);
  }
  sc = (emax > -emin) ? ldexp(1.0, -emax) : ldexp(1.0, 1 - emin);
  for (i = 2, h = 1.0, w = 0.0; i >= 0; i--) {
    h *= sc;
    b[i] *= h;
    w = MAX(w, fabs(b[i]));
  }

  w += 1.0;				/* radius of root circle */
  h = -1.0 / 3.0 * b[2];		/* inflection point */
  h = (b[0] + h * (b[1] + h * (b[2] + h)) > 0.0) ? -w : w;
  do {					/* find 1st root by Halley's method */
    y = b[2] + h;			/* evaluate function and derivatives*/
    y1 = 2 * y + h;
    y2 = y1 + 3 * h;
    y1 = h * y1 + b[1];
    y = (h * y + b[1]) * h + b[0];	/* ...looks odd, but saves CPU time */
    dx = y * y1 / (SQ(y1) - 0.5 * y * y2);
    h -= dx;
  } while (fabs(dx) > fabs(PREC * w));
  x[0] = h / sc;

  c[1] = h + b[2];			/* deflation; c[2] is 1 */
  c[0] = c[1] * h + b[1];
  p = 0.5 * c[1];			/* solve quadratic equation */
  dis = SQ(p) - c[0];
  if (dis >= 0.0) {
    dis = sqrt(dis);
    x[1] = (-p - dis) / sc;
    x[2] = (-p + dis) / sc;
    if (x[0] > x[1]) {			/* sort results */
      SWAP(x[0], x[1], h);
      if (x[1] > x[2]) SWAP(x[1], x[2], h);
    }
    nreal = 3;
  }
  else
    nreal = 1;
#endif	/* CARDANO */

  return (nreal);
}
