/*************************************************************05.03.2002*/
/* eqn_quartic: solution of a biquadratic equation			*/
/* Equations of lesser degree are solved by the appropriate formulas.	*/
/* The solutions are arranged in ascending order.			*/
/*									*/
/* Ref.: R. Zurmuehl, Praktische Mathematik fuer Ingenieure und		*/
/* Physiker, Springer, Berlin 1965, p. 60ff				*/
/*									*/
/* with coefficient scaling in order to minimize round-off errors	*/
/*======================================================================*/
/* a		(i)  array containing the polynomial coefficients
   x		(o)  result array
   RESULT	(o)  number of valid solutions				*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

int eqn_quartic(const double *a, double *x) {
  int i, nreal, e, emin, emax;
  double b[5], c[4], d0, d1, h, t, z, *px, f, sc;

  if (a[4] == 0.0) return (eqn_cubic(a, x));

  b[4] = 1.0;				// reduce and scale
  for (i = 0, emin = emax = 0; i < 4; i++) {
    b[i] = a[i] / a[4];
    f = frexp(b[i], &e);
    e /= (4 - i);			// faster than pwr(b[i], e/(4-i))
    emax = MAX(emax, e);
    emin = MIN(emin, e);
  }
  sc = (emax > -emin) ? ldexp(1.0, -emax) : ldexp(1.0, 1 - emin);
  for (i = 3, h = 1.0; i >= 0; i--) {
    h *= sc;
    b[i] *= h;
  }

  c[3] = 1.0;				/* cubic resolvent */
  c[2] = b[2];
  c[1] = b[1] * b[3] - 4 * b[0];
  c[0] = SQ(b[1]) + b[0] * (SQ(b[3]) - 4 * b[2]);
  i = eqn_cubic(c, x);
  z = x[i - 1];				/* use largest (positive!) root */

  nreal = 0;
  px = x;
  t = sqrt(0.25 * SQ(z) - b[0]);
  for (i = -1; i <= 1; i += 2) {
    d0 = -0.5 * z + i * t;		/* coeffs. of quadratic factor */
    d1 = (t != 0.0) ? i * 0.5 * (d0 * b[3] - b[1]) / t
    : 0.5 * b[3] + i * sqrt(0.25 * SQ(b[3]) - z - b[2]);
    h = 0.25 * SQ(d1) - d0;
    if (h >= 0.0) {
      h = sqrt(h);
      nreal += 2;
      *px++ = (-0.5 * d1 - h) / sc;
      *px++ = (-0.5 * d1 + h) / sc;
    }
  }

  for (i = 0; i < nreal; i++) {		/* single Newton step */
    d0 = a[0] + x[i] * (a[1] + x[i] * (a[2] + x[i] * (a[3] + x[i] * a[4])));
    d1 = a[1] + x[i] * (2 * a[2] + x[i] * (3 * a[3] + 4 * x[i] * a[4]));
    if (d1 != 0.0) x[i] -= d0 / d1;
  }

  if (nreal == 4) {			/* sort results */
    if (x[2] < x[0]) SWAP(x[0], x[2], h);
    if (x[3] < x[1]) SWAP(x[1], x[3], h);
    if (x[1] < x[0]) SWAP(x[0], x[1], h);
    if (x[3] < x[2]) SWAP(x[2], x[3], h);
    if (x[2] < x[1]) SWAP(x[1], x[2], h);
  }
  return (nreal);
}
