/************************************************************************/
/* Legendre polynomials							*/
/************************************************************************/

#include <math.h>
#include <mathc.h>


/*************************************************************24.08.2005*/
/* f_legendre: Legendre polynomial P_{l}(x)				*/
/*									*/
/* Legendre polynomials are orthogonal functions over [-1, +1]. 	*/
/* This function calculates the value of the polynomial. See		*/
/* f_legendre_c() if the coefficients are needed.			*/
/*									*/
/* note: Int (P_l(x))^2 dx = 2/(2l + 1)					*/
/*======================================================================*/
/* x		(i)  distance
   l		(i)  order (l >= 0)
   RESULT	(o)  P_l(x)						*/
/************************************************************************/
double f_legendre(double x, int l) {
  int k;
  double p0, p1, p2;

  p1 = 1.0;
  if (l == 0) return (p1);

  p2 = x;
  if (l == 1) return (p2);
  for (k = 2; k <= l; k++) {
    p0 = p1;
    p1 = p2;
    p2 = ((2 * k - 1) * x * p1 - (k - 1) * p0) / ((double) k);
  }
  return (p2);
}

/*************************************************************24.08.2005*/
/* f_legendre_c: Legendre polynomial P_{l}(x)				*/
/*									*/
/* This functions calculates the coefficents of Legendre polynomials.	*/
/* See f_legendre(), if only the function value is needed.		*/
/*======================================================================*/
/* l		(i)  order (l >= 0)
   c		(o)  array of coefficients (must have length l + 1)
   RESULT	(o)  error code: MTH_ILL = l negative
				 MTH_NOMEM = l too large for memory	*/
/************************************************************************/
int f_legendre_c(int l, double *c) {
  int i, k;
  dvector p0, p1, p2;

  if (l < 0) return (MTH_ILL);

  if (dvector_create(&p0, 0, l) || dvector_create(&p1, 0, l))
    return (MTH_NOMEM);

  dvector_overlay(&p2, c, 0, l);

  dvector_set(0.0, &p2);
  p2.a[0] = 1.0;			/* P_0(x) = 1 */
  if (l == 0) return (MTH_GOOD);

  dvector_copy(&p2, &p1);
  p2.a[0] = 0.0;
  p2.a[1] = 1.0;			/* P_1(x) = x */
  if (l == 1) return (MTH_GOOD);

  for (k = 2; k <= l; k++) {
    dvector_copy(&p1, &p0);
    dvector_copy(&p2, &p1);
    p2.a[0] = -((double) (k - 1)) / ((double) k) * p0.a[0];
    for (i = 1; i <= k; i ++) p2.a[i] = ((2 * k - 1) * p1.a[i - 1]
    - (k - 1) * p0.a[i]) / ((double) k);
  }
  dvector_free(&p0);
  dvector_free(&p1);
  return (MTH_GOOD);
}
