/*************************************************************19.12.2001*/
/* fit_marquardt: least squares approximation for systems of nonlinear	*/
/*		  equations by a modified Marquardt method		*/
/*======================================================================*/
/* f(x,y,aux)	(i)  subroutine containing the object functions: f
		     calculates the deviation vector y(n) for a given
		     parameter set x(m); an error code is returned.
   m		(i)  number of parameters to be optimized; if 0, only
		     ynorm is calculated.
   x[m]		(io) initial values/results of the parameters
   n		(i)  number of equations
   aux		(i)  structure containing auxiliary data for the
		     object function (only passed through)
   maxit	(i)  max. number of iteration cycles (recomm'd: 20*m)
   iprot	(i)  protocol control: 0 = no protocol
				       1 = print intermediate results
   ynorm	(o)  root of mean squared deviations
   RESULT	(o)  error code:
		       MTH_OBJ = object function error
		       MTH_LIB = matrix inverter: parameter #k irrelevant
		       MTH_NOCONV = too many cycles required
		       MTH_ILL = more unknowns than equations
		       MTH_NOMEM = memory allocation problem

  This subroutine uses dynamic memory allocations using malloc.

  xl is the Marquardt parameter; it is increased or decreased by
  xfact at every iteration step. If xl becomes less than xlcrit, it
  is set to zero. The Jacobian matrix is recomputed after m success-
  ful iteration steps, or after each unsuccessful step.			*/
/************************************************************************/
#include <math.h>
#include <stdio.h>
#include <mathc.h>

#define XINC		1.0E-5		/* differentiation increment */
#define EPS		1.0E-6		/* termination criterion */
#define EPSAUX		1.0E-4
#define XLINIT		0.1		/* Marquardt parameter initial value */
#define XLMAX		2000.0		/* Marquardt parameter max. value */
#define XLTHR		0.001		/* Marquardt parameter threshold */
#define FACTINIT	10.0		/* Marquardt parameter init. factor */

int fit_marquardt(int (*f)(), int m, double *x, int n, void *aux, int maxit,
int iprot, double *ynorm) {
  int i, j, k, age, inv_flag, success, err, dyn_flag;
  double xl, xlcrit, xfact, ynorm2, h, delta, gscal, grdnt, yscal;
  dvector x0, y0, y, yh, jacrow, grad, trace, dx, xnew;
  dmatrix jacobian, asq, ainv;

  dyn_flag = 0;
  if (m > n || n <= 0 ) return (MTH_ILL);/* more unknowns than equations?*/
				
  dvector_overlay(&x0, x, 0, m - 1);	/* create vector from x */
  if (dvector_create(&y0, 0, n - 1) != 0) return (MTH_NOMEM);
  dyn_flag++;				/* remember to deallocate y0 */

  if ((err = (*f)(x0.data, y0.data, aux)) != 0) {
    *ynorm = DHIGH;			/* trial evaluation of f() */
    err = MTH_OBJ(err);
    goto ex;
  }
  *ynorm = sqrt(dvector_dot(&y0, &y0) / n);
  if (m <= 0) goto ex;			/* only rms calculation? */

  dyn_flag++;				/* memory allocations */
  if (dvector_create(&trace, 0, m - 1) != 0
  || dvector_create(&grad, 0, m - 1) != 0
  || dvector_create(&dx, 0, m - 1) != 0
  || dvector_create(&xnew, 0, m - 1) != 0
  || dmatrix_create(&asq, 0, m - 1, 0, m - 1) != 0
  || dmatrix_create(&ainv, 0, m - 1, 0, m - 1) != 0
  || dmatrix_create(&jacobian, 0, m - 1, 0, n - 1) != 0
  || dvector_create(&y, 0, n - 1) != 0
  || dvector_create(&yh, 0, n - 1) != 0) {
    err = MTH_NOMEM;
    goto ex;
  }

  xl = XLINIT;				/* set control parameters */
  xfact = FACTINIT;
  xlcrit = XLTHR;
  age = m;
  dvector_set(0.0, &trace);

  for (k = 0; k < maxit; k++) {		/* main iteration loop */
    if (iprot == 1) {
      printf("\nrms: %e\n", *ynorm);
      for (i = 0; i < m; i++) printf("%16.7e", x0.a[i]);
      printf("\n");
    }
    if (age >= m) {			/* calc. new Jacobian */
      age = 0;
      for (i = 0; i < m; i++) {		/* differentiate */
	h = *ynorm * sqrt(m * trace.a[i]);
	delta = MAX(XINC, h);		/* differentiation increment */
	delta = XINC * MAX(delta, fabs(x0.a[i]));
	h = x0.a[i];
	x0.a[i] = h + delta;
	if ((err = (*f)(x0.data, y.data, aux)) != 0) {
	  err = MTH_OBJ(err);
	  goto ex;
	}
	x0.a[i] = h - delta;
	if ((err = (*f)(x0.data, yh.data, aux)) != 0) {
	  err = MTH_OBJ(err);
	  goto ex;
	}
	x0.a[i] = h;
	dvector_overlay(&jacrow, jacobian.a[i], 0, n - 1);
	dvector_sub(&y, &yh, &jacrow);
	dvector_smul(&jacrow, 1.0 / (2 * delta), &jacrow);
      }

      dmatrix_mult(&jacobian, &jacobian, &asq);

      for (i = 0; i < m; i++) {
	if (asq.a[i][i] <= 0.0) {	/* parameter irrelevant? */
	  err = MTH_LIB(i);
	  goto ex;
	}
	trace.a[i] = 1.0 / sqrt(asq.a[i][i]);/* scale square matrix */
	for (j = 0; j < m; j++) asq.a[i][j] *= trace.a[i];
	for (j = 0; j < m; j++) asq.a[j][i] *= trace.a[i];
      }
      dmatrix_copy(&asq, &ainv);
      inv_flag = 1;
    }
    dmatrix_mulv(&jacobian, &y0, &grad);/* calc. actual gradient vector */
    dvector_mul(&grad, &trace, &grad);
    gscal = dvector_dot(&grad, &grad);
    grdnt = sqrt(gscal / m) / (XINC + *ynorm);

    do {				/* determine correction vector */
      do {
	if (xl > 2.0) {			/* by gradient method */
	  dvector_smul(&grad, 1.0 / (1.0 + xl), &dx);
	  err = MTH_GOOD;
	}
	else {				/* by Marquardt's method */
	  for (i = 0; i < m; i++) asq.a[i][i] = 1.0 + xl;
	  if ((err = linsys_gauss(&asq, &grad, &dx)) != MTH_GOOD)
	    xl = (xl < xlcrit) ? xlcrit : xl * xfact;
	}
      } while (err != MTH_GOOD);
      yscal = dvector_dot(&grad, &dx);
      if (iprot == 1) printf("lambda=%e   gradient=%e\n", xl, grdnt);
      if (xl > XLMAX) goto ex;		/* no gradient? finish!*/
      dvector_mul(&dx, &trace, &dx);
      dvector_sub(&x0, &dx, &xnew);	/* new parameter set - try it!*/

      if ((err = (*f)(xnew.data, y.data, aux)) != 0) {
	xl = (xl < xlcrit)? xlcrit : xl*xfact;	/* get out of forbidden */
	success = 0;				/* area */
      }
      else if ((ynorm2 = sqrt(dvector_dot(&y, &y) / n))
      > (*ynorm * (1.0 - EPS))) {		/* no improvement? */
	success = 0;
	if (xl < xlcrit) {		/* calc. new lambda threshold */
	  if (inv_flag != 0) inv_flag = dmatrix_invert(&ainv);
	  xlcrit = m/sqrt(dmatrix_dot(&ainv, &ainv));
	}
	xl = (xl < xlcrit) ? xlcrit : xl * xfact;/* increase lambda */
	age *= m;			/* Jacobian matrix too old? */
	if (age >= 1) break;
	if (grdnt < 1.0E-3 || xl > 20.0) {/* convergence? */
	  for (i = 0, success = 1; i < m; i++) {
	    if (fabs(dx.a[i]) / (EPSAUX + fabs(x0.a[i])) > EPS) success = 0;
	  }
	  if (success) goto ex;
	}
      }
      else {				/* successful step */
	success = 1;
	dvector_copy(&xnew, &x0);
	dvector_copy(&y, &y0);
	if (ynorm2 <= 0.0) goto ex;	/* exact solution? */
	xfact = 2.0 + n * fabs((SQ(ynorm2) - SQ(*ynorm)) / yscal);
	if (xfact < 2.0) xfact = 2.0;
	if (xfact > 10.0) xfact = 10.0;
	xl /= xfact;			/* decrease Marquardt parameter */
	if (xl > 1.0) xl = 1.0;
	if (xl < xlcrit) xl = 0.0;
	*ynorm = ynorm2;
	age++;
      }
    } while (! success);
  }
  err = MTH_NOCONV;			/* too many steps required */

  ex: if (dyn_flag >= 2) {		/* deallocate memory */
    dvector_free(&trace);
    dvector_free(&grad);
    dvector_free(&dx);
    dvector_free(&xnew);
    dmatrix_free(&asq);
    dmatrix_free(&ainv);
    dmatrix_free(&jacobian);
    dvector_free(&y);
    dvector_free(&yh);
  }
  if (dyn_flag >= 1) dvector_free(&y0);
  return (err);
}
