/*************************************************************28.12.2001*/
/* fit_powell: nonlinear parameter fitting by Powell's method 		*/
/*======================================================================*/
/* obj(x,y,aux)	(i)  object function, calculates residual vector y[n] for
		     parameter array x[m], returns error flag; y*y will
		     be minimized.
   m		(i)  number of parameters to be optimimized
   x[m]		(io) parameter vector
   n		(i)  number of equations
   aux		(i)  structure containing auxiliary data for the object
		     function (just passed through)
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
				    1 = print intermediate results
   rms		(o)  rms deviation at the minimum
   RESULT	(o)  error code:
		       MTH_OBJ = failure of object function
		       MTH_LIB = failure of intern. minimizer
		       MTH_NOMEM = memory allocation failure
		       MTH_ILL = more unknowns than equations

   Memory for the residuals is allocated by malloc.			*/
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mathc.h>


#define EPS		1.0E-7		/* desired rms precision */
#define MAXIT		100		/* max. bracketing search steps */
#define GOLD		1.618034	/* sectio aurea ratio */
#define CGOLD		(2.0-GOLD)
#define GLIMIT		100.0		/* limit for parabola extrapol. */


/* auxiliary structures */

struct auxdata {			/* for auxobj() */
  dvector *y;
  void *aux;
  int (*obj)();
};

struct auxdata1 {			/* for auxobj1() */
  dvector *x, *u, *xa;
  void *aux;
  int (*f)();
};


/* prototypes of auxiliary subroutines */

static int linmin(int (*obj)(), dvector *x, const dvector *u, 
void *aux, double *y);
static int auxobj(double *x, double *y, struct auxdata *com);
static int auxf(double q, double *y, struct auxdata1 *com);


int fit_powell(int (*obj)(), int m, double *x, int n, void *aux, int maxit,
int iprot, double *rms) {
  int err;
  dvector y;
  struct auxdata com;

  if (m > n) return (MTH_ILL);

  if (dvector_create(&y, 0, n - 1) != 0) return (MTH_NOMEM);

  com.y = &y;
  com.aux = aux;
  com.obj = obj;
  if ((err = min_powell(auxobj, m, x, &com, maxit, iprot, rms)) == 0)
  *rms = sqrt(*rms / n);
  dvector_free(&y);
  return (err);
}

/*************************************************************31.12.2001*/
/* min_powell: minimization of a multidimensional function by Powell's	*/
/*	       method (successive 1-D minimization with optimizing the	*/
/*	       search direction)					*/
/*	       adapted from "Numerical Recipes"				*/
/*======================================================================*/
/* f(x,y,aux)	(i)  object function, calculates scalar property y for
		     parameter array x[m], returns error flag; y will
		     be minimized.
   m		(i)  number of parameters to be optimimized
   x[m]		(io) parameter array
   aux		(i)  structure containing auxiliary data for the object
		     function (just passed through)
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
				    1 = print intermediate results
   y		(o)  f() value at the minimum
   RESULT	(o)  error code: see fit_powell()

   Memory for the residuals is allocated by malloc.			*/
/************************************************************************/
int min_powell(int (*f)(), int m, double *x, void *aux, int maxit,
int iprot, double *y) {
  int i, j, k, err, iopt;
  double y2, y1, y0, gain, h;
  dvector *u, uavg, xact, xinit, xextr;

  err = MTH_NOMEM;			/* memory allocations */
  if ((u = (dvector *) malloc((size_t) m * sizeof(dvector))) == NULL)
  goto ex;
  for (i = 0; i < m; i++) {
    if (dvector_create(&u[i], 0, m - 1) != 0) goto ex;
  }
  if (dvector_create(&xinit, 0, m - 1) != 0
  || dvector_create(&uavg, 0, m - 1) != 0
  || dvector_create(&xextr, 0, m - 1)) goto ex;
  dvector_overlay(&xact, x, 0, m - 1);

  for (i = 0; i < m; i++) {		/* set direction vectors */
    dvector_set(0.0, &u[i]);
    u[i].a[i] = 1.0;
  }

  if ((err = f(xact.data, &y2, aux)) != 0) {
    err = MTH_OBJ(err);			/* 1st evaluation of f() */
    goto ex;
  }

  for (k = 0; k < maxit; k++) {		/* main iteration loop */
    dvector_copy(&xact, &xinit);	/* save the initial parameters */
    y0 = y2;
    gain = 0.0;
    for (i = 0, iopt = 0; i < m; i++) {	/* minimize along each */
      y1 = y2;				/* direction successively */
      if ((err = linmin(f, &xact, &u[i], aux, &y2)) != 0) {
	err = MTH_LIB(err);
	goto ex;
      }
      if (iprot > 0) {
	printf("%d %d y = %e\n", k, i, y2);
	for (j = 0; j < m; j++) printf("%14.5e", xact.a[j]);
	printf("\n");
      }
      if (y1 - y2 > gain) {		/* Did this give the largest */
	gain = y1 - y2;			/* decrease in value? */
	iopt = i;
      }
    }

    if (y0 - y2 < EPS * 0.5 * fabs(y0 + y2)) {
      *y = y2;				/* worthwhile to continue? */
      goto ex;
    }

    dvector_sub(&xact, &xinit, &uavg);	/* average direction */
    dvector_add(&xact, &uavg, &xextr);	/* extrapolate along uavg */
    if ((err = f(xextr.data, &y1, aux)) != 0) y1 = DHIGH;
    if (iprot > 0) {
      printf("%d %d y = %e\n", k, m, y1);
      for (i = 0; i < m; i++) printf("%14.5e", xextr.a[i]);
      printf("\n");
    }
    if (y1 < y2) {			/* successful extrapolation */
      dvector_copy(&xextr, &xact);	/* use it as new starting point */
      y2 = y1;
      h = 2 * (y0 - 2 * y2 + y1) * SQ(y0 - y2 - gain) - gain * SQ(y0 - y1);
      if (h <= 0.0) {			/* minimize along average dir.*/
	if ((err = linmin(f, &xact, &uavg, aux, &y2)) != 0) {
	  err = MTH_LIB(err);
	  goto ex;
	}
	if (iprot > 0) {
	  printf("%d %d y = %e\n", k, m + 1, y2);
	  for (i = 0; i < m; i++) printf("%14.5e", xact.a[i]);
	  printf("\n");
	}
	dvector_copy(&uavg, &u[iopt]);
      }
    }
  }
  err = MTH_NOCONV;			/* too many iterations ...*/

ex: dvector_free(&xinit);
  dvector_free(&uavg);
  dvector_free(&xextr);
  for (i = 0; i < m; i++) dvector_free(&u[i]);
  free(u);
  return (err);
}

/*************************************************************31.12.2001*/
/* linmin: one-dimensional minimization of a multidimensional function	*/
/*======================================================================*/
/* f(x,y,aux)	(i)  object function
   x		(io) parameter vector
   u		(i)  direction vector
   aux		(i)  structure containing auxiliary data for the object
		     function (just passed through)
   y		(o)  object function value at the minimum
   RESULT	(o)  error code: MTH_OBJ = object function error
				 MTH_ILL = no local minimum
				 MTH_FAIL = improper direction vector	*/
/************************************************************************/
static int linmin(int (*f)(), dvector *x, const dvector *u, 
void *aux, double *y) {
  int i, k, err;
  double d, pxlim;
  point2 a, b, c, p;
  dvector xa;
  struct auxdata1 com;

  if (dvector_create(&xa, 0, x->n - 1) != 0) return (MTH_NOMEM);

  com.x = x;				/* pass-through data for auxf */
  com.xa = &xa;
  com.u = (dvector *) u;
  com.aux = aux;
  com.f = f;
  /* The problem is to find a scalar parameter q for which the function
     value at x + q*u is a minimum.
     The first step is the calculation of brackets for this minimum. */

  a.x = 0.0;				/* 1st guess */
  if ((err = auxf(a.x, &a.y, &com)) != 0) {
    err = MTH_OBJ(err);			/* give up, if even this fails */
    goto ex;
  }
  for (i = 0, b.x = 0.0; i < x->n; i++) {
    d = fabs(u->a[i])/(fabs(x->a[i]) + EPS);/* nonzero direction vector? */
    if (d > b.x) b.x = d;
  }
  if (b.x == 0.0) return (MTH_FAIL);
  b.x = EPS / b.x;			/* 2nd guess */
  /* safety catch in case f() cannot be evaluated */
  while (auxf(b.x, &b.y, &com) != 0) b.x = 0.5 * (a.x + b.x);
  if (b.y > a.y) SWAP(a, b, c); 	/* wrong direction? */
  c.x = b.x + GOLD * (b.x - a.x);	/* 3rd guess */
  while (auxf(c.x, &c.y, &com) != 0) c.x = 0.5 * (b.x + c.x);

  for (k = 0; k < MAXIT; k++) {		/* search for brackets */
    if (b.y < c.y) break;		/* ... succeeded! */
    err = min_parabola(a, b, c, &p.x);	/* otherwise extrapolate */
    pxlim = b.x + GLIMIT * (c.x - b.x);	/* limit of safe extrapolation */

    /* Now there are several possibilities to investigate: */
    if ((b.x - p.x) * (p.x - c.x) > 0.0) {/* p between b and c */
      if (auxf(p.x, &p.y, &com) == 0) {
	if (p.y < c.y) {		/* minimum detected? */
	  a = b;			/* ... between b and c */
	  b = p;
	  break;
	}
	else if (p.y > b.y) {
	  c = p;			/* ... between a and p */
	  break;
	}
      }
    }
    else if ((c.x - p.x) * (p.x - pxlim) > 0.0) {/* p between c and limit */
      if (auxf(p.x, &p.y, &com) == 0) {
	if (p.y < c.y) {
	  b = c;
	  c = p;
	}
      }
    }
    p.x = ((p.x - pxlim) * (pxlim - c.x) >= 0.0) ? pxlim
    : c.x + GOLD * (c.x - b.x);
    while (auxf(p.x, &p.y, &com) != 0) p.x = 0.5 * (p.x + c.x);
    a = b;				/* discard oldest point */
    b = c;
    c = p;
  }
  if (b.y >= a.y || b.y >= c.y) return (MTH_ILL);/* no valid minimum found */

  /* Now that it is certain that the points a, b, and c bracket a
     minimum, this can be located by Brent's method. */
  if ((err = min_brent(auxf, &com, a.x, b.x, c.x, DEPS * fabs(b.x), &p.x, y))
  != 0) goto ex;
  dvector_smul(u, p.x, &xa);
  dvector_add(x, &xa, x);

ex: dvector_free(&xa);
  return (err);
}

/*************************************************************27.12.2001*/
/* auxobj: auxiliary object function					*/
/*									*/
/* auxobj "translates" an object function of a curve-fitting problem,	*/
/* obj(x[m], y[n]), into a function for a minimization subroutine,	*/
/* f(x[m], sum(y^2)).							*/
/*======================================================================*/
/* x		(i)  parameter vector
   ysq		(o)  sum of squared residuals
   com		(i)  auxiliary data for obj()
   RESULT	(o)  error code of user-supplied object function	*/
/************************************************************************/
static int auxobj(double *x, double *ysq, struct auxdata *com) {
  int err;

  err = (*com->obj)(x, com->y->data, com->aux);
  if (err == 0) *ysq = dvector_dot(com->y, com->y);
  return (err);
}

/*************************************************************31.12.2001*/
/* auxf: auxiliary object function					*/
/*									*/
/* auxf "translates" a multi-dimensional minimization problem into a	*/
/* 1-dimensional minimization along a chosen direction.			*/
/*======================================================================*/
/* q		(i)  abscissa
   y		(o)  function value
   com		(i)  pass-through data
   RESULT	(o)  error code of user-supplied object function	*/
/************************************************************************/
static int auxf(double q, double *y, struct auxdata1 *com) {
  int err;

  dvector_smul(com->u, q, com->xa);
  dvector_add(com->x, com->xa, com->xa);
  err = (*com->f)(com->xa->data, y, com->aux);
  return (err);
}
