/*************************************************************12.11.2006*/
/* fit_simplex: multidimensional parameter fitting with the simplex	*/
/*		method of Nelder and Mead				*/
/*									*/
/* J.A. Nelder, R. Mead, The Computer Journal ?? (????) 308-313.	*/
/*									*/
/* W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,		*/
/* Numerical Recipes, Cambridge University Press, Cambridge 1986,	*/
/* section 7.1								*/
/*									*/
/* This algorithm finds minima where many other methods fail, but	*/
/* convergence is AWFULLY slow, of 1st order at best. It should never	*/
/* be used to determine final values of model parameters, but merely	*/
/* to generate initial values for the Marquardt-Levenberg or Powell	*/
/* methods								*/
/*======================================================================*/
/* f(x,y,aux)	(i)  object function, calculates residual vector y[n] for
		     parameter array x[m], returns error flag; y*y will
		     be minimized
   m		(i)  number of parameters to be optimized
   x[m]		(io) parameter vector
   dx[m]	(i)  initial step size of x
   n		(i)  number of equations
   aux		(i)  structure containing auxiliary data for f(), i.e.,
   		     data simply passed through
   maxit	(i)  max. number of iterations
   iprot	(i)  protocol flag: 0 = no protocol
   				    1 = print intermediate results
   rms		(o)  rms deviation at the minimum
   RESULT	(o)  error code:
   		       MTH_OBJ = failure of object function
		       MTH_NOMEM = failure of memory allocation
		       MTH_ILL = too many unknowns, or dx[i] = 0
		       MTH_NOCONV = no convergence iterations		*/
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mathc.h>

#define	ALPHA		1.0		/* reflection factor */
#define BETA		0.5		/* contraction factor */
#define GAMMA		2.0		/* expansion factor */
#define EPS		1.0E-02		/* relative accuracy */


int fit_simplex(int (*f)(), int m, double *x, double *dx, int n, void *aux,
int maxit, int iprot, double *rms) {
  int i, j, err, ilo, ihi, ihi2, k;
  double sh, shh, *s;
  dvector *u, uh, uhh, uc, y;

  if (m > n) return (MTH_ILL);
  for (i = 0; i < m; i++) {
    if (dx[i] == 0.0) return (MTH_ILL);
  }

  if ((s = (double *) malloc((m + 1) * sizeof(double))) == NULL
  || dvector_create(&y, 0, n - 1) != 0
  || dvector_create(&uh, 0, m - 1) != 0
  || dvector_create(&uhh, 0, m - 1) != 0
  || dvector_create(&uc, 0, m - 1) != 0
  || (u = (dvector *) malloc((m + 1) * sizeof(dvector))) == NULL)
  return (MTH_NOMEM);

  ilo = ihi = 0;
  for (i = 0; i <= m; i++) {		/* set up the 1st simplex */
    if (dvector_create(&u[i], 0, m - 1) != 0) return (MTH_NOMEM);
    for (j = 0; j < m; j++) u[i].a[j] = x[j];
    if (i < m) u[i].a[i] += dx[i];	/* u[m] contains original x vector */
    if ((err = f(u[i].data, y.data, aux)) != 0) goto ex;
    s[i] = dvector_dot(&y, &y);
    if (s[i] < s[ilo]) ilo = i;
    if (s[i] > s[ihi]) ihi = i;
  }
  for (i = 0, ihi2 = ilo; i <= m; i++) {
    if (i == ihi || i == ilo) continue;
    if (s[i] > s[ihi2]) ihi2 = i;
  }
      
  for (k = 1; k < maxit; k++) {		/* main iteration loop */
    if (iprot) {
      printf("%d smin = %e\n", k, s[ilo]);
      for (i = 0; i < m; i++) printf("%14.5e", u[ilo].a[i]);
      printf("\n");
    }

    err = MTH_FAIL;
    if (ilo == ihi) goto ex;		/* cannot distinguish best and worst */
    for (j = i = 0; j < m; j++) {	/* convergence? */
      if (fabs(u[ihi].a[j] - u[ilo].a[j]) > 0.5 * EPS
      * fabs(u[ihi].a[j] + u[ilo].a[j])) {
        i = 1;
	break;
      }
    }
    if (i == 0) {
      for (j = 0; j < m; j++) x[j] = u[ilo].a[j];
      *rms = sqrt(s[ilo] / n);
      err = MTH_GOOD;
      goto ex;
    }

    for (j = 0; j < m; j++) {		/* reflection away from u[ihi] */
      for (i = 0, uc.a[j] = 0.0; i <= m; i++) {
        if (i != ihi) uc.a[j] += u[i].a[j];
      }
      uc.a[j] /= m;			/* center of plane opposite u[ihi] */

      uh.a[j] = (1.0 + ALPHA) * uc.a[j] - ALPHA * u[ihi].a[j];
    }
    sh = (f(uh.data, y.data, aux) != 0) ? DHIGH : dvector_dot(&y, &y);

    if (sh < s[ilo]) {			/* successful move? */
      for (j = 0, uhh.a[j] = 0.0; j < m; j++) uhh.a[j] = (1.0 - GAMMA)
      * uc.a[j] + GAMMA * uh.a[j];	/* expansion step */
      shh = (f(uhh.data, y.data, aux) != 0) ? DHIGH : dvector_dot(&y, &y);
      if (shh < sh) {			/* 2nd move successful? */
        dvector_copy(&uhh, &u[ihi]);	/* ... yes, replace high point */
	s[ihi] = shh;
      }
      else {
        dvector_copy(&uh, &u[ihi]);	/* ... no, use uh as new high */
	s[ihi] = sh;
      }
      ilo = ihi;
      ihi = ihi2;
      for (i = 0, ihi2 = ilo; i <= m; i++) {
        if (i == ihi || i == ilo) continue;
        if (s[i] > s[ihi2]) ihi2 = i;
      }
    }
    else if (sh <= s[ihi2]) {		/* new point better than 2nd-highest? */
      dvector_copy(&uh, &u[ihi]);	/* ... yes, discard previous high */
      s[ihi] = sh;
      ihi = ihi2;
      for (i = 0, ihi2 = ilo; i <= m; i++) {
        if (i == ihi || i == ilo) continue;
        if (s[i] > s[ihi2]) ihi2 = i;
      }
    }
    else {
      if (sh <= s[ihi]) {		/* new point better than high one? */
	dvector_copy(&uh, &u[ihi]);	/* ... yes, discard previous high */
	s[ihi] = sh;
      }
      for (j = 0; j < m; j++) uhh.a[j] = BETA * u[ihi].a[j]
      + (1.0 - BETA) * uc.a[j];		/* contraction */
      shh = (f(uhh.data, y.data, aux) != 0) ? DHIGH : dvector_dot(&y, &y);
      if (shh <= s[ihi]) {		/* improvement? */
	dvector_copy(&uhh, &u[ihi]);	/* ... yes; use new point */
	s[ihi] = shh;
      }
      else {				/* if all fails */
        for (i = 0; i <= m; i++) {	/* contract towards low point */
	  if (i == ilo) continue;
	  for (j = 0; j < m; j++) u[i].a[j] = 0.5 * (u[i].a[j] + u[ilo].a[j]);
	  if ((err = f(u[i].data, y.data, aux)) != 0) goto ex;
	  s[i] = dvector_dot(&y, &y);
	}
      }
      for (i = 1, ilo = ihi = 0; i <= m; i++) {
        if (s[i] > s[ihi]) ihi = i;
        if (s[i] < s[ilo]) ilo = i;
      }
      for (i = 0, ihi2 = ilo; i <= m; i++) {
        if (i == ihi || i == ilo) continue;
        if (s[i] > s[ihi2]) ihi2 = i;
      }
    }
  }

  for (j = 0; j < m; j++) x[j] = u[ilo].a[j];
  *rms = sqrt(s[ilo] / n);
  err = MTH_NOCONV;

ex: free(s);
  dvector_free(&uh);
  dvector_free(&uhh);
  dvector_free(&uc);
  dvector_free(&y);
  for (j = 0; j <= m; j++) dvector_free(&u[j]);
  free(u);
  return (err);
}
