/*************************************************************14.08.1998*/
/* ipl_akima: interpolation by Akima's method				*/
/* Ref.: c't 1989, 206-214						*/
/*									*/
/* If x[i]<x<x[i+1], yinterpol = y[i] + b[i]*h + c[i]*h^2 + d[i]*h^3	*/
/* with h = x[i+1] - x[i].						*/
/* Akima calculates the polynomial coefficients. Use ipl_eval to obtain	*/
/* interpolated values.							*/
/*======================================================================*/
/* n		(i)  number of data points
   x[n], y[n]	(i)  abscissa and ordinate values
   b, c, d[n]	(o)  Akima polynomial coefficient arrays
   RESULT	(o)  error code: 0 = no error
		       MTH_ILL = too few data points (n < 3)
		       MTH_RANGE = abscissa values not monotonous	*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

static double akslope(double s1, double s2, double s3, double s4);

int ipl_akima(int n, double *x, double *y, double *b, double *c,
double *d) {
  int i, n1;
  double f, yl1, sl1, yl2, sl2, yr1, sr1, yr2;

  if (n < 3) return (MTH_ILL);		/* enough data points? */
  n1 = n - 1;

  for (i = 0; i < n1; i++) {		/* abscissae monotonous? */
    d[i] = x[i + 1] - x[i];
    if (d[i] <= 0.0) return (MTH_RANGE);
  }

  for (i = 0; i < n1; i++) c[i] = (y[i + 1] - y[i]) / d[i];
  yl1 = d[1] * (c[1] - 2 * c[0]) + y[0];/* extrapolate 2 values left */
  sl1 = (y[0] - yl1) / d[1];
  yl2 = d[0] * (c[0] - 2 * sl1) + yl1;
  sl2 = (yl1 - yl2) / d[0];

  yr1 = d[n1 - 2] * (2 * c[n1 - 1] - c[n1 - 2]) + y[n1];/* ...and right */
  c[n1] = (yr1 - y[n1]) / d[n1-2];
  yr2 = d[n1 - 1] * (2 * c[n1] - c[n1 - 1]) + yr1;
  sr1 = (yr2 - yr1) / d[n1 - 1];

  b[0] = akslope(sl2, sl1, c[0], c[1]);	/* calculate smoothed slopes */
  b[1] = akslope(sl1, c[0], c[1], c[2]);
  for (i = 2; i < n1; i++) b[i] = akslope(c[i - 2], c[i - 1], c[i], c[i + 1]);
  b[n1] = akslope(c[n1 - 2], c[n1 - 1], c[n1], sr1);

  for (i = 0; i < n1; i++) {		/* calc. the other coefficients */
    f = d[i];
    d[i] = (b[i + 1] + b[i] - 2 * c[i]) / SQ(f);
    c[i] = (3 * c[i] - 2 * b[i] - b[i + 1]) / f;
  }
  c[n1] = d[n1] = 0.0;
  return (MTH_GOOD);
}

/*************************************************************21.01.1993*/
/* akslope: Akima's slope formula					*/
/*======================================================================*/
/* s1, ..., s4	(i)  four slopes
   RESULT	(i)  averaged slope					*/
/************************************************************************/
static double akslope(double s1, double s2, double s3, double s4) {
  double sla, sra;

  sla = fabs(s2 - s1);
  sra = fabs(s4 - s3);
  if (sla == 0.0 && sra == 0.0)
    return (0.5 * (s2 + s3));
  else
    return ((sra * s2 + sla * s3) / (sla + sra));
}
