/*************************************************************17.04.2003*/
/* ipl_pspline: interpolation by a periodic cubic spline function:	*/
/* if x[i] < x < x[i+1], yspline = y[i] + b[i]*h + c[i]*h^2 + d[i]*h^3	*/
/* with h = x[i+1] - x[i].						*/
/*									*/
/* This function calculates the polynomial coefficients only. Use	*/
/* ipl_eval to generate interpolated values.				*/
/*									*/
/* If less than 3 data points are given, ipl_pspline sets the polynomial*/
/* cofficients to zero.							*/
/*======================================================================*/
/* n		(i)  number of data points
   x[n], y[n]	(i)  abscissa and ordinate values; the last point is
		     regarded as periodic image of the first one
   b, c, d[n]	(o)  spline coefficient arrays
   RESULT	(o)  error code:
		       MTH_RANGE = abscissa values not monotonous	*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

int ipl_pspline(int n, double *x, double *y, double *b, double *c,
double *d) {
  int i, i1, n1, n2;
  double f, h;

  if (n <= 2) {				/* enough data points? */
    for (i = 0; i < n; i++)
      b[i] = c[i] = d[i] = 0.0;		/* ... no, just clear coefficients */
  }
  else {				/* ... yes, calc. cubic spline */
    n1 = n - 1;
    n2 = n1 - 1;
    for (i = 0; i < n1; i++) {		/* abscissa values monotonous? */
      if ((c[i] = x[i+1] - x[i]) <= 0.0) return (MTH_RANGE);
    }
    c[n1] = c[0];

    /* The spline coefficients c are obtained from a system of linear
       equations. The matrix representing the lhs. is symmetric and
       cyclic tridiagonal (diagonal elements: b, off-diagonal elements: c).
       The rhs. is given by d. */

    for (i = 1; i < n1; i++) {		/* set up system of equations */
      i1 = i - 1;
      d[i] = 3 * ((y[i + 1] - y[i]) / c[i] - (y[i] - y[i1]) / c[i1]);
      b[i] = 2 * (c[i1] + c[i]);
    }
    d[n1] = 3 * ((y[1] - y[0]) / c[0] - (y[n1] - y[n2]) / c[n2]);
    b[n1] = 2 * (c[n2] + c[n1]);

    h = c[1];				/* simplified Gauss elimination */
    for (i = 2; i < n1; i++) {
      f = -h / b[i - 1];
      b[i] += f * h;
      d[i] += f * d[i - 1];
      h = c[i];
      if (i == n2)
	c[i] += f *c[i - 1];
      else
	c[i] = f * c[i - 1];
    }
    f = -h / b[n2];
    b[n1] += f * c[n2];
    d[n1] += f * d[n2];

    for (i = n1 - 2; i >= 1; i--) {
      f = -(x[i + 1] -  x[i]) / b[i + 1];
      c[i] += f * c[i + 1];
      d[i] += f * d[i + 1];
    }

    f = -c[0] / b[1];
    b[n1] += f * c[1];
    d[n1] += f * d[1];

    c[0] = c[n1] = d[n1] / b[n1];
    for (i = n2; i >= 1; i--) {
      c[i] = (d[i] - c[i] * c[n1]) / b[i];
      f = x[i + 1] - x[i];
      b[i] = (y[i + 1] - y[i]) / f - (c[i + 1] + 2 * c[i]) * f / 3.0;
      d[i] = (c[i + 1] - c[i]) / (3 * f);
    }
    f = x[1] - x[0];			/* deal with outer intervals */
    b[0] = b[n1] = (y[1] - y[0]) / f - (c[1] + 2 * c[0]) * f / 3.0;
    d[0] = d[n1] = (c[1] - c[0]) / (3 * f);
  }
  return (MTH_GOOD);
}
