/*************************************************************16.01.2000*/
/* ipl_spline: interpolation by a natural cubic spline function:	*/
/* if x[i] < x < x[i+1], yspline = y[i] + b[i]*h + c[i]*h^2 + d[i]*h^3	*/
/* with h = x[i+1] - x[i].						*/
/*									*/
/* This function calculates the polynomial coefficients only. Use	*/
/* ipl_eval to generate interpolated values.				*/
/*									*/
/* If less than 3 data points are given, ipl_spline calculates a linear	*/
/* interpolation or sets the polynomial coefficients to zero.		*/
/*======================================================================*/
/* n		(i)  number of data points
   x[n], y[n]	(i)  abscissa and ordinate values
   b, c, d[n]	(o)  spline coefficient arrays
   RESULT	(o)  error code:
		       MTH_RANGE = abscissa values not monotonous	*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

int ipl_spline(int n, double *x, double *y, double *b, double *c,
double *d) {
  int i, i1, n1, n2;
  double f;

  if (n <= 1)				/* enough data points? */
    b[0] = c[0] = d[0] = 0.0;		/* ... no, just clear coefficients */
  else if (n == 2) {			/* ... no, try linear interpolation */
    f = x[1] - x[0];
    if (f <= 0.0) return (MTH_RANGE);
    b[0] = b[1] = (y[1] - y[0]) / f;
    c[0] = c[1] = d[0] = d[1] = 0.0;
  }
  else {				/* ... yes, calc. cubic spline */
    n1 = n - 1;
    n2 = n1 - 1;
    for (i = 0; i < n1; i++) {		/* abscissa values monotonous? */
      if ((c[i] = x[i + 1] - x[i]) <= 0.0) return (MTH_RANGE);
    }
    c[n1] = 0.0;

    /* The spline coefficients c are obtained from a system of linear
       equations. The matrix representing the lhs. is symmetric and
       tridiagonal (diagonal elements: b, off-diagonal elements: c).
       The rhs. is given by d. */

    for (i = 1; i < n1; i++) {		/* set up matrix */
      i1 = i - 1;
      d[i] = 3 * ((y[i + 1] - y[i]) / c[i] - (y[i] - y[i1]) / c[i1]);
      b[i] = 2 * (c[i1] + c[i]);
    }

    for (i = 1; i < n2; i++) {		/* simplified Gauss elimination */
      f = -c[i] / b[i];
      b[i + 1] += f * c[i];
      d[i + 1] += f * d[i];
    }
    for (i = n2; i >= 1; i--) {
      f = c[i];
      c[i] = (d[i] - c[i + 1] * f) / b[i];
      b[i] = (y[i + 1] - y[i]) / f - (c[i + 1] + 2 * c[i]) * f / 3.0;
      d[i] = (c[i + 1] - c[i]) / (3 * f);
    }

    f = x[1] - x[0];			/* deal with 1st interval */
    b[0] = (y[1] - y[0]) / f - f * c[1] / 3.0;
    c[0] = 0.0;
    d[0] = c[1] / (3 * f);
    f = x[n1] - x[n2];			/* ... and with last one */
    b[n1] = b[n2] + f * (2 * c[n2] + 3 * f * d[n2]);
    d[n1] = 0.0;
  }
  return (MTH_GOOD);
}
