/*************************************************************08.05.2008*/
/* lin_jacobi: eigenvalues and -vectors of a symmetric matrix by	*/
/*	       Jacobi's method						*/
/*									*/
/* adapted from "Numerical Recipes"					*/
/*======================================================================*/
/* a		(i)  matrix for which eigenvalues are sought
   d		(o)  vector of eigenvalues
   v		(o)  matrix of which each column is an eigenvector
   RESULT	(o)  error code						*/
/************************************************************************/
#include <math.h>
#include <mathc.h>


#define NITER		50		/* number of sweeps */


int lin_jacobi(const dmatrix *a, dvector *d, dmatrix *v) {
  int i, j, k, l, n;
  double thresh, theta, tau, t, sum, s, c, g, h;
  dvector b, z;
  dmatrix q;

  n = a->nrow;

  if (dmatrix_create(&q, 0, n - 1, 0, n - 1) != 0
  || dvector_create(&b, 0, n - 1) != 0
  || dvector_create(&z, 0, n - 1) != 0) return MTH_NOMEM;

  dmatrix_copy(a, &q);
  dmatrix_set_u(v);
  for (i = 0; i < n; i++) b.a[i] = d->a[i] = q.a[i][i];

  for (k = 0; k < NITER; k++) {		/* main loop */
    dvector_set(0.0, &z);
    sum = 0.0;				/* sum of off-diagonal elements */
    for (i = 0; i < n - 1; i++) {
      for (j = i + 1; j < n; j++) sum += fabs(q.a[i][j]);
    }
    if (sum == 0.0) {			/* matrix diagonalized? */
      dvector_free(&b);
      dvector_free(&z);
      dmatrix_free(&q);
      return MTH_GOOD;
    }
    thresh = (k < 4) ? 0.2 * sum / SQ(n) : 0.0;
    for (i = 0; i < n - 1; i++) {
      for (j = i + 1; j < n; j++) {
        g = 100.0 * fabs(q.a[i][j]);
	/* skip rotation for small off-diagonal elements after 4 sweeps */
	if (k > 4 && fabs(d->a[i] + g) == fabs(d->a[i])
	&& fabs(d->a[j] + g) == fabs(d->a[j]))
	  q.a[i][j] = 0.0;
	else if (fabs(q.a[i][j]) > thresh) {
	  h = d->a[j] - d->a[i];
	  if (fabs(h) + g == fabs(h))
	    t = q.a[i][j] / h;
	  else {
	    theta = 0.5 * h / q.a[i][j];
	    t = 1.0 / (fabs(theta) + sqrt(1.0 + SQ(theta)));
	    if (theta < 0.0) t = -t;
	  }
          c = 1.0 / sqrt(1 + SQ(t));	/* cos */
	  s = t * c;			/* sin */
	  tau = s / (1.0 + c);
	  h = t * q.a[i][j];
	  z.a[i] -= h;
	  z.a[j] += h;
	  d->a[i] -= h;
	  d->a[j] += h;
	  q.a[i][j] = 0.0;
	  for (l = 0; l < i; l++) {	/* rotation for 0 <= l < i */
	    g = q.a[l][i];
	    h = q.a[l][j];
	    q.a[l][i] = g - s * (h + g * tau);
	    q.a[l][j] = h + s * (g - h * tau);
	  }
	  for (l = i + 1; l < j; l++) {	/* rotation for i < l < j */
	    g = q.a[i][l];
	    h = q.a[l][j];
	    q.a[i][l] = g - s * (h + g * tau);
	    q.a[l][j] = h + s * (g - h * tau);
	  }
	  for (l = j + 1; l < n; l++) {	/* rotation for j < l < n */
	    g = q.a[i][l];
	    h = q.a[j][l];
	    q.a[i][l] = g - s * (h + g * tau);
	    q.a[j][l] = h + s * (g - h * tau);
	  }
	  for (l = 0; l < n; l++) {	/* update eigenvectors */
	    g = v->a[l][i];
	    h = v->a[l][j];
	    v->a[l][i] = g - s * (h + g * tau);
	    v->a[l][j] = h + s * (g - h * tau);
	  }
	}
      }
    }
    dvector_add(&b, &z, &b);
    dvector_copy(&b, d);
  }
  dvector_free(&b);
  dvector_free(&z);
  dmatrix_free(&q);
  return MTH_NOCONV;
}
