/*************************************************************03.09.2001*/
/* linsys_gauss: solution of a system of linear equations by elimination*/
/*									*/
/* In each elimination cycle the absolutely largest column element is	*/
/* chosen as pivot. The input arrays are not changed.			*/
/*======================================================================*/
/* a		(i)  coefficient matrix
   b[n]		(i)  right hand side of the system of equations
   x[n]		(o)  result vector
   RESULT	(o)  error code:
		       MTH_FAIL = singular coefficient matrix
		       MTH_NOMEM = memory allocation failed
		       MTH_ILL = sizes of a, b, and x incompatible	*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

int linsys_gauss(const dmatrix *a, const dvector *b, dvector *x) {
  int i, k, n, imax, elim, *pi, err;
  double *pa, *pb, *pw, pivot, f;
  ivector ipv;
  dmatrix w;

  n = a->nrow;
  if (n != a->ncol || n != b->n || n != x->n) return (MTH_ILL);

  /* allocate memory for sort vector and workspace matrix */
  if (ivector_create(&ipv, 0, n - 1) != 0 
  || dmatrix_create(&w, 0, n - 1, 0, n) != 0) return (MTH_NOMEM);

  for (i = 0, pw = w.data, pa = a->data, pb = b->data, pi = ipv.data,
  err = MTH_GOOD; i < n; i++) {
    *pi++ = i;				/* preset pivot table */
    for (k = 0; k < n; k++) *pw++ = *pa++;/* copy a to workspace */
    *pw++ = *pb++;			/* copy b to workspace */
  }

  for (elim = 0; elim < n; elim++) {	/* begin Gauss elimination */
    imax = elim;			/* find pivot element */
    pivot = w.a[ipv.a[imax]][elim];
    for (i = elim + 1; i < n; i++) {
      if (fabs(w.a[ipv.a[i]][elim]) > fabs(pivot)) {
	imax = i;
	pivot = w.a[ipv.a[i]][elim];
      }
    }
    if (pivot == 0.0) {
      err = MTH_FAIL;
      break;
    }
    pi = &ipv.a[elim];
    i = *pi;				/* update pivot table */
    *pi = ipv.a[imax];
    ipv.a[imax] = i;
    for (i = 0; i < n; i++) {		/* eliminations */
      if (i != *pi) {
	f = w.a[i][elim] / pivot;
	for (k = elim; k <= n; k++) w.a[i][k] -= w.a[*pi][k] * f;
      }
    }
  }

  if (err == MTH_GOOD) {		/* move results to x */
    for (i = 0, pa = x->data; i < n; i++) {
      k = ipv.a[i];
      *pa++ = w.a[k][n] / w.a[k][i];
    }
  }
  ivector_free(&ipv);
  dmatrix_free(&w);
  return (err);
}
