/*************************************************************20.12.2001*/
/* matrix operations							*/
/************************************************************************/
#define _XOPEN_SOURCE 600
#include <stdlib.h>
#include <math.h>
#include <mathc.h>

/*************************************************************03.09.2001*/
/* dmatrix_mul: matrix product c = a * b				*/
/*======================================================================*/
/* a, b		(i)  input matrices (a[l][m], b[m][n])
   c[l][n]	(o)  output matrix	
   RESULT	(o)  error code:
		       MTH_ILL = incompatible matrix dimensions
		       Note that we check for compatible row and colum
		       *numbers* only, not for agreement of index ranges.*/
/************************************************************************/
int dmatrix_mul(const dmatrix *a, const dmatrix *b, dmatrix *c) {
  int i, j, k;
  double sum, *pa, *pb, *pc, *ar, *bc;

  if (a->ncol != b->nrow || a->nrow != c->nrow || b->ncol != c->ncol)
  return (MTH_ILL);

  for (i = 0, ar = a->data, pc = c->data; i < a->nrow; i++, ar += a->ncol) {
    for (j = 0, bc = b->data; j < b->ncol; j++, bc++) {
      for (k = 0, sum = 0.0, pa = ar, pb = bc; k < a->ncol; k++) {
	sum += *pa++ * *pb;
	pb += b->ncol;
      }
      *pc++ = sum;
    }
  }
  return (MTH_GOOD);
}

/*************************************************************03.09.2001*/
/* dmatrix_mult: matrix product c = a * transpose(b)			*/
/*======================================================================*/
/* a, b		(i)  input matrices (a[l][m], b[n][m])
   c[l][n]	(o)  result matrix
   RESULT	(o)  error code:
		       MTH_ILL = incompatible matrix dimensions		*/
/************************************************************************/
int dmatrix_mult(const dmatrix *a, const dmatrix *b, dmatrix *c) {
  int i, j ,k;
  double sum, *pa, *ar, *pb, *br, *pc;

  if (a->ncol != b->ncol || a->nrow != c->nrow || b->nrow != c->ncol)
  return (MTH_ILL);

  for (i = 0, ar = a->data, pc = c->data; i < a->nrow; i++, ar += a->ncol) {
    for (j = 0, br = b->data; j < b->nrow; j++, br += b->ncol) {
      for (k = 0, sum = 0.0, pa = ar, pb = br; k < a->ncol; k++)
      sum += *pa++ * *pb++;
      *pc++ = sum;
    }
  }
  return (MTH_GOOD);
}

/*************************************************************12.09.2004*/
/* dmatrix_muls: matrix scalar product c = a * s			*/
/*======================================================================*/
/* a		(i)  input matrix (a[l][m])
   s		(i)  scalar
   c		(o)  output matrix
   RESULT	(o)  error code:
		       MTH_ILL = incompatible matrix dimensions		*/
/************************************************************************/
int dmatrix_muls(const dmatrix *a, const double s, dmatrix *c) {
  int i, j;
  double sum, *pa, *pc, *pcmax;

  if (a->n != c->n) return (MTH_ILL);

  pa = a->data;
  pc = c->data;
  pcmax = pc + c->n;
  while (pc < pcmax) *pc++ = *pa++ * s;
  return (MTH_GOOD);
}

/*************************************************************20.12.2001*/
/* dmatrix_mulv: matrix vector product c = a * b			*/
/*======================================================================*/
/* a		(i)  input matrix (a[l][m])
   b		(i)  input vector (b[m])
   c		(o)  output vector (c[l])
   RESULT	(o)  error code:
		       MTH_ILL = incompatible matrix dimensions		*/
/************************************************************************/
int dmatrix_mulv(const dmatrix *a, const dvector *b, dvector *c) {
  int i, j;
  double sum, *pa, *pb, *pc;

  if (a->ncol != b->n || a->nrow != c->n) return (MTH_ILL);

  for (i = 0, pa = a->data, pc = c->data; i < c->n; i++) {
    for (j = 0, pb = b->data, sum = 0.0; j < b->n; j++) sum += *pa++ * *pb++;
    *pc++ = sum;
  }
  return (MTH_GOOD);
}

/*************************************************************20.12.2001*/
/* dmatrix_set: set matrix to fixed value				*/
/*======================================================================*/
/* x		(i)  scalar
   b		(o)  target matrix					*/
/************************************************************************/
void dmatrix_set(double x, dmatrix *b) {
  double *pb, *pmax;

  pb = b->data;
  pmax = pb + b->n;
  while (pb < pmax) *pb++ = x;
}

/*************************************************************24.08.2004*/
/* dmatrix_set_u: set matrix to unit matrix				*/
/*======================================================================*/
/* b		(o)  target matrix					*/
/************************************************************************/
void dmatrix_set_u(dmatrix *b) {
  double *pb, *pmax;

  pb = b->data;
  pmax = pb + b->n;
  while (pb < pmax) *pb++ = 0.0;
  pb = b->data;
  while (pb < pmax) {
    *pb = 1.0;
    pb += b->nrow + 1;
  }
}

/*************************************************************20.12.2001*/
/* dmatrix_copy: copy matrix to another (existing) matrix		*/
/*======================================================================*/
/* a		(i)  input matrix
   b		(o)  target matrix
   RESULT	(o)  error code:
		       MTH_ILL = incompatible matrix dimensions		*/
/************************************************************************/
int dmatrix_copy(const dmatrix *a, dmatrix *b) {
  double *pa, *pb, *pmax;

  if (a->ncol != b->ncol || a->nrow != b->nrow) return (MTH_ILL);
  pa = a->data;
  pb = b->data;
  pmax = pb + b->n;
  while (pb < pmax) *pb++ = *pa++;
  return (MTH_GOOD);
}

/*************************************************************20.12.2001*/
/* dmatrix_dot: sum of element products (analogous to scalar product)	*/
/*======================================================================*/
/* a, b		(i)  input matrices
   RESULT	(o)  Sum_ik(a_ik * b_ik)
		     Note: We do not check for size compatibility here!	*/
/************************************************************************/
double dmatrix_dot(const dmatrix *a, const dmatrix *b) {
  double *pa, *pb, *pmax, h;

  pa = a->data;
  pb = b->data;
  pmax = pb + b->n;
  h = 0.0;
  while (pb < pmax) h += *pb++ * *pa++;
  return (h);
}

/*************************************************************03.09.2001*/
/* dmatrix_invert: in-situ matrix inversion by means of the Gauss-Jordan*/
/*		   algorithm						*/
/*======================================================================*/
/* a		(io) matrix to be inverted/result matrix
   RESULT	(o)  error code:
		       MTH_FAIL = matrix is singular
		       MTH_NOMEM = memory allocation problem
		       MTH_ILL = not a square matrix			*/
/************************************************************************/
int dmatrix_invert(dmatrix *a) {
  int i, k, elim, n, imax, kmax, *pi, *pk, err;
  double pivot, f, **pc;
  ivector ipv, kpv;			/* pivot element coordinates */
  dvector b;				/* sort space */
  dmatrix c;

  n = a->nrow;
  if (n != a->ncol) return (MTH_ILL);

  if ((pc = (double **) malloc((size_t) (n * sizeof(double *)))) == NULL)
  return (MTH_NOMEM);			/* give matrix index ranges 0:n-1 */
  dmatrix_overlay(&c, pc, a->data, 0, n - 1, 0, n - 1);
  if (ivector_create(&ipv, 0, n - 1) != 0
  || ivector_create(&kpv, 0, n - 1) != 0
  || dvector_create(&b, 0, n - 1) != 0) return (MTH_NOMEM);

  for (i = 0; i < n; i++) ipv.a[i] = kpv.a[i] = i; /* preset pivot list */

  for (elim = 0, err = MTH_GOOD; elim < n; elim++) {/* begin elimination */
    imax = kmax = elim;			/* find pivot element */
    pivot = c.a[ipv.a[imax]][kpv.a[imax]];
    for (i = elim; i < n; i++) {
      for (k = elim; k < n; k++) {
	if (fabs(c.a[ipv.a[i]][kpv.a[k]]) > fabs(pivot)) {
	  imax = i;
	  kmax = k;
	  pivot = c.a[ipv.a[i]][kpv.a[k]];
        }
      }
    }
    if (pivot == 0.0) {			/* matrix singular? */
      err = MTH_FAIL;
      break;
    }
    pi = &ipv.a[elim];
    i = *pi;				/* record pivot indices */
    *pi = ipv.a[imax];			/* ... pivot row */
    ipv.a[imax] = i;
    pk = &kpv.a[elim];
    k = *pk;
    *pk = kpv.a[kmax];			/* ... pivot column */
    kpv.a[kmax] = k;
    for (i = 0; i < n; i++) {		/* Gauss-Jordan transformation */
      if (i != *pi) {
	f = c.a[i][*pk] / pivot;
	for (k = 0; k < n; k++) {
	  if (k != *pk)
	    c.a[i][k] -= c.a[*pi][k] * f;
	  else
	    c.a[i][k] = f;
	}
      }
    }
    for (k = 0; k < n; k++) {
      if (k != *pk) c.a[*pi][k] /= -pivot;
    }
    c.a[*pi][*pk] = 1.0 / pivot;
  }

  if (err == MTH_GOOD) {
    for (i = 0; i < n; i++) {		/* sort matrix elements */
      for (k = 0, pi = &ipv.a[0], pk = &kpv.a[0]; k < n; k++)
      b.a[*pi++] = c.a[i][*pk++];
      for (k = 0; k < n; k++) c.a[i][k] = b.a[k];
    }
    for (k = 0; k < n; k++) {
      for (i = 0, pi = &ipv.a[0], pk = &kpv.a[0]; i < n; i++)
      b.a[*pk++] = c.a[*pi++][k];
      for (i = 0; i < n; i++) c.a[i][k] = b.a[i];
    }
  }
  ivector_free(&ipv);
  ivector_free(&kpv);
  dvector_free(&b);
  free((void *) pc);
  return (err);
}

/*************************************************************03.09.2001*/
/* dmatrix_create: allocate a double matrix with user-defined subscript	*/
/*		   range: m[nrl..nrh][ncl..nch]				*/
/* Ref.: W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,	*/
/* Numerical Recipes in C, Cambridge University Press, 1992		*/
/*======================================================================*/
/* m		(o)  matrix to be allocated
   nrl, nrh	(i)  lower and upper row indices
   ncl, nch	(i)  lower and upper column indices
   RESULT	(o)  error code:
		       MTH_NOMEM = memory allocation failed		*/
/************************************************************************/
int dmatrix_create(dmatrix *m, int nrl, int nrh, int ncl, int nch) {
  int i;
  const size_t
    size_d = sizeof(double),
    size_dp = sizeof(double *),
    size_vp = sizeof(void *),
    align = (size_vp > size_d) ? size_vp : size_d;
  
  m->nrl = nrl;
  m->nrh = nrh;
  m->nrow = nrh - nrl + 1;
  m->ncl = ncl;
  m->nch = nch;
  m->ncol = nch - ncl + 1;
  m->n = m->nrow * m->ncol;
#ifdef USE_VALLOC			/* use old (pre-POSIX) method */
  if ((m->data = (double *) valloc(m->n * size_d)) == NULL
  || (m->row = (double **) valloc((m->nrow * size_dp))) == NULL)
  return (MTH_NOMEM);			/* space for data & row pointers */
#else					/* POSIX-style method */
  if (posix_memalign((void **) &m->data, align, size_d * m->n)
  || posix_memalign((void **) &m->row, align, size_dp * m->nrow))
  return (MTH_NOMEM);			/* space for data & row pointers */
#endif

  m->a = m->row - nrl;
  m->row[0] = m->data - ncl;
  for (i = 1; i < m->nrow; i++) m->row[i] = m->row[i - 1] + m->ncol;
  return (MTH_GOOD);
}

/*************************************************************03.09.2001*/
/* dmatrix_free: deallocate a double matrix				*/
/*======================================================================*/
/* m		(o)  matrix to be deallocated				*/
/************************************************************************/
void dmatrix_free(dmatrix *m) {
  free((char *) m->data);
  free((char *) m->row);
}

/*************************************************************03.09.2001*/
/* imatrix_create: allocate an int matrix with user-defined subscript	*/
/*		   range: m[nrl..nrh][ncl..nch]				*/
/* Ref.: W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,	*/
/* Numerical Recipes in C, Cambridge University Press, 1992		*/
/*======================================================================*/
/* m		(o)  matrix to be allocated
   nrl, nrh	(i)  lower and upper row indices
   ncl, nch	(i)  lower and upper column indices
   RESULT	(o)  error code:
		       MTH_NOMEM = memory allocation failed		*/
/************************************************************************/
int imatrix_create(imatrix *m, int nrl, int nrh, int ncl, int nch) {
  int i;
  const size_t
    size_i = sizeof(int),
    size_ip = sizeof(int *),
    size_vp = sizeof(void *),
    align = (size_vp > size_i) ? size_vp : size_i;
  
  m->nrl = nrl;
  m->nrh = nrh;
  m->nrow = nrh - nrl + 1;
  m->ncl = ncl;
  m->nch = nch;
  m->ncol = nch - ncl + 1;
  m->n = m->nrow * m->ncol;
#ifdef USE_VALLOC			/* use old (pre-POSIX) method */
  if ((m->data = (int *) valloc(m->n * size_i)) == NULL
  || (m->row = (int **) valloc((m->nrow * size_ip))) == NULL)
  return (MTH_NOMEM);			/* space for data & row pointers */
#else					/* POSIX-style method */
  if (posix_memalign((void **) &m->data, align, size_i * m->n)
  || posix_memalign((void **) &m->row, align, size_ip * m->nrow))
  return (MTH_NOMEM);			/* space for data & row pointers */
#endif

  m->a = m->row - nrl;
  m->row[0] = m->data - ncl;
  for (i = 1; i < m->nrow; i++) m->row[i] = m->row[i - 1] + m->ncol;
  return (MTH_GOOD);
}

/*************************************************************03.09.2001*/
/* imatrix_free: deallocate an integer matrix				*/
/*======================================================================*/
/* m		(o)  matrix to be deallocated				*/
/************************************************************************/
void imatrix_free(imatrix *m) {
  free((char *) m->data);
  free((char *) m->row);
}

/*************************************************************03.09.2001*/
/* dmatrix_overlay: overlay a double array with a matrix		*/
/*									*/
/* This is almost the same as dmatrix_create(), except that here no	*/
/* dynamic memory allocations are used.					*/
/*									*/
/* usage:								*/
/*   dmatrix a;								*/
/*   double *(apointer[3]), aspace[12];					*/
/*   dmatrix_overlay(&a, apointer, aspace, -1, 1, 1, 4);		*/
/*   h = a.a[-1][1] + a.a[-1][4] + a.a[1][1] + a.a[1][4];		*/
/*									*/
/* WARNING: The array boundary ranges must match the array size. There	*/
/*	    is no way to check this at runtime!				*/
/*======================================================================*/
/* m		(o)   dmatrix to be generated
   p		(io)  array of pointers (1 for each row)
   a		(i)   array to be organized as a matrix
   nrl, nrh	(i)   lower and upper row boundaries
   ncl, nch	(i)   lower and upper column boundaries			*/
/************************************************************************/
void dmatrix_overlay(dmatrix *m, double **p, double *a, int nrl, int nrh,
int ncl, int nch) {
  int i;

  m->nrl = nrl;
  m->nrh = nrh;
  m->nrow = nrh - nrl + 1;
  m->ncl = ncl;
  m->nch = nch;
  m->ncol = nch - ncl + 1;
  m->n = m->nrow * m->ncol;
  m->data = a;
  m->row = p;

  m->a = m->row - nrl;
  m->row[0] = m->data - ncl;
  for (i = 1; i < m->nrow; i++) m->row[i] = m->row[i - 1] + m->ncol;
}

/*************************************************************03.09.2001*/
/* imatrix_overlay: overlay an integer array with a matrix		*/
/*									*/
/* usage: see dmatrix_overlay()						*/
/*======================================================================*/
/* m		(o)   imatrix to be generated
   p		(io)  array of pointers (1 for each row)
   a		(i)   array to be organized as a matrix
   nrl, nrh	(i)   lower and upper row boundaries
   ncl, nch	(i)   lower and upper column boundaries			*/
/************************************************************************/
void imatrix_overlay(imatrix *m, int **p, int *a, int nrl, int nrh,
int ncl, int nch) {
  int i;

  m->nrl = nrl;
  m->nrh = nrh;
  m->nrow = nrh - nrl + 1;
  m->ncl = ncl;
  m->nch = nch;
  m->ncol = nch - ncl + 1;
  m->n = m->nrow * m->ncol;
  m->data = a;
  m->row = p;

  m->a = m->row - nrl;
  m->row[0] = m->data - ncl;
  for (i = 1; i < m->nrow; i++) m->row[i] = m->row[i - 1] + m->ncol;
}
