/*************************************************************07.05.2005*/
/* minimum1: find the global minimum of a 1-dimensional function	*/
/*									*/
/* An interval is scanned with a step size that is passed as a para-	*/
/* meter. If a minimum is detected, i.e. if an y value is found that is	*/
/* lower than the y values at adjacent positions, this minimum is	*/
/* located by iterative parabola fitting.				*/
/*======================================================================*/
/* f		(i)  function fo which the minimum is sought
   s		(i)  structure for passing additional data to f()
   a, b		(i)  interval for the search (a < b!)
   step		(i)  step size of initial search
   eps		(i)  desired accuracy of xmin
   xmin, ymin	(o)  abscissa and ordinate of minimum
   RESULT	(o)  error code:
		       MTH_ILL = no relative minimum; one of the
				 interval boundaries is returned
		       MTH_RANGE = search went astray (step too large?)	*/
/************************************************************************/
#include <math.h>
#include <mathc.h>

int minimum1(double (*f)(), void *s, double a, double b, double step,
double eps, double *xmin, double *ymin) {
  int err, localmin;
  double dx, det, a1, a2;
  struct point {
    double x, y;
  } p0, p1, p2, px;

  p0.x = *xmin = a;			/* start scan at left boundary */
  p0.y = *ymin = (*f)(a, s);
  err = MTH_ILL;
  if ((p1.x = a + step) >= b) return (MTH_RANGE);
  p1.y = (*f)(p1.x, s);

  while (p1.x < b) {			/* main scanning loop */
    p2.x = MIN(p1.x + step, b);
    p2.y = (*f)(p2.x, s);

    do {				/* construct parabola */
      localmin = 0;
      det = (p2.x - p1.x) * (p0.x - p1.x) * (p2.x - p0.x);
      a2 = ((p2.y - p1.y) * (p0.x - p1.x) - (p0.y - p1.y)
      * (p2.x - p1.x)) / det;
      if (a2 <= 0.0) break;		/* wrong curvature? */
      a1 = ((p0.y - p1.y) * SQ(p2.x - p1.x) - (p2.y - p1.y)
      * SQ(p0.x - p1.x)) / det;

      dx = -0.5 * a1 / a2;		/* calculate the parabola minimum */
      px.x = p1.x + dx;
      if (px.x <= p0.x || px.x >= p2.x) break;
      px.y = (*f)(px.x, s);
      if (px.x < p1.x) {
        if (px.y < p0.y && px.y < p1.y) {
	  p2 = p1;
	  p1 = px;
	  localmin = 1;
	}
	else
	  break;
      }
      else {
        if (px.y < p1.y && px.y < p2.y) {
	  p0 = p1;
	  p1 = px;
	  localmin = 1;
	}
	else
	  break;
      }
    } while (fabs(dx) > eps);

    if (localmin) {
      if (p1.y < *ymin) {		/* compare with previous minimum */
	*xmin = p1.x;
	*ymin = p1.y;
	err = MTH_GOOD;
      }
      p0 = p1;				/* resume scanning */
      p1.x = MIN(p0.x + step, b);
      p1.y = (*f)(p1.x, s);
    }
    else {
      p0 = p1;
      p1 = p2;
    }
  }

  if (p1.y < *ymin) {			/* minimum on interval boundary? */
    *xmin = p1.x;
    *ymin = p1.y;
    err = MTH_ILL;
  }
  return (err);
}
