/*************************************************************08.03.2004*/
/* ode_runge_kutta: ordinary differential equation solver using the	*/
/*		    method of Runge and Kutta				*/
/*		    with automatic control of the step size		*/
/*									*/
/* G. Jordan-Engeln, F. Reutter, Formelsammlung zur Numerischen		*/
/* Mathematik, BI Hochschultaschenbcher Vol. 106, Bibliographisches	*/
/* Institut, Mannheim 1976.						*/
/*======================================================================*/
/* f(x, y, s)	(i)  function defining the ODE, i.e. dy/dx = f(x, y, s).
   s		(i)  structure containing pass-through arguments for f()
   x0, y0	(i)  initial values
   xend		(i)  end of integration interval
   yend		(o)  function value at xend				*/
/************************************************************************/
#include <math.h>
#include <mathc.h>


#define CRIT_HI		0.2		/* Lipshitz criterion for step size */
#define CRIT_LO		0.005


void ode_runge_kutta(double (*f)(), void *s, double x0, double y0,
double xend, double *yend) {
  int at_max;
  double h, yh[4], crit;

  h = xend - x0;			/* largest possible step size */
  at_max = 1;
  while (h * (x0 - xend) < 0.0) {
    while (1) {
      if ((x0 + h - xend) * h > 0.0) {	/* make sure not exceed xend */
	h = xend - x0;
	at_max = 1;
      }
      yh[0] = f(x0, y0, s) * h;
      yh[1] = f(x0 + 0.5 * h, y0 + 0.5 * yh[0], s) * h;
      yh[2] = f(x0 + 0.5 * h, y0 + 0.5 * yh[1], s) * h;
      yh[3] = f(x0 + h, y0 + yh[2], s) * h;
      crit = 2 * fabs((yh[1] - yh[2]) / (yh[0] - yh[1]));
      if (crit > CRIT_HI) {		/* step size adjustment */
	h *= 0.5;
	at_max = 0;
      }
      else if (crit < CRIT_LO && ! at_max)
	h *= 2.0;
      else
	break;
    }
    x0 += h;
    y0 += (yh[0] + 2 * yh[1] + 2 * yh[2] + yh[3]) / 6.0;
  }
  *yend = y0;
}
