/*************************************************************14.08.1998*/
/* ran_ecuyer: random number generator using l'Ecuyer's method		*/
/*									*/
/* This routine uses a double congruential algorithm described by	*/
/* l'Ecuyer. In order to destroy sequential correlation, the shuffling	*/
/* table method is used in addition.					*/
/* Ref.: W.H. Press, B.P. Flannery, S.A. Teukolsky, W.T. Vetterling,	*/
/*	 "Numerical Recipes", Cambridge University Press, Cambridge	*/
/*	 1986, section 7.1						*/
/*	 P. l'Ecuyer, ACM Comm. 31, 742 (1988)				*/
/************************************************************************/
#include <time.h>
#include <math.h>
#include <mathc.h>

#define M1	2147483563		/* constants of 1st generator */
#define A1	40014			/* note: M1 = A1*Q1 + R1 */
#define Q1	53668
#define R1	12211
#define M2	2147483399		/* constants of 2nd generator */
#define A2	40692
#define Q2	52774
#define R2	3791
#define NTAB	32			/* shuffling table size */
#define NDIV	(1 + (M1 - 1)/NTAB)

double ran_ecuyer(void) {
  int j, k;
  static int irand, i1, i2, v[NTAB], linit = 0;

  if (linit == 0) {			/* initialize subroutine */
    linit = 1;
    i1 = time(NULL);			/* get seconds after 1900 */
    if (i1 % 2 == 0) i1--;
    if (i1 % 3 == 0) i1 -= 2;
    for (j = 0; j < NTAB; j++) {
      k = i1 / Q1;
      i1 = A1 * (i1 - k * Q1) - k * R1;
      if (i1 < 0) i1 += M1;
      v[j] = i1;
    }
    irand = v[NTAB-1];
  }

  k = i1 / Q1;
  i1 = A1 * (i1 - k * Q1) - k * R1;
  if (i1 < 0) i1 += M1;			/* 1st random number */
  k = i2 / Q2;
  i2 = A2 * (i2 - k * Q2) - k * R2;
  if (i2 < 0) i2 += M2;			/* 2nd random number */

  /* calculation of a random number without sequential correlation:
     use previous random number to select a table entry */
  k = irand / NDIV;			/* table index */
  irand = v[k] - i2;
  v[k] = i1;
  if (irand < 1) irand += M1 - 1;
  return (((double) irand) / ((double) M1));
}
