/*************************************************************20.12.2001*/
/* vector operations							*/
/************************************************************************/
#define _XOPEN_SOURCE 600
#include <stdlib.h>
#include <math.h>
#include <mathc.h>

/*************************************************************20.12.2001*/
/* vector_ops: simple vector operations					*/
/*======================================================================*/
/* v1, v2	(i)  input vectors					*/
/* v3		(o)  output vector					*/
/* a		(i)  scalar operand					*/
/* f(x)		(i)  1-dimensional function				*/
/* RESULT	(o)  error code: 0 = no error				*/
/*				 1 = incompatible vector dimensions	*/
/*		     Note: Only vector sizes are compared, not index	*/
/*		     ranges.						*/
/*									*/
/* operations:	+			    subroutine:	add		*/
/*		-					sub		*/
/*		* (element by element)			mul		*/
/*		/					div		*/
/*		scalar multiplication v3 = a * v1	smul		*/
/* 		scalar addition				sadd		*/
/*		scalar assignment v3=(a)		set		*/
/*		copy					copy		*/
/*		function evaluation v3=f(v1)		vfunc		*/
/*		dot product a = v1 . v2			dot		*/
/*		cross product				cross		*/
/************************************************************************/
int dvector_add(const dvector *v1, const dvector *v2, dvector *v3) {
  double *p1, *p2, *p3, *pmax;

  if (v1->n != v2->n || v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p2 = v2->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++ + *p2++;
  return (MTH_GOOD);
}
 
int dvector_sub(const dvector *v1, const dvector *v2, dvector *v3) {
  double *p1, *p2, *p3, *pmax;

  if (v1->n != v2->n || v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p2 = v2->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++ - *p2++;
  return (MTH_GOOD);
}
 
int dvector_mul(const dvector *v1, const dvector *v2, dvector *v3) {
  double *p1, *p2, *p3, *pmax;

  if (v1->n != v2->n || v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p2 = v2->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++ * *p2++;
  return (MTH_GOOD);
}
 
int dvector_div(const dvector *v1, const dvector *v2, dvector *v3) {
  double *p1, *p2, *p3, *pmax;

  if (v1->n != v2->n || v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p2 = v2->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++ / *p2++;
  return (MTH_GOOD);
}
 
int dvector_smul(const dvector *v1, double a, dvector *v3) {
  double *p1, *p3, *pmax;

  if (v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++ * a;
  return (MTH_GOOD);
}
 
int dvector_sadd(const dvector *v1, double a, dvector *v3) {
  double *p1, *p3, *pmax;

  if (v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++ + a;
  return (MTH_GOOD);
}
 
int dvector_set(double a, dvector *v3) {
  double *p3, *pmax;

  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = a;
  return (MTH_GOOD);
}
 
int dvector_copy(const dvector *v1, dvector *v3) {
  double *p1, *p3, *pmax;

  if (v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = *p1++;
  return (MTH_GOOD);
}

int dvector_func(double (*f)(), const dvector *v1, dvector *v3) {
  double *p1, *p3, *pmax;

  if (v1->n != v3->n) return (MTH_ILL);
  p1 = v1->data;
  p3 = v3->data;
  pmax = p3 + v3->n;
  while (p3 < pmax) *p3++ = (*f)(*p1++);
  return (MTH_GOOD);
}

double dvector_dot(const dvector *v1, const dvector *v2) {
  double *p1, *p2, *pmax, a;

  p1 = v1->data;
  pmax = p1 + v1->n;
  p2 = v2->data;
  a = 0.0;
  while (p1 < pmax) a += *p1++ * *p2++;	/* assume that v1->n == v2->n */
  return (a);
}

int dvector_cross(const dvector *v1, const dvector *v2, dvector *v3) {
  if (v1->n != 3 || v2->n != 3 || v3->n != 3) return (MTH_ILL);

  v3->data[0] = v1->data[1] * v2->data[2] - v1->data[2] * v2->data[1];
  v3->data[1] = v1->data[2] * v2->data[0] - v1->data[0] * v2->data[2];
  v3->data[2] = v1->data[0] * v2->data[1] - v1->data[1] * v2->data[0];
  return (MTH_GOOD);
}

/*************************************************************03.09.2001*/
/* dvector_create: allocate a double vector with user-defined subscript	*/
/*	    range: v[nl..nh]						*/
/*======================================================================*/
/* v		(o)  vector to be allocated
   nl, nh	(i)  lower and upper indices
   RESULT	(o)  error code:
		       MTH_NOMEM = memory allocation failure		*/
/************************************************************************/
int dvector_create(dvector *v, int nl, int nh) {
  const size_t
    size_d = sizeof(double),
    size_vp = sizeof(void *),
    align = (size_vp > size_d) ? size_vp : size_d;

  v->nl = nl;
  v->nh = nh;
  v->n = nh - nl + 1;
#ifdef USE_VALLOC			/* use old (pre-POSIX) method */
  if ((v->data = (double *) valloc(v->n * size_d)) == NULL)
  return (MTH_NOMEM);
#else					/* POSIX method */
  if (posix_memalign((void **) &v->data, align, v->n * size_d))
  return (MTH_NOMEM);
#endif
  v->a = v->data - nl;
  return (MTH_GOOD);
}

/*************************************************************03.09.2001*/
/* dvector_free: deallocate a double vector				*/
/*======================================================================*/
/* v		(i)  vector to be deallocated				*/
/************************************************************************/
void dvector_free(dvector *v) {
  free((char *) v->data);
}

/*************************************************************03.09.2001*/
/* ivector_create: allocate an int vector with user-defined subscript	*/
/*	    range: v[nl..nh]						*/
/*======================================================================*/
/* v		(o)  vector to be allocated
   nl, nh	(i)  lower and upper indices
   RESULT	(o)  error code:
		       MTH_NOMEM = memory allocation failure		*/
/************************************************************************/
int ivector_create(ivector *v, int nl, int nh) {
  const size_t
    size_i = sizeof(int),
    size_vp = sizeof(void *),
    align = (size_vp > size_i) ? size_vp : size_i;

  v->nl = nl;
  v->nh = nh;
  v->n = nh - nl + 1;
#ifdef USE_VALLOC			/* use old (pre-POSIX) method */
  if ((v->data = (int *) valloc(v->n * size_i)) == NULL) return (MTH_NOMEM);
#else					/* POSIX method */
  if (posix_memalign((void **) &v->data, align, v->n * size_i))
  return (MTH_NOMEM);
#endif
  v->a = v->data - nl;
  return (MTH_GOOD);
}

/*************************************************************03.09.2001*/
/* ivector_free: deallocate an integer vector				*/
/*======================================================================*/
/* v		(i)  vector to be deallocated				*/
/************************************************************************/
void ivector_free(ivector *v) {
  free((char *) v->data);
}

/*************************************************************03.09.2001*/
/* dvector_overlay: overlay a double array with a vector		*/
/*									*/
/* This is almost the same as dvector_create(), except that here no 	*/
/* dynamic memory allocations are used.					*/
/*									*/
/* usage:								*/
/*   dvector a;								*/
/*   double aspace[12];							*/
/*   dvector_overlay(&a, aspace, -2, 9);				*/
/*   h = a.a[-2] + a.a[9];						*/
/*									*/
/* WARNING: The array boundary ranges must match the array size. There	*/
/*	    is no way to check this at runtime!				*/
/*======================================================================*/
/* v		(o)   dvector to be generated
   a		(i)   array to be organized as a vector
   nl, nh	(i)   lower and upper boundaries			*/
/************************************************************************/
void dvector_overlay(dvector *m, double *a, int nl, int nh) {
  m->nl = nl;
  m->nh = nh;
  m->n = nh - nl + 1;
  m->data = a;
  m->a = m->data - nl;
}

/*************************************************************03.09.2001*/
/* ivector_overlay: overlay an integer array with a vector		*/
/*									*/
/* usage: see dvector_overlay()						*/
/*======================================================================*/
/* v		(o)   ivector to be generated
   a		(i)   array to be organized as a vector
   nl, nh	(i)   lower and upper boundaries			*/
/************************************************************************/
void ivector_overlay(ivector *m, int *a, int nl, int nh) {
  m->nl = nl;
  m->nh = nh;
  m->n = nh - nl + 1;
  m->data = a;
  m->a = m->data - nl;
}

/*************************************************************28.04.2004*/
/* dvector_resize: change the size of an existing vector		*/
/*======================================================================*/
/* v		(o)  vector to be changed
   nh		(i)  new upper index
   RESULT	(o)  error code:
		       MTH_NOMEM = memory allocation failure		*/
/************************************************************************/
int dvector_resize(dvector *v, int nh) {
  v->nh = nh;
  v->n = v->nh - v->nl + 1;
  v->data = (double *) realloc(v->data, (size_t) (v->n * sizeof(double)));
  if (v->data == NULL) return (MTH_NOMEM);
  v->a = v->data - v->nl;
  return (MTH_GOOD);
}

/*************************************************************28.04.2004*/
/* ivector_resize: change the size of an existing vector		*/
/*======================================================================*/
/* v		(o)  vector to be changed
   nh		(i)  new upper index
   RESULT	(o)  error code:
		       MTH_NOMEM = memory allocation failure		*/
/************************************************************************/
int ivector_resize(ivector *v, int nh) {
  v->nh = nh;
  v->n = v->nh - v->nl + 1;
  v->data = (int *) realloc(v->data, (size_t) (v->n * sizeof(int)));
  if (v->data == NULL) return (MTH_NOMEM);
  v->a = v->data - v->nl;
  return (MTH_GOOD);
}
